/*->c.border */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "akbd.h"
#include "bbc.h"
#include "font.h"

#include "transform.h"
#include "xprocess.h"
#include "flex.h"
#include "deb.h"


#include "wos.h"


#include "err.h"
#include "xext.h"
#include "poll.h"
#include "key.h"
#include "temp.h"
#include "colour.h"
#include "xmath.h"
#include "drag.h"
#include "conf.h"
#include "redraw.h"
#include "etc.h"
#include "bits.h"
#include "wimptext.h"
#include "filetype.h"
#include "fsx.h"
#include "bf.h"
#include "scanplus.h"
#include "dbhi.h"



#define DRAWFILE_INTERNAL
#include "DrawFile1.h"
#include "DrawLevel0.h"
#include "DrawLevel1.h"
#include "DrawFile2.h"
#include "DrawErrors.h"



#include "str.h"
#include "constants.h"
#include "reslink.h"

#include "xredraw.h"
#include "xrender.h"
#include "xxcolour.h"
#include "xcolour.h"
#include "colsep.h"
#include "object.h"

#include "file.h"
#include "view.h"
#include "viewr.h"
#include "undo.h"
#include "select.h"
#include "frameflow.h"
#include "curve.h"
#include "geometry.h"
#include "dscript.h"

#include "prstr.h"
#include "xprint.h"
#include "fxdraw.h"

#include "border.h"


/**************************************************************************/

#define UBORDERS         0xC000
#define UBORDERF         0xC001



typedef struct linestylestr
{
 int style;
 int stylescale;
 int join;
 int id;
} linestylestr;


static linestylestr deflinestyle=
{
 -1,
  0,
  0,
};



static linestylestr * linestyles;
static int            linestylen;

static char         * border;
static int            bordersize;


typedef struct borderdefstr
{
 int   offset;
 int   name;
 int   type;
} borderdefstr;



static char         * fancyborder;
static int            fancybordersize;


typedef struct fancyborderstr
{
 int     offset;
 int     length;
 int     n;
 int     object[8];
 aboxstr abox[8];
} fancyborderstr;



static borderdefstr    * borderdefs;
       int               borderdefn;



#define BORDERCHUNK    0x100
#define LINECHUNK      (8*sizeof(linestylestr))
#define BORDERDEFCHUNK (8*sizeof(borderdefstr))
#define FANCYBORDERCHUNK 0x1000



static os_error *fancysample(int x,int y,int w,int h,int d,
                             fancyborderstr *fs,int selected);



static linestylestr * findlinestyle(int id)
{
 int i;

 for(i=0;i<linestylen;i++)
 {
  if(linestyles[i].id==id)
  {
   return(&linestyles[i]);
  }
 }

 return(&deflinestyle);
}




int getborderinfo(int n,int * width)
{
 int              type;
 fancyborderstr * fs;

 type=borderdefs[n].type;

 if(type)
 {
  fs=(fancyborderstr*)(border+borderdefs[n].offset);
  if(width) *width=(fs->abox[0].x1-fs->abox[0].x0);
 }

 return(type);
}



/* x,y top left corner of frame sample */
/* w   total width of frame            */
/* h   total height                    */
/* d   border thickness                */
/* n   border number                   */


os_error * sampleborder(int x,int y,int w,int h,int d,int n,int selected)
{
 os_error     * err;
 char         * p;
 int            len;
 int            path[16];
 int          * q;
 int            c;
 int            t;
 linestylestr * style;


 startredraw();

 err=NULL;

 if(n>=0 && n<borderdefn)
 {
  p=border+borderdefs[n].offset;
  len=strlen(p);

  if(borderdefs[n].type==0)
  {
   d=d/len;

   while((c=*p)!=0)
   {
    t=0;
    while(*p==c) {t+=d;p++;}

    if(c!='_')
    {
     style=findlinestyle(c);

     q=path;

     *q++=DRAWMOVE;
     *q++=x+t/2;
     *q++=y-h;
     *q++=DRAWDRAW;
     *q++=x+t/2;
     *q++=y-t/2;
     *q++=DRAWDRAW;
     *q++=x+w-t/2;
     *q++=y-t/2;
     *q++=DRAWDRAW;
     *q++=x+w-t/2;
     *q++=y-h;
     *q++=DRAWEND;

     wimp_setcolour(selected?WIMPWHITE:WIMPBLACK);
     rdrawpath(path,t,style->style,style->stylescale,style->join);
    }

    x+=t;
    y-=t;
    w-=2*t;
    h-=t;
   }
  }
  else
  {
   err=fancysample(x,y,w,h,d,
                   (fancyborderstr*)(border+borderdefs[n].offset),selected);
  }
 }

 return(err);
}






/* a token is a non space, followed by a space or zero */

static char * gettok(char * string,int * len)
{
 int    tlen;
 char * p;
 char * q;

 p=string;
 while(*p==' ') p++;
 q=p;
 while(*q>' ') q++;

 tlen=q-p;
 if(*q)
 {
  *q=0;
  tlen++;
 }

 *len=tlen;

 return(p);
}





static os_error * parsefancyborder(char * fancyborder,int length,
                                   fancyborderstr * fs,char * name)
{
 os_error    * err;
 Draw_objhdr * hdr;
 int           offset;
 int           n;

 err=NULL;

 hdr=NULL;  /* compiler */

 memset(fs,0,sizeof(fancyborderstr));

 n=0;

 offset=sizeof(Draw_fileheader);

 while(offset<length)
 {
  hdr=(Draw_objhdr*)(fancyborder+offset);

  if(hdr->tag!=draw_OBJFONTLIST && hdr->tag!=draw_OBJOPTIONS)
  {
   fs->object[n]=offset;
   fs->abox[n]=*((aboxstr*)&hdr->bbox);
   n++;
   if(n>=8) break;
  }
  offset+=hdr->size;
 }

 fs->length=length;
 fs->n=n;

 if(n<1) err=geterrorf(EBDR1,name);

 return(err);
}





static int           ftransparent;
static colourvalue   fcv;
static filestr     * ffile;
static viewstr     * fview;


static os_error * fancyfn(int handle,void * d,void * s,int size,
                                       int dtype,int stype,cmsdatastr * scms,cmsdatastr * dcms)
{
 int         r;
 int         g;
 int         b;
 int         w;
 colourvalue cv;
 int         tint;

 if(ftransparent)
 {
  if(fview && fview->process)
  {
   ((processfn3)fview->process)((int)fview,d,s,size,dtype,stype,scms,dcms);
  }
 }
 else
 {
  if(stype==PROCESST_PAL && dtype==PROCESST_PAL)
  {
   w=*((int*)s);

   r=(w>>8)&0xFF;
   g=(w>>16)&0xFF;
   b=(w>>24)&0xFF;

   tint=255*16-((r<<2)+(b<<1)+(g<<3)+(g<<1));   /* max 255*16 */

   cv=fcv;
   cv.data.tint.tint=(cv.data.tint.tint*tint)/(255*16);

   *((int*)d)=vfindcolour2(fview,ffile,&cv);
  }
 }

 return(NULL);
 USE(handle);
}






static os_error * renderfancyborder(fancyborderstr *fs,borderdatastr * data,
                                    curvestr * curve,
                                    transformstr * tr,viewredrawstr * v)
{
 os_error    * err;
 draw_objptr   dummy;
 Draw_diag     diag;
 transformstr  tr1;
 transformstr  tr2;
 transformstr  tr3;
 transformstr  trs;
 transformstr  trx;
 transformstr  trw;
 transformstr  drtr;
 Draw_error    error;
 processstr    pr;
 Draw_box      box;
 int           w;
 int           bits;


// cprintf("fancy border");

 err=NULL;

 w=data->border[BORDER].width;

 if(data->border[SHADOW].active & BDRACT) bits=0xF;
 else                                     bits=data->border[BORDER].mode;

 ffile=v->file;
 fview=v->view;

 fcv=data->border[BORDER].colour;
 ftransparent=transparentcolour(&data->border[BORDER].colour,ffile);


 memset(&pr,0,sizeof(processstr));

 if(!ftransparent || fview->process)
 {
  pr.mode=PROCESSFN;
  pr.process=fancyfn;
 }

 pr.isprinter=v->isprinter;
// pr.vdupal=&cvdu.vdupal;
// pr.deltax=cvdu.deltax;
// pr.deltay=cvdu.deltay;
// pr.lndeltax=cvdu.lndeltax;
// pr.lndeltay=cvdu.lndeltay;
 pr.xdpi=cvdu.xdpi;
 pr.ydpi=cvdu.ydpi;
 pr.ln2bpp=cvdu.ln2bpp;


 if(pr.isprinter && (printbits & PRB_FLAT))
 {
  pr.scale=((prjobstr*)v->prjob)->scale;
 }
 else
 {
  pr.scale.x=pr.scale.y=0x10000;
 }


 pr.handle2=NULL; // view;
 pr.cmyk2rgb=(processcmyk2rgbfn)xcmyk2rgb3232;
 pr.cmyk2bgr=(processcmyk2rgbfn)xcmyk2bgr3232;


 fxdsetview(v);

 diag.data=fancyborder+fs->offset;
 diag.length=fs->length;

 dummy.bytep=0;
 note_fontcat(dummy);

 getdrtr(&drtr,tr,v);

 if(bits & BDRLEFT)
 {
  tr_setshift(&trs,-fs->abox[0].x0,-fs->abox[0].y1);
  tr_setscale(&tr2,fs->abox[0].x1-fs->abox[0].x0,w,
                   fs->abox[0].y1-fs->abox[0].y0,-curve->box.h);
  tr_multiply(&tr3,&trs,&tr2);

  tr_setshift(&tr1,curve->box.x-w,curve->box.y);
  tr_multiply(&trx,&tr3,&tr1);


  tr_multiply(&trw,&trx,&drtr);

  box=*(Draw_box*)(&fs->abox[0]);

  Draw_renderObjects(&diag,fs->object[0],fs->object[0],&box,&trw,&pr,&error);


  if(bits & 0x1)
  {                       
   tr_setscale(&tr2,fs->abox[0].x1-fs->abox[0].x0,w,
                    fs->abox[0].x1-fs->abox[0].x0,-w);
   tr_multiply(&tr3,&trs,&tr2);

   tr_setshift(&tr1,curve->box.x-w,curve->box.y);
   tr_multiply(&trx,&tr3,&tr1);


   tr_multiply(&trw,&trx,&drtr);


   box=*(Draw_box*)(&fs->abox[1]);

   Draw_renderObjects(&diag,fs->object[1],Draw_LastObject,&box,&trw,
                                                          &pr,&error);
  }
 }


 if(bits & BDRTOP)
 {
  tr_setshift(&tr1,-fs->abox[0].x1,-fs->abox[0].y1);

  tr_setidentity(&tr2);
  tr_setrotationa(&tr2,-90*0x10000);

  tr_multiply(&trs,&tr1,&tr2);


  tr_setscale(&tr2,fs->abox[0].y1-fs->abox[0].y0,curve->box.w,
                   fs->abox[0].x1-fs->abox[0].x0,-w);
  tr_multiply(&tr3,&trs,&tr2);

  tr_setshift(&tr1,curve->box.x+curve->box.w,curve->box.y);
  tr_multiply(&trx,&tr3,&tr1);

  tr_multiply(&trw,&trx,&drtr);

  box=*(Draw_box*)(&fs->abox[0]);

  Draw_renderObjects(&diag,fs->object[0],fs->object[0],&box,&trw,&pr,&error);


  if(bits & 0x2)
  {
   tr_setscale(&tr2,fs->abox[0].x1-fs->abox[0].x0,w,
                    fs->abox[0].x1-fs->abox[0].x0,-w);
   tr_multiply(&tr3,&trs,&tr2);

   tr_setshift(&tr1,curve->box.x+curve->box.w,curve->box.y);
   tr_multiply(&trx,&tr3,&tr1);


   tr_multiply(&trw,&trx,&drtr);


   box=*(Draw_box*)(&fs->abox[1]);

   Draw_renderObjects(&diag,fs->object[1],Draw_LastObject,&box,&trw,
                                                          &pr,&error);
  }
 }




 if(bits & BDRRIGHT)
 {
  tr_setshift(&tr1,-fs->abox[0].x1,-fs->abox[0].y1);

  tr_setidentity(&tr2);
  tr_setrotationa(&tr2,-180*0x10000);

  tr_multiply(&trs,&tr1,&tr2);


  tr_setscale(&tr2,fs->abox[0].x1-fs->abox[0].x0,w,
                   fs->abox[0].y1-fs->abox[0].y0,-curve->box.h);
  tr_multiply(&tr3,&trs,&tr2);

  tr_setshift(&tr1,curve->box.x+curve->box.w,curve->box.y+curve->box.h);
  tr_multiply(&trx,&tr3,&tr1);


  tr_multiply(&trw,&trx,&drtr);

  box=*(Draw_box*)(&fs->abox[0]);

  Draw_renderObjects(&diag,fs->object[0],fs->object[0],&box,&trw,&pr,&error);

  if(bits & 0x4)
  {
   tr_setscale(&tr2,fs->abox[0].x1-fs->abox[0].x0,w,
                    fs->abox[0].x1-fs->abox[0].x0,-w);
   tr_multiply(&tr3,&trs,&tr2);

   tr_setshift(&tr1,curve->box.x+curve->box.w,curve->box.y+curve->box.h);
   tr_multiply(&trx,&tr3,&tr1);


   tr_multiply(&trw,&trx,&drtr);


   box=*(Draw_box*)(&fs->abox[1]);

   Draw_renderObjects(&diag,fs->object[1],Draw_LastObject,&box,&trw,
                                                          &pr,&error);
  }
 }



 if(bits & 0x4)
 {
  tr_setshift(&tr1,-fs->abox[0].x0,-fs->abox[0].y1);

  tr_setidentity(&tr2);
  tr_setrotationa(&tr2,90*0x10000);

  tr_multiply(&trs,&tr1,&tr2);


  tr_setscale(&tr2,fs->abox[0].y1-fs->abox[0].y0,curve->box.w,
                   fs->abox[0].x1-fs->abox[0].x0,-w);
  tr_multiply(&tr3,&trs,&tr2);

  tr_setshift(&tr1,curve->box.x,curve->box.y+curve->box.h+w);
  tr_multiply(&trx,&tr3,&tr1);


  tr_multiply(&trw,&trx,&drtr);

  box=*(Draw_box*)(&fs->abox[0]);

  Draw_renderObjects(&diag,fs->object[0],fs->object[0],&box,&trw,&pr,&error);


  if(bits & 0x8)
  {
   tr_setscale(&tr2,fs->abox[0].x1-fs->abox[0].x0,w,
                    fs->abox[0].x1-fs->abox[0].x0,-w);
   tr_multiply(&tr3,&trs,&tr2);

   tr_setshift(&tr1,curve->box.x,curve->box.y+curve->box.h+w);
   tr_multiply(&trx,&tr3,&tr1);


   tr_multiply(&trw,&trx,&drtr);


   box=*(Draw_box*)(&fs->abox[1]);

   Draw_renderObjects(&diag,fs->object[1],Draw_LastObject,&box,&trw,
                                                          &pr,&error);
  }
 }

 fxdsetview(NULL);

 return(err);
}




static os_error * samplefn(int selected,void * d,void * s,int size,
                                      int dtype,int stype,cmsdatastr * scms,cmsdatastr * dcms)
{
 int r;
 int g;
 int b;
 int w;

 if(stype==PROCESST_PAL && dtype==PROCESST_PAL)
 {
  while(size--)
  {
   w=*((int*)s);

   if(selected)
   {
    r=(w>>8)&0xFF;
    g=(w>>16)&0xFF;
    b=(w>>24)&0xFF;
    r=255-r;
    g=255-g;
    b=255-b;
    w=(b<<24)|(g<<16)|(r<<8);
   }

   *((int*)d)=w;

   s=((char*)s)+sizeof(int);
   d=((char*)d)+sizeof(int);
  }
 }

 return(NULL);
}




static os_error * fancysample(int x,int y,int w,int h,int d,
                                        fancyborderstr *fs,int selected)
{
 os_error    * err;
 draw_objptr   dummy;
 Draw_diag     diag;
 transformstr  tr1;
 transformstr  tr2;
 transformstr  tr3;
 transformstr  trs;
 transformstr  trx;
 Draw_error    error;
 processstr    pr;
 Draw_box      box;


 err=NULL;

 memset(&pr,0,sizeof(processstr));

 pr.mode=PROCESSFN;
// pr.vdupal=&cvdu.vdupal;
// pr.deltax=cvdu.deltax;
// pr.deltay=cvdu.deltay;
// pr.lndeltax=cvdu.lndeltax;
// pr.lndeltay=cvdu.lndeltay;
 pr.xdpi=cvdu.xdpi;
 pr.ydpi=cvdu.ydpi;
 pr.ln2bpp=cvdu.ln2bpp;
 pr.scale.x=pr.scale.y=0x10000;
 pr.process=samplefn;
 pr.handle=selected;


 diag.data=fancyborder+fs->offset;
 diag.length=fs->length;

 dummy.bytep=0;
 note_fontcat(dummy);


 tr_setshift(&tr1,-fs->abox[0].x1,-fs->abox[0].y1);

 tr_setidentity(&tr2);
 tr_setrotationa(&tr2,-90*0x10000);

 tr_multiply(&trs,&tr1,&tr2);


 tr_setscale(&tr2,fs->abox[0].y1-fs->abox[0].y0,w-2*d,
                  fs->abox[0].x1-fs->abox[0].x0,d);
 tr_multiply(&tr3,&trs,&tr2);

 tr_setshift(&tr1,x+w-d,y-d-(h-d)/2);
 tr_multiply(&trx,&tr3,&tr1);

 box=*(Draw_box*)(&fs->abox[0]);

 Draw_renderObjects(&diag,fs->object[0],fs->object[0],&box,&trx,&pr,&error);


  
 tr_setscale(&tr2,fs->abox[0].x1-fs->abox[0].x0,d,
                  fs->abox[0].x1-fs->abox[0].x0,d);
 tr_multiply(&tr3,&trs,&tr2);

 tr_setshift(&tr1,x+w-d,y-d-(h-d)/2);
 tr_multiply(&trx,&tr3,&tr1);


 box=*(Draw_box*)(&fs->abox[1]);

 Draw_renderObjects(&diag,fs->object[1],Draw_LastObject,&box,&trx,&pr,&error);


 tr_setshift(&trs,-fs->abox[0].x0,-fs->abox[0].y1);
 tr_setscale(&tr2,fs->abox[0].x1-fs->abox[0].x0,d,
                  fs->abox[0].x1-fs->abox[0].x0,d);

 tr_multiply(&tr3,&trs,&tr2);

 tr_setshift(&tr1,x,y-d-(h-d)/2);
 tr_multiply(&trx,&tr3,&tr1);

 Draw_renderObjects(&diag,fs->object[1],Draw_LastObject,&box,&trx,&pr,&error);


 return(err);
}






static char * borderdir="<HKEY_LOCAL_MACHINE\\Software\\DavidPilling\\OvationPro\\>\\Borders";
static char * borderbasedir="<HKEY_LOCAL_MACHINE\\Software\\DavidPilling\\OvationPro\\>\\Borders";


static os_error * loadborders(void)
{
 os_error     * err;
 buffer         bf;
 char           string[512];
 int            eof;
 int            len;
 char         * p;
 char         * q;
 int            len1;
 linestylestr   style;
 fxstat         f;
 fancyborderstr fs;
 fsscanhandle   handle;


          err=flex_alloc((flex_ptr)&linestyles,LINECHUNK);
 if(!err) err=flex_alloc((flex_ptr)&border,BORDERCHUNK);
 if(!err) err=flex_alloc((flex_ptr)&borderdefs,BORDERDEFCHUNK);
 if(!err) err=flex_alloc((flex_ptr)&fancyborder,FANCYBORDERCHUNK);

 if(!err)
 {
  err=fs_startsortscan(borderdir,TEXT("*.*"),0,&handle);
  if(!err)
  {
   while(1)
   {
    err=fs_nextsortitem(handle,&f,&eof);
 			if(err) break;
 			if(eof) break;


    sprintf(string,"%s\\%s",borderdir,f.name);

    if(f.f.type==DRAW)
    {
     len1=strlen(f.name)+1;
     if(len1<=BORDERNAMELEN)
     {
      len=(bordersize+3) & (~0x3);                            // make sure we're on a word boundary before copying
      err=flex_chunk((flex_ptr)&border,len,BORDERCHUNK);      // structures to it
      bordersize=len;                                         // DP 19/3/2007 report of change in !Borders leading to
                                                              // crash on Draw file border in scrolly sample pane
                                                              // apparently not a problem on Windows though

      len=(len1+3) & (~0x3);
      len1=sizeof(fancyborderstr);

      err=flex_chunk((flex_ptr)&border,bordersize+len+len1,BORDERCHUNK);
      if(!err) err=flex_chunk((flex_ptr)&borderdefs,
                        (borderdefn+1)*sizeof(borderdefstr),BORDERDEFCHUNK);
      if(err) break;

      borderdefs[borderdefn].name=bordersize;
      strcpy(border+bordersize,f.name);

      p=fs_extension(border+bordersize);
      if(p && (p>border+bordersize))
      {
       *(p-1)=0;
      }

      bordersize+=len;
      borderdefs[borderdefn].offset=bordersize;
      borderdefs[borderdefn].type=1;

      err=flex_chunk((flex_ptr)&fancyborder,
                                fancybordersize+f.f.length,FANCYBORDERCHUNK);
      if(err) break;

      err=fs_loadblock(string,fancyborder+fancybordersize);
      if(err) break;

      err=parsefancyborder(fancyborder+fancybordersize,f.f.length,&fs,f.name);
      if(!err)
      {
       fs.offset=fancybordersize;

       XMEMCPYX(border+bordersize,&fs,sizeof(fancyborderstr));

       fancybordersize+=f.f.length;
       bordersize+=len1;
       borderdefn++;
      }
      if(err) break;
     }
    }
    else
    {
     err=bf_open(string,'r',DEFBUFFSIZE,&bf);
     if(!err)
     {
      eof=0;

      while(!eof)
      {
       err=bf_getstring(&bf,string,sizeof(string),&eof);
       if(err) break;

       p=gettok(string,&len);
       if(len)
       {
        if(strncmp(p,"//",2))
        {
         if(!cstrcmp(p,"Line"))
         {
          p=gettok(p+len,&len);
          if(len==2)
          {
           style.id=*p;
           p=gettok(p+len,&len);
           if(len)
           {
            if(sscanf(p,"%x",&style.style)==1)
            {
             p=gettok(p+len,&len);
             if(len)
             {
              if(sscanf(p,"%x",&style.stylescale)==1)
              {
               p=gettok(p+len,&len);
               if(len)
               {
                if(sscanf(p,"%x",&style.join)==1)
                {
                 err=flex_chunk((flex_ptr)&linestyles,
                                (linestylen+1)*sizeof(linestylestr),LINECHUNK);
                 if(err) break;

                 linestyles[linestylen++]=style;
                }
               }
              }
             }
            }
           }
          }
         }
         else
         if(!cstrcmp(p,"Border"))
         {
          p=gettok(p+len,&len1);
          if(len)
          {
           q=gettok(p+len1,&len);

           err=flex_chunk((flex_ptr)&border,bordersize+len+len1+2,BORDERCHUNK);     // DP 19/3/2007, added extra 2
           if(!err) err=flex_chunk((flex_ptr)&borderdefs,
                       (borderdefn+1)*sizeof(borderdefstr),BORDERDEFCHUNK);
           if(err) break;

           borderdefs[borderdefn].name=bordersize;
           strcpy(border+bordersize,p);
           bordersize+=len1+1;
           borderdefs[borderdefn].offset=bordersize;
           borderdefs[borderdefn].type=0;
           strcpy(border+bordersize,q);
           bordersize+=len+1;
           borderdefn++;
          }
         }
        }
       }
      }
      err=bf_close(&bf,err);
     }
    }
   }
   if(!err) err=flex_extend((flex_ptr)&linestyles,
                                       linestylen*sizeof(linestylestr));

   if(!err) err=flex_extend((flex_ptr)&border,bordersize);
  
   if(!err) err=flex_extend((flex_ptr)&borderdefs,
                                       borderdefn*sizeof(borderdefstr));

   if(!err) err=flex_extend((flex_ptr)&fancyborder,fancybordersize);

   fs_endsortscan(&handle);
  }
 }
 return(err);
}



static int getborderindexs(char * name)
{
 int i;

 for(i=0;i<borderdefn;i++)
 {
  if(!cstrcmp(name,border+borderdefs[i].name)) return(i);
 }

 return(-1);
}


int getborderindex(char * name)
{
 int n;

 n=getborderindexs(name);
 if(n<0) n=0;

 return(n);
}



void getbordername(char * name,int index)
{
 if(index<0 || index>=borderdefn) index=0;
 if(!borderdefn) *name=0;
 else strcpy(name,border+borderdefs[index].name);
}



/* return width of border */

os_error * getbordersize(int border,filestr * file,int * width,curvestr * c)
{
 borderlinestr * borderp;
 int             i;
 int             bbits;
 int             sbits;


 if(border && (borderp=(borderlinestr*)fob_find(border,file))!=NULL)
 {
  bbits=(borderp->data.border[BORDER].active & BDRACT)?
                                      borderp->data.border[BORDER].mode:0;
  sbits=(borderp->data.border[SHADOW].active & BDRACT)?
                                      borderp->data.border[SHADOW].mode:0;

  if(c->type!=CBOX) 
  {
   sbits=0;
   if(bbits && bbits!=0xF) bbits=0xF;
  }

  for(i=1;i<0xF;i<<=1)
  {
   *width++=((bbits&i)?borderp->data.border[BORDER].width:0)+
            ((sbits&i)?borderp->data.border[SHADOW].width:0);
  }
 }
 else
 {
  memset(width,0,4*sizeof(int));
 }

 return(NULL);
}



void borderchange(modifyborderdatastr * borderp,borderdatastr * xnew,
                                                borderdatastr * old)
{
 int i;
 int acc;

 borderp->border=*xnew;

 borderp->xname=(cstrcmp(xnew->name,old->name)!=0);

 for(i=BORDER;i<=SHADOW;i++)
 {
  acc=(xnew->border[i].active ^ old->border[i].active);

  borderp->bits[i].xactive =(acc & BDRACT)!=0;
  borderp->bits[i].xrepeat =(acc & BDRREP)!=0;
  borderp->bits[i].xstretch=(acc & BDRSTR)!=0;
  borderp->bits[i].xdesign =(acc & BDRDES)!=0;

  borderp->bits[i].xwidth=(xnew->border[i].width!=old->border[i].width);
  borderp->bits[i].xcolour=(cmpcolourval(&xnew->border[i].colour,
                                         &old->border[i].colour)!=0);
  borderp->bits[i].xmode=(xnew->border[i].mode!=old->border[i].mode);
 }
}



os_error * createborder(modifyborderdatastr * borderp,
                                    filestr * file,int * border,int oldborder)
{
 os_error      * err;
 borderlinestr * nborder;
 int             newkey;
 int             prevkey;
 borderdatastr   data;
 int             i;

 err=NULL;

 if(oldborder) 
 {
  nborder=(borderlinestr*)fob_find(oldborder,file);
  data=nborder->data;
 }
 else
 {
  memset(&data,0,sizeof(borderdatastr));
 }

 for(i=BORDER;i<=SHADOW;i++)
 {
  if(borderp->bits[i].xactive)
  {
   data.border[i].active&=~BDRACT;
   data.border[i].active|=(borderp->border.border[i].active & BDRACT);
  }

  if(borderp->bits[i].xrepeat)
  {
   data.border[i].active&=~BDRREP;
   data.border[i].active|=(borderp->border.border[i].active & BDRREP);
  }

  if(borderp->bits[i].xdesign)
  {
   data.border[i].active&=~BDRDES;
   data.border[i].active|=(borderp->border.border[i].active & BDRDES);
  }

  if(borderp->bits[i].xstretch)
  {
   data.border[i].active&=~BDRSTR;
   data.border[i].active|=(borderp->border.border[i].active & BDRSTR);
  }


  if(borderp->bits[i].xwidth  || !oldborder)
       data.border[i].width=borderp->border.border[i].width;
  if(borderp->bits[i].xcolour || !oldborder)
       data.border[i].colour=borderp->border.border[i].colour;
  if(borderp->bits[i].xmode   || !oldborder)
       data.border[i].mode=borderp->border.border[i].mode;
 }

 if(borderp->xname || !oldborder) xzstrncpy(data.name,borderp->border.name,sizeof(data.name));

 if((data.border[BORDER].active & BDRACT) || 
    (data.border[SHADOW].active & BDRACT))
 {
  err=fob_create(sizeof(borderlinestr),file,OBORDER,(fobjptr)&nborder);
  nborder->data=data;
  *border=newkey=nborder->key;

  prevkey=file->hdr.border;

  nborder->chain.next=prevkey;
  file->hdr.border=newkey;

  if(prevkey)
  {
   nborder=(borderlinestr*)fob_find(prevkey,file);
   nborder->chain.prev=newkey;
  }
 }
 else
  *border=0;

 return(err);
}



os_error * deleteborder(int border,filestr * file)
{
 borderlinestr * nborder;
 int             next;
 int             prev;

 if(border)
 {
  nborder=(borderlinestr*)fob_find(border,file);
  next=nborder->chain.next;
  prev=nborder->chain.prev;

  if(next)
  {
   nborder=(borderlinestr*)fob_find(next,file);
   nborder->chain.prev=prev;
  }

  if(prev)
  {
   nborder=(borderlinestr*)fob_find(prev,file);
   nborder->chain.next=next;
  }
  else
  {
   file->hdr.border=next;
  }

  return(fob_delete(border,file));
 }

 return(NULL);
}


os_error * copyborderm(int * dn,filestr * dfile,
                       int sn,filestr * sfile,int copymode)
{
 os_error      * err;
 borderlinestr * borderp;
 int             newkey;
 int             prevkey;
 subborderstr    sub[2];

 err=NULL;

 if(sn)
 {
  err=fob_copy((fobjptr)&borderp,dfile,sn,sfile);
  if(!err)
  {
   *dn=newkey=borderp->key;
   prevkey=dfile->hdr.border;

   borderp->chain.next=prevkey;
   borderp->chain.prev=0;
   dfile->hdr.border=newkey;

   if(sfile!=dfile)
   {
    XMEMCPYX(sub,borderp->data.border,2*sizeof(subborderstr));
    mapcolour(sfile,dfile,&sub[0].colour,copymode);
    mapcolour(sfile,dfile,&sub[1].colour,copymode);
    fob_mod(newkey,dfile,(fobjptr)&borderp);
    XMEMCPYX(borderp->data.border,sub,2*sizeof(subborderstr));
   }

   if(prevkey)
   {
    borderp=(borderlinestr*)fob_find(prevkey,dfile);
    borderp->chain.prev=newkey;
   }
  }
 }
 else
 {
  *dn=0;
 }

 return(err);
}


os_error * copyborder(int * dn,int sn,filestr * file)
{
 return(copyborderm(dn,file,sn,file,0));
}



int bordercmp(int * width1,int * width2)
{
 return(memcmp(width1,width2,4*sizeof(int)));
}


os_error * bordergetcolour(int border,int * message,
                                             void * data,filestr * file)
{
 os_error      * err;
 borderlinestr * borderp;


 err=NULL;

 if(border)
 {
  borderp=(borderlinestr*)fob_find(border,file);

  if(*((int*)data)==SELECTION_BORDERCOLOUR)
  {
   *message=0;
   *((colourvalue*)data)=borderp->data.border[BORDER].colour;
  }
  else
  if(*((int*)data)==SELECTION_SHADOWCOLOUR)
  {
   *message=0;
   *((colourvalue*)data)=borderp->data.border[SHADOW].colour;
  }
 }

 return(err);
}








os_error * bordersetcolour(int border,void * data,filestr * file)
{
 os_error      * err;
 borderlinestr * borderp;


 err=NULL;

 if(border)
 {
  borderp=(borderlinestr*)fob_find(border,file);

  if(*((int*)data)==SELECTION_BORDERCOLOUR)
  {
   borderp->data.border[BORDER].colour=*((colourvalue*)(((int*)data)+1));
  }
  else
  if(*((int*)data)==SELECTION_SHADOWCOLOUR)
  {
   borderp->data.border[SHADOW].colour=*((colourvalue*)(((int*)data)+1));
  }
 }

 return(err);
}







/****************************************************************************/


static void borderbox(int width,int drop,int bits,coordstr * cx,int * path)
{
 int * p;
 int   dx;
 int   dy;
 int   mask;
 int   pmask;
 int   nmask;
 int   i;
 int   j;


 p=path;

 mask=0x1;
 pmask=0x8;

 for(i=0;i<4;i++)
  if(!(bits & (1<<i))) break;

 if(i<4)
 {
  j=i;

  while(1)
  {
   j=(j+1) & 0x3;
   if(j==i) break;

   mask=(1<<j);
   pmask=(1<<((j+3) & 0x3));
   nmask=(1<<((j+1) & 0x3));

   if(bits & mask)
   {
    dx=dy=0;

    if(j==0) dx=width/2;
    else
    if(j==1) dy=width/2;
    else
    if(j==2) dx=-width/2;
    else
    if(j==3) dy=-width/2;

    if(drop)
    {
     dx*=4;
     dy*=4;
    }


    if(!(bits & pmask))
    {
     *p++=DRAWMOVE;
     *p++=cx[j].x+dx;
     *p++=cx[j].y+dy;
    }
    else
    {
     *p++=DRAWDRAW;
     *p++=cx[j].x;
     *p++=cx[j].y;
    }

    if(!(bits & nmask))
    {
     *p++=DRAWDRAW;
     *p++=cx[(j+1) & 0x3].x-dx;
     *p++=cx[(j+1) & 0x3].y-dy;
    }
   }
  }
 }

 *p++=DRAWEND;
}




static void borderpath(borderdatastr * data,int shadow,
                                  curvestr * curve,int * path,int t)
{
 coordstr cx[4];
 int      bits;
 int      width;

 curve->data.rbox.xradius+=t;
 curve->data.rbox.yradius+=t;
 curve->box.w+=2*t;
 curve->box.h+=2*t;
 curve->box.x-=t;
 curve->box.y-=t;

 width=data->border[shadow].width;

 if(curve->type==CBOX)
 {
  if(!shadow && (data->border[SHADOW].active & BDRACT)) bits=0xF;
  else                                       bits=data->border[shadow].mode;

  if(bits==0xF)     curvepath(curve,path);
  else
  {
   box2coord(cx,&curve->box);
   borderbox(width,shadow,bits,cx,path);
  }
 }
 else
 {
  curvepath(curve,path);
 }
}



os_error * renderborder(int bordern,curvestr * fcurve,transformstr * tr,
                                                          viewredrawstr * v)
{
 os_error      * err;
 borderdatastr   data;
 int             path[64];
 transformbitstr b;
 int             bits;
 curvestr        curve;
 curvestr        xcurve;
 int             i;
 int             n;
 int             len;
 char          * p;
 int             c;
 pathstyle       pstyle;
 linestylestr  * style;
 filestr       * file;
 viewstr       * view;
 int             transparent;
 int             t;
 int             t1;
 int             t2;
 transformstr    trx;
 int           * spath;
 int           * dpath;
 int             w;
 int             dw;
 int             type;


 err=NULL;


 if(bordern)
 {
  file=v->file;
  view=v->view;

  data=((borderlinestr*)fob_find(bordern,file))->data;

  n=getborderindex(data.name);

  p=border+borderdefs[n].offset;
  type=borderdefs[n].type;

  w=data.border[BORDER].width;

  if(type==0)
  {
   len=strlen(p);
   p+=len-1;
   i=len;
   dw=(data.border[BORDER].width/=len);
  }
  else
  {
   dw=0;
   i=0; /* compiler */
  }


 
  t2=0;

  if(fcurve->type==CPATH && (data.border[BORDER].active & BDRACT) && type==0)
  {
   err=pathflex(&spath,fcurve->data.path.path,file);
   if(!err)
   {    
    pathtransform(spath,tr);

    while(i)
    {
     c=*p;
     t=0;

     while(i && (*p==c)) {t+=data.border[BORDER].width;p--;i--;}

     if(c!='_')
     {
      t1=t/2+t2;

      dpath=NULL;

      err=parallelpath2(file,0,&spath,0,&dpath,t1);
      if(!err && dpath)
      {
       style=findlinestyle(c);

       rcreatepathstyle(&pstyle,0);
 
       pstyle.thickness=t;
       pstyle.stylescale=style->stylescale;
       pstyle.style=style->style;
       pstyle.join=style->join;

//       err=vsetcolour(&data.border[BORDER].colour,&transparent,view);
       err=render_setcolour(&data.border[BORDER].colour,&transparent,v);
       if(!err)
       {
        if(!transparent) err=rpathline(dpath,&pstyle,v,NULL);
       }

       flex_free((flex_ptr)&dpath);
      }
     }
     t2+=t;
    }

    flex_free((flex_ptr)&spath);
   }
  }
  else
  {
   bits=tr_getbits(tr,&b);
   bits&=~(TR_XSCALE|TR_YSCALE);
   tr_setbits(&trx,&b,bits);

   curve.type=fcurve->type;
   curve.box.x=scale(fcurve->box.x,b.xscale,0x10000);
   curve.box.y=scale(fcurve->box.y,b.yscale,0x10000);
   curve.box.w=scale(fcurve->box.w,b.xscale,0x10000);
   curve.box.h=scale(fcurve->box.h,b.yscale,0x10000);
   curve.data.rbox.xradius=scale(fcurve->data.rbox.xradius,b.xscale,0x10000);
   curve.data.rbox.yradius=scale(fcurve->data.rbox.yradius,b.yscale,0x10000);

   if(fcurve->type==CBOX && (data.border[SHADOW].active & BDRACT))
                                                    /* do drop shadow */
   {
    xcurve=curve;

    if(data.border[BORDER].active & BDRACT)
    {
     t=data.border[BORDER].width;

     xcurve.box.x-=w-dw/2;
     xcurve.box.y-=w-dw/2;
     xcurve.box.w+=2*w-dw;
     xcurve.box.h+=2*w-dw;
    }

    data.border[SHADOW].width+=dw/2;
    data.border[SHADOW].width/=2;

    borderpath(&data,1,&xcurve,path,data.border[SHADOW].width);

    rcreatepathstyle(&pstyle,0);
    pstyle.thickness=2*data.border[SHADOW].width;
    pstyle.join=0;

//    err=vsetcolour(&data.border[SHADOW].colour,&transparent,view);
    err=render_setcolour(&data.border[SHADOW].colour,&transparent,v);
    if(!err)
    {
     if(!transparent)
     {
      err=rpathline(path,&pstyle,v,&trx);
     }
    }
   }

   if(data.border[BORDER].active & BDRACT)
   {
    if(type==0)
    {
     while(i)
     {
      c=*p;
      t=0;
      while(i && (*p==c)) {t+=data.border[BORDER].width;p--;i--;}

      if(c!='_')
      {
       t1=t/2+t2;

       style=findlinestyle(c);

       borderpath(&data,0,&curve,path,t1);

       rcreatepathstyle(&pstyle,0);

       pstyle.thickness=t;
       pstyle.stylescale=style->stylescale;
       pstyle.style=style->style;
       pstyle.join=style->join;

//       err=vsetcolour(&data.border[BORDER].colour,&transparent,view);
       err=render_setcolour(&data.border[BORDER].colour,&transparent,v);
       if(!err)
       {
        if(!transparent)
        {
         err=rpathline(path,&pstyle,v,&trx);
        }
       }
       t2-=t1;
      }
      t2+=t;
     }
    }
    else
    if(fcurve->type==CBOX)
    {        /* render fancy border */

     err=renderfancyborder((fancyborderstr*)(border+borderdefs[n].offset),
                                                        &data,&curve,&trx,v);
    }
   }
  }
 }

 return(err);
}





os_error * getframemaxbox(sobjectstr ** sobp,int frame,filestr * file,
                                                      transformstr * tr)
{
 os_error       * err;
 borderdatastr    data;
 sobjectstr     * sob;
 int              key;
 transformbitstr  b;
 int              bits;
 transformstr     trx;
 curvestr         curve;
 int              width[4];
 int            * spath;
 int            * dpath;
 int              noborder;
 int              t;
 int              t1;
 int              c;
 char           * p;
 int              len;
 coordstr         cx[2];
 pathstyle        pstyle;
 linestylestr   * style;
 int              n;


 sob=*sobp;
 key=sob->key;

 if(frame)
 {
  if(sob->fcurve.type==CPATH)
  {
   noborder=1;

   data=((borderlinestr*)fob_find(frame,file))->data;

   if(data.border[BORDER].active & BDRACT)
   {
    n=getborderindex(data.name);
    p=border+borderdefs[n].offset;
    if(borderdefs[n].type==0)
    {
     len=strlen(p);
     t1=data.border[BORDER].width;
     data.border[BORDER].width/=len;

     err=pathflex(&spath,sob->fcurve.data.path.path,file);
     if(!err)
     {    
      pathtransform(spath,tr);

      c=*p;
      t=0;

      while(*p && (*p==c)) 
      {
       t+=data.border[BORDER].width;
       p++;
      }

      if(c!='_')
      {
       t1-=t/2;

       dpath=NULL;

       err=parallelpath2(file,0,&spath,0,&dpath,t1);
       if(!err && dpath)
       {
        style=findlinestyle(c);

        rcreatepathstyle(&pstyle,0);
 
        pstyle.thickness=t;
        pstyle.stylescale=style->stylescale;
        pstyle.style=style->style;
        pstyle.join=style->join;
     
        drawboxsize(cx,dpath,&pstyle);
        noborder=0;

        sob->box.x=cx[0].x;
        sob->box.y=cx[0].y;
        sob->box.w=cx[1].x-cx[0].x;
        sob->box.h=cx[1].y-cx[0].y;

        flex_free((flex_ptr)&dpath);
       }
      }

      flex_free((flex_ptr)&spath);
     }
    }
   }
   if(noborder) getmaxbox(&sob->box,&sob->fcurve.box,tr);
  }
  else
  {
   bits=tr_getbits(tr,&b);
   bits&=~(TR_XSCALE|TR_YSCALE);
   tr_setbits(&trx,&b,bits);

   curve.type=sob->fcurve.type;
   curve.box.x=scale(sob->fcurve.box.x,b.xscale,0x10000);
   curve.box.y=scale(sob->fcurve.box.y,b.yscale,0x10000);
   curve.box.w=scale(sob->fcurve.box.w,b.xscale,0x10000);
   curve.box.h=scale(sob->fcurve.box.h,b.yscale,0x10000);

   getbordersize(frame,file,width,&curve);

   curve.box.w+=width[1]+width[3];
   curve.box.h+=width[0]+width[2];
   curve.box.x-=width[3];
   curve.box.y-=width[0];

   getmaxbox(&sob->box,&curve.box,&trx);
  }
 }
 else
 {
  getmaxbox(&sob->box,&sob->fcurve.box,tr);
 }

 return(NULL);
}



/****************************************************************************/


/* tagged to UBORDERS */
/* this function is always called last in undo/redo */

static os_error * unborders(filestr * file)
{

 frflflush(file);
 vxflush(file,VRBACK,selpagek,selsectk);

 frflvalid(file);

 frflscrap(file);
 vxscrap();


 return(undowritefn(file,UBORDERF,NULL,0));
}


/* tagged to UBORDERF */

static os_error * unborderf(filestr * file)
{

 frflstart(file);
 vxstart(file);

 return(undowritefn(file,UBORDERS,NULL,0));
}




static os_error * borderselectionsub(modifyborderdatastr * border,
                                                               filestr * file)
{
 os_error * err;
 int        message;

 err=NULL;

 if(selectn)
 {
  err=modfile(selfile,selsectk);
  if(!err)
  {
   unborderf(file);

   message=OM_NEWBORDER;
   selectionmessage(file,&message,border);

   unborders(file);
  }
 }

 return(err);
}




os_error * borderselection(modifyborderdatastr * borderp,filestr * file)
{
 os_error * err;

 err=NULL;

 if(file==selfile)
 {
  err=undostart(selfile);

  err=borderselectionsub(borderp,file);

  undoend(selfile);
 }

 return(err);
}


/**************************************************************************/


static os_error * dsborder(int fp,int argc,objectloc * loc)
{
 int           sp;
 int           nsp;
 dsobjectstr * ds;
 dsobjectstr * ds1;
 int           argn;
 subborderstr  border;

 sp=fp+2; /* skip leading type/len */

 memset(&border,0,sizeof(subborderstr));

 argn=0;

 while(argn++<argc)
 {
  ds=(dsobjectstr*)(dsstack+sp);
  nsp=sp+ds->len/sizeof(int);
  ds1=(dsobjectstr*)(dsstack+sp+2);

  switch(ds->type)
  {
   case  DSTINT:
                if(argn==1) border.active=ds->data.integer;
                else        border.mode=ds->data.integer;
                break;

      case DSTW:
                border.width=ds1->data.integer;
                break;

   case DSTCOLOURVALUE:
                       border.colour.data.tint.key=ds1->data.integer;
                       ds1=(dsobjectstr*)(dsstack+sp+2+3);
                       border.colour.data.tint.tint=ds1->data.integer;
                       if(nsp>(sp+2+6))
                       {
                        ds1=(dsobjectstr*)(dsstack+sp+2+6);
                        border.colour.data.tint.overprint=ds1->data.integer;
                       }
                       break;
  }

  sp=nsp;
 }

 ds=(dsobjectstr*)(dsstack+fp);
 XMEMCPYX(ds->data.array,&border,sizeof(subborderstr));
 ds->len=2*sizeof(int)+sizeof(subborderstr);

 return(NULL);
 USE(loc);
}



static os_error * dsframeborder(int fp,int argc,objectloc * loc)
{
 os_error       *    err;
 int                 sp;
 int                 nsp;
 dsobjectstr       * ds;
 int                 key;
 modifyborderdatastr data;

 sp=fp+2; /* skip leading type/len */

 memset(&data,0,sizeof(modifyborderdatastr));
 data.bits[BORDER].xactive=data.bits[SHADOW].xactive=1;

 while(argc--)
 {
  ds=(dsobjectstr*)(dsstack+sp);
  nsp=sp+ds->len/sizeof(int);

  switch(ds->type)
  {
   case    DSTSTR:
                  xzstrncpy(data.border.name,ds->data.string,sizeof(data.border.name));
                  break;

   case DSTBORDER:
                  XMEMCPYX(&data.border.border[BORDER],ds->data.array,
                                              sizeof(subborderstr));
                  break;

   case DSTSHADOW:
                  XMEMCPYX(&data.border.border[SHADOW],ds->data.array,
                                              sizeof(subborderstr));
                  break;
  }

  sp=nsp;
 }

 err=createborder(&data,loc->file,&key,0);

 ds=(dsobjectstr*)(dsstack+fp);
 ds->data.integer=key;
 ds->len=3*sizeof(int);

// cprintf("border created %d",key);

 return(err);
}




static int dsnulllist[]={
                         DSTENDLIST
                        };



static dscreator dsdefs[]=
{
 0,0,0,0, "frameborder",  dsframeborder, 0, DSTFRAMEBORDER, dsnulllist,
 0,0,0,0, "border",       dsborder,      0, DSTBORDER,      dsnulllist,
 0,0,0,0, "shadow",       dsborder,      0, DSTSHADOW,      dsnulllist,
};





static os_error * dswbsub(char * name,subborderstr * border,
                                             buffer * bf,filestr * file)
{
 return(bf_printf(bf,"{%s %d 0x%x %s {w %d}}",
                  name,
                  border->active,
                  border->mode,
                  dscvstring(&border->colour,file),
                  border->width));
}



os_error * dswriteframeborder(int bordern,buffer * bf,filestr * file)
{
 os_error      * err;
 borderlinestr   border;

 err=NULL;

 if(bordern)
 {
  border=*((borderlinestr*)fob_find(bordern,file));

  err=bf_printf(bf,"{frameborder \"%s\" ",border.data.name);
  if(!err) err=dswbsub("border",&border.data.border[BORDER],bf,file);
  if(!err) err=dswbsub("shadow",&border.data.border[SHADOW],bf,file);
  if(!err) err=bf_printfeol(bf,"}");
 }

 return(err);
}


/****************************************************************************/


static os_error * borderopen(int eventn,int user,filestr * file,int data1)
{
 borderlinestr * borderp;
 int             next;

 next=file->hdr.border;

 while(next)
 {
  borderp=(borderlinestr*)fob_find(next,file);
  next=borderp->chain.next;

  if(getborderindexs(borderp->data.name)<0)
  {
   werr(WERRMSG,"{BDR0}",borderp->data.name);
   break;
  }
 }

 return(NULL);
 USE(eventn);
 USE(user);
 USE(data1);
}








/***************************************************************************/

#define UNDOS 2

static undofnstr undos[UNDOS]=
{
 unborders,              NULL, UBORDERS,      "{UBORDER}",
 unborderf,              NULL, UBORDERF,      "{UBORDER}",
};



os_error * borderinit(void)
{
 os_error * err;

          err=loadborders();
 if(!err) err=undoaddblock(undos,UNDOS);
 if(!err) err=addevent(EVENT_FOPEN,(eventfn)borderopen,0);

 dsaddblock(dsdefs,sizeof(dsdefs)/sizeof(dscreator));

 return(err);
}

