/*->c.curve */

#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>


#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "bbc.h"
#include "flex.h"
#include "transform.h"
#include "BezierArc.h"
#include "deb.h"

#include "err.h"
#include "fsx.h"
#include "bf.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "colour.h"
#include "etc.h"
#include "xmath.h"


#include "constants.h"
#include "str.h"
#include "object.h"

#include "viewr.h"
#include "view.h"
#include "file.h"
#include "chain.h"
#include "xcolour.h"
#include "xredraw.h"
#include "geometry.h"


#include "posn.h"



#include "curvelo.h"
#include "curve.h"



/*****************************************************************************/


void newsegment(segstr * seg)
{
 seg->index=seg->oldindex=seg->bindex=0;
 seg->type=0;
}


/* return segment type 0==end 1==line 2==curve 3=close loop */

os_error * nextsegment(segstr * seg,int * q)
{
 int      x;
 int    * p;

 seg->oldindex=seg->index;

 p=q+seg->index;

 if(seg->type==1) seg->cx[0]=seg->cx[1];
 else
 if(seg->type==2) seg->cx[0]=seg->cx[3];

 while(1)
 {
  x=*p;

  if(x==DRAWEND)
  {
   if((seg->cx[0].x!=seg->bx.x || seg->cx[0].y!=seg->bx.y) && seg->index)
   {
    seg->cx[1]=seg->bx;
    seg->type=1;
    break;
   }

   if(seg->type!=3 && seg->index)
   {
    seg->type=3;
    break;
   }

   seg->type=0;
   break;
  }
  else
  if(x==DRAWCLOSE)
  {
   if(seg->cx[0].x!=seg->bx.x || seg->cx[0].y!=seg->bx.y)
   {
    seg->cx[1]=seg->bx;
    seg->type=1;
    break;
   }

   p++;
   seg->type=3;
   break;
  }
  else
  if(x==DRAWDRAW)
  {
   seg->cx[1].x=*(p+1);
   seg->cx[1].y=*(p+2);
   p+=3;
   seg->type=1;
   break;
  }
  else
  if(x==DRAWCURVE)
  {
   seg->cx[1].x=*(p+1);
   seg->cx[1].y=*(p+2);
   seg->cx[2].x=*(p+3);
   seg->cx[2].y=*(p+4);
   seg->cx[3].x=*(p+5);
   seg->cx[3].y=*(p+6);
   p+=7;
   seg->type=2;
   break;
  }
  else
  if(x==DRAWMOVE)
  {
   seg->bindex=p-q;
   seg->cx[0].x=seg->bx.x=*(p+1);
   seg->cx[0].y=seg->bx.y=*(p+2);
   p+=3;
  }
 }

 seg->index=p-q;

 return(NULL);
}


os_error * nextsegmentpath(segstr * seg,int path,filestr * file)
{
 pathstr * pathp;
 pathp=(pathstr*)fob_find(path,file);
 return(nextsegment(seg,pathp->data));
}



static void seggrad(segstr * seg,int second,int * dx,int * dy)
{
 if(seg->type==1)
 {
  *dx=seg->cx[1].x-seg->cx[0].x;
  *dy=seg->cx[1].y-seg->cx[0].y;
 }
 else
 if(seg->type==2)
 {
  if(second)
  {
   *dx=seg->cx[3].x-seg->cx[2].x;
   *dy=seg->cx[3].y-seg->cx[2].y;
  }
  else
  {
   *dx=seg->cx[1].x-seg->cx[0].x;
   *dy=seg->cx[1].y-seg->cx[0].y;
  }
 }
}



/****************************************************************************/

static int insiderbox(int x,int y,curvestr * c)
{
 int    xradius;
 int    yradius;
 boxstr box;
 int    cx;
 int    cy;

 xradius=c->data.rbox.xradius;
 if(xradius>c->box.w/2) xradius=c->box.w/2;

 yradius=c->data.rbox.yradius;
 if(yradius>c->box.h/2) yradius=c->box.h/2;

 box=c->box;
 box.h-=2*yradius;
 box.y+=yradius;

 if(insidebox(x,y,&box)) return(1);

 box=c->box;
 box.w-=2*xradius;
 box.x+=xradius;

 if(insidebox(x,y,&box)) return(1);

 box=c->box;

 cx=box.x+xradius;
 cy=box.y+yradius;
 if(pythag(x-cx,scale(y-cy,xradius,yradius))<xradius) return(1);

 cx=box.x+box.w-xradius;
 if(pythag(x-cx,scale(y-cy,xradius,yradius))<xradius) return(1);

 cy=box.y+box.h-yradius;
 if(pythag(x-cx,scale(y-cy,xradius,yradius))<xradius) return(1);

 cx=box.x+xradius;
 if(pythag(x-cx,scale(y-cy,xradius,yradius))<xradius) return(1);

 return(0);
}


static int insideellipse(int x,int y,curvestr * c)
{
 int cx;
 int cy;
 int ht;
 int hb;
 int wl;
 int wr;
 int h;
 int w;

 ht=c->box.h/2;
 hb=c->box.h-ht;

 wl=c->box.w/2;
 wr=c->box.w-wl;

 cx=c->box.x+wl;
 cy=c->box.y+ht;

 x-=cx;
 y-=cy;

 if(x<=0) w=wl;
 else     w=wr;

 if(y>=0) h=ht;
 else     h=hb;

 if(w>h)
 {
  x=scale(x,h,w);
  return(pythag(x,y)<h);
 }
 else
 {
  y=scale(y,w,h);
  return(pythag(x,y)<w);
 }
}


/* generate a line from (x,y) to outside the curve */
/* count the intersections => inside?              */

static int insidepath(int x,int y,curvestr * c,filestr * file)
{
 pathstr * p;
 int     * q;
 int       cuts;
 int       type;
 segstr    seg;


 cuts=0;


 p=(pathstr*)fob_find(c->data.path.path,file);
 if(p)
 {
  q=p->data;

  newsegment(&seg);


  while(1)
  {
   nextsegment(&seg,q);
   if((type=seg.type)==0) break;

   if(type==1) 
   {
    if(hintersection(x,y,seg.cx)==INTER) cuts++;
   }
   else
   if(type==2)
   {
    cuts+=hbintersection(x,y,seg.cx);
   }
  }
 }

 return(cuts & 0x1);
}


int insidecurve(int x,int y,curvestr * c,transformstr * tr,filestr * file)
{
 coordstr coord;

 coord.x=x;
 coord.y=y;

 if(tr) tr_inverseset(&coord,1,tr);

 if(insidebox(coord.x,coord.y,&c->box))
 {
  switch(c->type)
  {
      case CBOX:
                return(1);
                break;

     case CBOXR:
                return(insiderbox(coord.x,coord.y,c));
                break;

  case CELLIPSE:
                return(insideellipse(coord.x,coord.y,c));
                break;

     case CPATH:
                return(insidepath(coord.x,coord.y,c,file));
                break;
  }
 }
 return(0);
}


/***************************************************************************/

/* project out to left and right, find nearest cut */
/* project out to top  and bot,   find nearest cut */

static int nearpathp(int * p,int x,int y,int dx,int dy,curvestr * c)
{
 int     * q;
 int       type;
 segstr    seg;
 coordstr  ray[2];
 coordstr  result;
 cutstr    cut[MAXBCUT];
 int       i;
 int       n;
 int       w;
 int       h;
 int       ds;

 ds=MAX(dx,dy);

 q=p;
 newsegment(&seg);

 w=ABS(c->box.w);
 if(w<ds) w=ds;

 ray[0].x=c->box.x-w;
 ray[0].y=y;
 ray[1].x=ray[0].x+2*w;
 ray[1].y=ray[0].y;


 while(1)
 {
  nextsegment(&seg,q);
  if((type=seg.type)==0) break;

  if(type==1)
  {
   if(hintersectionr(ray,seg.cx,&result))
   {
    if(ABS(result.x-x)<=dx) return(1);
   }
  }
  else
  if(type==2)
  {
   n=hbintersectionr(ray,seg.cx,cut);
   for(i=0;i<n;i++)
   {
    if(ABS(cut[i].c.x-x)<=dx) return(1);
   }
  }
 }


 q=p;
 newsegment(&seg);

 h=ABS(c->box.h);
 if(h<ds) h=ds;

 ray[0].x=x;
 ray[0].y=c->box.y-h;
 ray[1].x=x;
 ray[1].y=ray[0].y+2*h;

 while(1)
 {
  nextsegment(&seg,q);
  if((type=seg.type)==0) break;

  if(type==1)
  {
   if(vintersectionr(ray,seg.cx,&result))
   {
    if(ABS(result.y-y)<=dy) return(1);
   }
  }
  else
  if(type==2)
  {
   n=vbintersectionr(ray,seg.cx,cut);
   for(i=0;i<n;i++)
   {
    if(ABS(cut[i].c.y-y)<=dy) return(1);
   }
  }
 }

 return(0);
}



static int nearpath(int x,int y,int dx,int dy,curvestr * c,filestr * file)
{
 pathstr * p;
 int     * q;

 p=(pathstr*)fob_find(c->data.path.path,file);
 q=p->data;

 return(nearpathp(q,x,y,dx,dy,c));
}


static int nearline(coordstr * cx,int dx,int dy,curvestr * c)
{
 int        path[16];

 curvepath(c,path);

 return(nearpathp(path,cx->x,cx->y,dx,dy,c));
}



int nearcurve(int x,int y,int dx,int dy,curvestr * c,
                                        transformstr * tr,filestr * file)
{
 coordstr      coord;
 transformstr  inv;
 int           ds;

 coord.x=x;
 coord.y=y;


 if(tr)
 {
  tr_inverse(tr,&inv);
  tr_transformset(&coord,1,&inv);
  tr_delta(&inv,&dx,&dy);
  dx=ABS(dx);
  dy=ABS(dy);
  ds=MAX(dx,dy);
  dx=dy=ds;
 }


 if(nearlyinsidebox(coord.x,coord.y,dx,dy,&c->box))
 {

  switch(c->type)
  {
      case CBOX:
                return(1);
                break;

     case CBOXR:
                return(insiderbox(coord.x,coord.y,c));
                break;

  case CELLIPSE:
                return(insideellipse(coord.x,coord.y,c));
                break;

     case CPATH:
                return(nearpath(coord.x,coord.y,dx,dy,c,file));
                break;

     case CLINE:
    case CLINEX:
    case CLINER:
   case CLINEXR:
                return(nearline(&coord,dx,dy,c));
                break;

  }
 }
 return(0);
}


/****************************************************************************/
#define CPCHUNK 16*sizeof(int)


/* round box */

static os_error * cphrbox(curvestr * f,boxstr * box,filestr * file,
                                                              clipbuff * buff)
{                                                                                os_error * err;
 int      * p;
 boxstr     temp;
 boxstr     result;
 int        x;
 int        y;
 int        h;
 int        w;
 int        xradius;
 int        yradius;

 err=NULL;
 p=buff->path;

 box->h=0;

 if(getminbox2(&result,&f->box,box))
 {
  xradius=f->data.rbox.xradius;
  if(xradius>f->box.w/2) xradius=f->box.w/2;

  yradius=f->data.rbox.yradius;
  if(yradius>f->box.h/2) yradius=f->box.h/2;

  temp=f->box;
  temp.h-=2*yradius;
  temp.y+=yradius;

  if(getminbox2(&result,box,&temp))
  {
   x=result.x;
   y=result.y;
   w=result.w;
  }
  else
  {
   y=box->y;

   if(box->y<=temp.y)          h=temp.y-box->y;
   else
   if(box->y>=(temp.y+temp.h)) h=box->y-(temp.y+temp.h);
   else                        h=0;

   w=xradius-pythagd(xradius,scale(h,xradius,yradius));
   x=box->x+w;
   w=box->w-2*w;
  }

  *p++=DRAWMOVE;
  *p++=x;
  *p++=y;
  *p++=DRAWDRAW;
  *p++=x+w;
  *p++=y;
 }

 *p++=DRAWEND;

 return(err);
 USE(file);
}



/* box */

static os_error * cphbox(curvestr * f,boxstr * box,filestr * file,
                                                             clipbuff * buff)
{
 os_error * err;
 int      * p;
 boxstr     result;

 err=NULL;
 p=buff->path;

 box->h=0;

 if(getminbox2(&result,box,&f->box))
 {
  *p++=DRAWMOVE;
  *p++=result.x;
  *p++=result.y;
  *p++=DRAWDRAW;
  *p++=result.x+result.w;
  *p++=result.y;
 }

 *p++=DRAWEND;

 return(err);
 USE(file);
}


static os_error * cphellipse(curvestr * f,boxstr * box,filestr * file,
                                                              clipbuff * buff)
{
 os_error * err;
 int      * p;
 int        cx;
 int        cy;
 int        h;
 int        w;
 int        x;
 int        y;

 err=NULL;

 p=buff->path;

 h=f->box.h/2;
 w=f->box.w/2;

 cx=f->box.x+w;
 cy=f->box.y+h;

 y=box->y-cy;

 if(y<=h)
 {
  y=scale(y,w,h);

  x=pythagd(w,y);

  *p++=DRAWMOVE;
  *p++=cx-x;
  *p++=box->y;
  *p++=DRAWDRAW;
  *p++=cx+x;
  *p++=box->y;
 }

 *p++=DRAWEND;

 return(err);
 USE(file);
}



static int vpcmp(const void * p1,const void * p2)
{
 return(*(((int*)p1)+2)-*(((int*)p2)+2));
}


static os_error * cphpath(curvestr * f,boxstr * box,filestr * file,
                                                              clipbuff * buff)
{
 os_error * err;
 int      * p;
 pathstr  * path;
 int      * q;
 coordstr   ray[2];
 cutstr     result[MAXBCUT];
 int        cuts;
 int        i;
 int        n;
 int        type;
 segstr     seg;
 coordstr   hres;

 err=NULL;

 p=buff->path;

 path=(pathstr*)fob_find(f->data.path.path,file);
 q=path->data;
 newsegment(&seg);

 ray[0].x=box->x-box->w;
 ray[0].y=box->y;
 ray[1].x=ray[0].x+2*box->w;
 ray[1].y=ray[0].y;

 n=0;
 while(1)
 {
  nextsegment(&seg,q);
  if((type=seg.type)==0) break;

  if(type==1) 
  {
   if(hintersectionr(ray,seg.cx,&hres))
   {
    *p++=DRAWMOVE;
    *p++=hres.x;
    *p++=hres.y;
    n++;
   }
  }
  else
  if(type==2)
  {
   cuts=bintersection(ray,seg.cx,result);

   for(i=0;i<cuts;i++)
   {
    *p++=DRAWMOVE;
    *p++=result[i].c.x;
    *p++=result[i].c.y;
    n++;
   }
  }
 }

 *p++=DRAWEND;

 p=buff->path;

 qsort(p,n,3*sizeof(int),vpcmp);

 for(i=0;i<n;i++)
 {
  if(i&0x1) *p=DRAWDRAW;
  p+=3;
 }

 return(err);
}




/* input  a curve, and a horizontal line from box x,y, x+w */
/* output a series of Draw paths to draw the clipped line  */ 

os_error * cliphline(curvestr * f,boxstr * box,filestr * file,clipbuff * buff)
{
 os_error * err;

 err=NULL;

 switch(f->type)
 {
     case CBOXR:
                cphrbox(f,box,file,buff);
                break;

      case CBOX:
                cphbox(f,box,file,buff);
                break;

  case CELLIPSE:
                cphellipse(f,box,file,buff);
                break;

     case CPATH:
                cphpath(f,box,file,buff);
                break;
 }

 return(err);
}





/* round box */

static os_error * cpvrbox(curvestr * f,boxstr * box,filestr * file,
                                                              clipbuff * buff)
{                                                                                os_error * err;
 int      * p;
 boxstr     temp;
 boxstr     result;
 int        x;
 int        y;
 int        h;
 int        w;
 int        xradius;
 int        yradius;

 err=NULL;
 p=buff->path;

 box->w=0;

 if(getminbox2(&result,&f->box,box))
 {
  xradius=f->data.rbox.xradius;
  if(xradius>f->box.w/2) xradius=f->box.w/2;

  yradius=f->data.rbox.yradius;
  if(yradius>f->box.h/2) yradius=f->box.h/2;

  temp=f->box;
  temp.w-=2*xradius;
  temp.x+=xradius;

  if(getminbox2(&result,box,&temp))
  {
   x=result.x;
   y=result.y;
   h=result.h;
  }
  else
  {
   x=box->x;

   if(box->x<=temp.x)          w=temp.x-box->x;
   else
   if(box->x>=(temp.x+temp.w)) w=box->x-(temp.x+temp.w);
   else                        w=0;

   h=yradius-pythagd(yradius,scale(w,yradius,xradius));
   y=box->y+h;
   h=box->h-2*h;
  }

  *p++=DRAWMOVE;
  *p++=x;
  *p++=y;
  *p++=DRAWDRAW;
  *p++=x;
  *p++=y+h;
 }

 *p++=DRAWEND;

 return(err);
 USE(file);
}


/* box */

static os_error * cpvbox(curvestr * f,boxstr * box,filestr * file,
                                                             clipbuff * buff)
{
 os_error * err;
 int      * p;
 boxstr     result;

 err=NULL;
 p=buff->path;

 box->w=0;

 if(getminbox2(&result,box,&f->box))
 {
  *p++=DRAWMOVE;
  *p++=result.x;
  *p++=result.y;
  *p++=DRAWDRAW;
  *p++=result.x;
  *p++=result.y+result.h;
 }

 *p++=DRAWEND;

 return(err);
 USE(file);
}


static os_error * cpvellipse(curvestr * f,boxstr * box,filestr * file,
                                                              clipbuff * buff)
{
 os_error * err;
 int      * p;
 int        cx;
 int        cy;
 int        h;
 int        w;
 int        x;
 int        y;

 err=NULL;

 p=buff->path;

 h=f->box.h/2;
 w=f->box.w/2;

 cx=f->box.x+w;
 cy=f->box.y+h;

 x=box->x-cx;

 if(x<=w)
 {
  x=scale(x,h,w);

  y=pythagd(h,x);

  *p++=DRAWMOVE;
  *p++=box->x;
  *p++=cy-y;
  *p++=DRAWDRAW;
  *p++=box->x;
  *p++=cy+y;
 }

 *p++=DRAWEND;

 return(err);
 USE(file);
}




static os_error * cpvpath(curvestr * f,boxstr * box,filestr * file,
                                                              clipbuff * buff)
{
 os_error * err;
 int      * p;
 pathstr  * path;
 int      * q;
 coordstr   ray[2];
 cutstr     result[MAXBCUT];
 int        cuts;
 int        i;
 int        n;
 int        type;
 segstr     seg;
 coordstr   vres;

 err=NULL;

 p=buff->path;

 path=(pathstr*)fob_find(f->data.path.path,file);
 q=path->data;
 newsegment(&seg);

 ray[0].x=box->x;
 ray[0].y=box->y-box->h;
 ray[1].x=ray[0].x;
 ray[1].y=ray[0].y+2*box->h;

 n=0;
 while(1)
 {
  nextsegment(&seg,q);
  if((type=seg.type)==0) break;

  if(type==1)
  {
   if(vintersectionr(ray,seg.cx,&vres))
   {
    *p++=DRAWMOVE;
    *p++=vres.x;
    *p++=vres.y;
    n++;
   } 
  }
  else
  if(type==2)
  {
   cuts=bintersection(ray,seg.cx,result);

   for(i=0;i<cuts;i++)
   {
    *p++=DRAWMOVE;
    *p++=result[i].c.x;
    *p++=result[i].c.y;
    n++;
   }
  }
 }

 *p++=DRAWEND;

 p=buff->path;

 qsort(p,n,3*sizeof(int),vpcmp);

 for(i=0;i<n;i++)
 {
  if(i&0x1) *p=DRAWDRAW;
  p+=3;
 }

 return(err);
}


/* input  a curve, and a vertical line from box x,y, y+h  */
/* output a series of Draw paths to draw the clipped line */ 

os_error * clipvline(curvestr * f,boxstr * box,filestr * file,clipbuff * buff)
{
 os_error * err;

 err=NULL;

 switch(f->type)
 {
     case CBOXR:
                cpvrbox(f,box,file,buff);
                break;

      case CBOX:
                cpvbox(f,box,file,buff);
                break;

  case CELLIPSE:
                cpvellipse(f,box,file,buff);
                break;

     case CPATH:
                cpvpath(f,box,file,buff);
                break;
 }

 return(err);
}



os_error * clipopen(clipbuff * buffer)
{
 os_error * err;

 err=flex_alloc((flex_ptr)&buffer->path,CPCHUNK);
 buffer->n=0;

 return(err);
}

os_error * clipreset(clipbuff * buffer)
{
 buffer->n=0;
 return(NULL);
}

os_error * clipclose(clipbuff * buffer)
{
 return(flex_free((flex_ptr)&buffer->path));
}

/****************************************************************************/

/*

  We clip two things.
  First column guides are lines that are clipped to the frame border.
  Second the vboxes for the curve are clipped to the text containing curve.
  they are then clipped by any overlapping text frames.

*/

/****************************************************************************/



/****************************************************************************/

os_error * rboxwidth(curvestr * c,int y,int * xlo,int * xhi)
{
 int xradius;
 int yradius;
 int xl;
 int xh;
 int yl;
 int yh;
 int h;
 int w;

 xradius=c->data.rbox.xradius;
 if(xradius>c->box.w/2) xradius=c->box.w/2;

 yradius=c->data.rbox.yradius;
 if(yradius>c->box.h/2) yradius=c->box.h/2;

 xl=c->box.x+xradius;
 xh=c->box.x+c->box.w-xradius;
 yl=c->box.y+yradius;
 yh=c->box.y+c->box.h-yradius;

 if(y>yl && y<yh)
 {
  *xlo=c->box.x;
  *xhi=c->box.x+c->box.w;
 }
 else
 if(y<yl)
 {
  h=(yl-y);
  w=pythagd(xradius,scale(h,xradius,yradius));
  *xlo=xl-w;
  *xhi=xh+w;
 }
 else    /* y>yh */
 {
  h=(y-yh);
  w=pythagd(xradius,scale(h,xradius,yradius));
  *xlo=xl-w;
  *xhi=xh+w;
 }

 return(NULL);
}


os_error * ellipsewidth(curvestr * c,int y,int * xlo,int * xhi)
{
 int cx;
 int cy;
 int ht;
 int hb;
 int wl;
 int wr;
 int x;


 ht=c->box.h/2;
 hb=c->box.h-ht;
 wl=c->box.w/2;
 wr=c->box.w-wl;

 cx=c->box.x+wl;
 cy=c->box.y+ht;

 y-=cy;

 if(y==-ht || y==hb)
 {
  *xlo=*xhi=cx;
 }
 else
 if(!y)
 {
  *xlo=cx-wl;
  *xhi=cx+wr;
 }
 else
 {
  y=scale(y,wl,ht);
  x=pythagd(wl,y);
  *xlo=cx-x;
  *xhi=cx+x;
 }


 return(NULL);
}




/*****************************************************************************/

static int pupmax;
static int pupmin;
static int pdownmax;
static int pdownmin;
static int seendown;
static int seenup;


static void pathboxsub(int type,coordstr * cx,int ylo,int yhi)
{
 int x0;
 int x1;
 int y0;
 int y1;


 if(type==1)
 {
  x0=cx[0].x;
  y0=cx[0].y;
  x1=cx[1].x;
  y1=cx[1].y;
 }
 else
 {
  bezgetext1(cx[0].x,cx[1].x,cx[2].x,cx[3].x,&x0,&x1,NULL,NULL);
  y0=cx[0].y;
  y1=cx[3].y;

/* dprintf(0,"area x0=%d x1=%d y0=%d y1=%d",x0,x1,y0,y1);  */

 }

 if(y0<y1) /* down in other words */
 {
  if(y0>=ylo && y1<=yhi)
  {
   if(!seendown)
   {
    pdownmax=MAX(x0,x1);
    pdownmin=MIN(x0,x1);
    seendown=1;
   }
   else
   {
    pdownmax=MAX(pdownmax,x0);
    pdownmax=MAX(pdownmax,x1);
    pdownmin=MIN(pdownmin,x0);
    pdownmin=MIN(pdownmin,x1);
   }
  }
 }
 else /* y0>=y1 */
 if(y0>y1)
 {
  if(y1>=ylo && y0<=yhi)
  {
   if(!seenup)
   {
    pupmax=MAX(x0,x1);
    pupmin=MIN(x0,x1);
    seenup=1;
   }
   else
   {
    pupmax=MAX(pupmax,x0);
    pupmax=MAX(pupmax,x1);
    pupmin=MIN(pupmin,x0);
    pupmin=MIN(pupmin,x1);
   }
  }
 }
}





static void pathboxbezsub(coordstr * cx,coordstr * tray,coordstr * bray)
{
 cutstr    result[2*MAXBCUT];
 bezier    a;
 bezier    b;
 coordstr  cy[4];
 int       cuts;
 int       u;
 int       i;


/* dprintf(2,"bezsub y0=%d y1=%d y2=%d y3=%d",cx[0].y,cx[1].y,cx[2].y,cx[3].y);
 dprintf(3,"bezsub x0=%d x1=%d x2=%d x3=%d",cx[0].x,cx[1].x,cx[2].x,cx[3].x);
  */

 if(cx[0].y<cx[3].y)
 {
  cuts =hbintersectionr(tray,cx,result);
  cuts+=hbintersectionr(bray,cx,result+cuts);
 }
 else
 {
  cuts =hbintersectionr(bray,cx,result);
  cuts+=hbintersectionr(tray,cx,result+cuts);
 }


/* dprintf(3,"cuts=%d",cuts);  */


 u=0;
 a=*((bezier *)cx);

 for(i=0;i<cuts;i++)
 {
  u=scale(result[i].u2-u,0x10000,0x10000-u);

 /* dprintf(0,"split u=%x",u);  */

  if(u)
  {
   bezsplit2(&a,(bezier *)cy,&b,u);

           cy[3].y=result[i].c.y;
   if(i>0) cy[0].y=result[i-1].c.y;

   pathboxsub(2,cy,tray->y,bray->y);

   a=b;
  }

 }

 *((bezier *)cy)=a;
 if(i>0) cy[0].y=result[i-1].c.y;

 pathboxsub(2,cy,tray->y,bray->y);
}





static os_error * pathbox(filestr * file,curvestr * c,int ylo,int yhi)
{
 pathstr   * p;
 int       * q;
 coordstr    tray[2];
 coordstr    bray[2];
 int         type;
 int         i;
 segstr      seg;
 int         xlo;
 int         xhi;
 int         cuts;
 int         code;
 int         n;
 coordstr    cx[2];
 coordstr    cy[4];
 bezsplitstr split[3];


/* dprintf(0,"pathbox ylo=%d yhi=%d",ylo,yhi); */


 seenup=seendown=0;

 p=(pathstr*)fob_find(c->data.path.path,file);

/* dprintf(0,"pathbox1 p=%x q=%d path=%d",p,p->data,c->data.path.path); */

 q=p->data;
 newsegment(&seg);

/* dprintf(0,"pathbox2"); */

 tray[0].x=c->box.x-10;
 tray[0].y=ylo;
 tray[1].x=c->box.x+c->box.w+10;
 tray[1].y=ylo;

 bray[0].x=c->box.x-10;
 bray[0].y=yhi;
 bray[1].x=c->box.x+c->box.w+10;
 bray[1].y=yhi;

 xlo=c->box.x;
 xhi=xlo+c->box.w;

 while(1)
 {
 /* dprintf(0,"NEXT SEG IN"); */

  nextsegment(&seg,q);

 /* dprintf(0,"type=%d",seg.type); */

  if((type=seg.type)==0) break;


  if(type==1) 
  {
   cuts=0;
   if(seg.cx[0].y<seg.cx[1].y)
   {
    code=hintersectionr(tray,seg.cx,&cy[cuts]);
    if(code && !(code & INTERCO)) cuts++;
    code=hintersectionr(bray,seg.cx,&cy[cuts]);
    if(code && !(code & INTERCO)) cuts++;
   }
   else
   {
    code=hintersectionr(bray,seg.cx,&cy[cuts]);
    if(code && !(code & INTERCO)) cuts++;
    code=hintersectionr(tray,seg.cx,&cy[cuts]);
    if(code && !(code & INTERCO)) cuts++;
   }

   cx[0]=seg.cx[0];

   /* this code can generate 0 height segments */

   if(cuts)
   {
    for(i=0;i<cuts;i++)
    {
     cx[1]=cy[i];
     pathboxsub(1,cx,ylo,yhi);
     cx[0]=cx[1];
    }
   }

   cx[1]=seg.cx[1];

   pathboxsub(1,cx,ylo,yhi);
  }
  else
  if(type==2)
  {
   bezchunks((bezier *)seg.cx,split,&n,NULL);

   for(i=0;i<n;i++)
   {
    pathboxbezsub((coordstr*)&split[i].b,tray,bray);
   }
  }
 }


/* dprintf(0,"pathbox seendown=%d seenup=%d downmax=%d upmin=%d",
               seendown,seenup,pdownmax,pupmin); */


 if(!seendown || !seenup)
 {
  c->box.w=0;
 }
 else
 if(pdownmax<pupmin)
 {
  c->box.x=pdownmax;
  c->box.w=pupmin-pdownmax;
 }
 else
 if(pupmax<pdownmin)
 {
  c->box.x=pupmax;
  c->box.w=pdownmin-pupmax;
 }
 else
 {
  c->box.w=0;
 }



/* if(yhi==(c->box.y+c->box.h))
 {
  dprintf(0,"pathbox x=%d w=%d",c->box.x,c->box.w);


 } */


 return(NULL);
}





os_error * curvebox(filestr * file,curvestr * c,int ylo,int yhi)
{
 os_error * err;
 int        ymin;
 int        ymax;
 int        xlo1;
 int        xhi1;
 int        xlo2;
 int        xhi2;

 err=NULL;

 ymin=c->box.y;
 ymax=ymin+c->box.h;

 if(ylo>ymax || yhi<ymin)
 {
  c->box.w=-1;
 }
 else
 {
  if(ylo<ymin) ylo=ymin;
  if(yhi>ymax) yhi=ymax;

  switch(c->type)
  {
      case CBOX:          /* done */
                break;

     case CBOXR:
                rboxwidth(c,ylo,&xlo1,&xhi1);
                rboxwidth(c,yhi,&xlo2,&xhi2);
                xlo1=MAX(xlo1,xlo2);
                xhi1=MIN(xhi1,xhi2);
                c->box.x=xlo1;
                c->box.w=xhi1-xlo1;
                break;

  case CELLIPSE:
                ellipsewidth(c,ylo,&xlo1,&xhi1);
                ellipsewidth(c,yhi,&xlo2,&xhi2);
                xlo1=MAX(xlo1,xlo2);
                xhi1=MIN(xhi1,xhi2);
                c->box.x=xlo1;
                c->box.w=xhi1-xlo1;
                break;

     case CPATH:
                err=pathbox(file,c,ylo,yhi);
                break;
  }

  c->box.y=ylo;
  c->box.h=yhi-ylo;
 }
 return(err);
}


/****************************************************************************/

#define CSBCHUNK 32*sizeof(csbuffelement)

os_error * createcsbuffer(csbuffstr * csbf)
{
 os_error * err;

          err=flex_alloc((flex_ptr)&csbf->c[CSTOP],CSBCHUNK);
 if(!err) err=flex_alloc((flex_ptr)&csbf->c[CSBOT],CSBCHUNK);

 csbf->n[CSTOP]=csbf->n[CSBOT]=0;

 return(err);
}

os_error * deletecsbuffer(csbuffstr * csbf)
{
 flex_free((flex_ptr)&csbf->c[CSTOP]);
 return(flex_free((flex_ptr)&csbf->c[CSBOT]));
}


void clearcsbuffer(csbuffstr * csbf)
{
 csbf->n[CSTOP]=csbf->n[CSBOT]=0;
}


static int cscmp(const void * p1,const void * p2)
{
 csbuffelement * q1;
 csbuffelement * q2;

 q1=(csbuffelement *)p1;
 q2=(csbuffelement *)p2;

 if(q1->cx.y==q2->cx.y)
 {
  if(q1->cx.x==q2->cx.x)
  {
   if(q1->c==q2->c)
   {
    if(q1->s.oldindex==q2->s.oldindex)
    {
     return(q1->u-q2->u);
    }
    else return(q1->s.oldindex-q2->s.oldindex);
   }
   else return(q1->c-q2->c);
  }
  else return(q1->cx.x-q2->cx.x);
 }
 else return(q1->cx.y-q2->cx.y);
}







static int cscmpb(const void * p1,const void * p2)
{
 csbuffelement * q1;
 csbuffelement * q2;
 big             temp1;
 big             temp2;
 big             temp3;
 int             du1;
 int             du2;
 int             s1;
 int             s2;



 q1=(csbuffelement *)p1;
 q2=(csbuffelement *)p2;

 if(q1->cx.y==q2->cx.y)
 {
  if(q1->cx.x==q2->cx.x)
  {
   if(q1->dy && q2->dy)
   {
    bigmul(&temp1,q1->dx,q2->dy);
    bigmul(&temp2,q2->dx,q1->dy);
    bigsub(&temp1,&temp2,&temp3);

 /*   dprintf(0,"cmpb dx1=%d dx2=%d dy1=%d dy2=%d -> lo=%d hi=%d",
                  q1->dx,q2->dx,q1->dy,q2->dy,temp3.lo,temp3.hi); */

    if((q1->dy ^ q2->dy)<0)
    {
     if(temp3.hi<0) return(1);
     else
     if(temp3.hi || temp3.lo) return(-1);
     else                     return(0);
    }
    else
    {
     if(temp3.hi<0) return(-1);
     else
     if(temp3.hi || temp3.lo) return(1);
     else                     return(0);
    }
   }
   else
   {
    if(q1->u==q1->umin) du1=1;
    else                du1=-1;
    s1=q1->dx ^ du1;

    if(q2->u==q2->umin) du2=1;
    else                du2=-1;
    s2=q2->dx ^ du2;

    if(s1<0 && s2>=0) return(-1);
    else
    if(s1>=0 && s2<0)  return(1);
    else
    {
     /* look at d2y/dx2 */






     return(0);
    }
   }
  }
  else return(q1->cx.x-q2->cx.x);
 }
 else return(q1->cx.y-q2->cx.y);
}




static int cscmpt(const void * p1,const void * p2)
{
 csbuffelement * q1;
 csbuffelement * q2;
 big             temp1;
 big             temp2;
 big             temp3;
 int             du1;
 int             du2;
 int             s1;
 int             s2;



 q1=(csbuffelement *)p1;
 q2=(csbuffelement *)p2;

 if(q1->cx.y==q2->cx.y)
 {
  if(q1->cx.x==q2->cx.x)
  {
   if(q1->dy && q2->dy)
   {
    bigmul(&temp1,q1->dx,q2->dy);
    bigmul(&temp2,q2->dx,q1->dy);
    bigsub(&temp1,&temp2,&temp3);


   /* dprintf(0,"cmpt dx1=%d dx2=%d dy1=%d dy2=%d -> lo=%d hi=%d",
                  q1->dx,q2->dx,q1->dy,q2->dy,temp3.lo,temp3.hi); */


    if((q1->dy ^ q2->dy)<0)
    {
     if(temp3.hi<0) return(-1);
     else
     if(temp3.hi || temp3.lo) return(1);
     else                     return(0);
    }
    else
    {
     if(temp3.hi<0) return(1);
     else
     if(temp3.hi || temp3.lo) return(-1);
     else                     return(0);
    }
   }
   else
   {
    if(q1->u==q1->umin) du1=1;
    else                du1=-1;
    s1=q1->dx ^ du1;

    if(q2->u==q2->umin) du2=1;
    else                du2=-1;
    s2=q2->dx ^ du2;

    if(s1<0 && s2>=0) return(-1);
    else
    if(s1>=0 && s2<0)  return(1);
    else
    {
     /* look at d2y/dx2 */





     return(0);
    }
   }
  }
  else return(q1->cx.x-q2->cx.x);
 }
 else return(q1->cx.y-q2->cx.y);
}





os_error * sortcsbuffer(csbuffstr * csbf)
{
 qsort(csbf->c[CSTOP],csbf->n[CSTOP],sizeof(csbuffelement),cscmpt);
 qsort(csbf->c[CSBOT],csbf->n[CSBOT],sizeof(csbuffelement),cscmpb);
 return(NULL);
}


static os_error * addintersection(csbuffstr * csbf,csbuffelement * ce,int line)
{
 os_error * err;

 err=flex_chunk((flex_ptr)&csbf->c[line],
                (csbf->n[line]+1)*sizeof(csbuffelement),CSBCHUNK);
 if(!err)
 {
  csbf->c[line][csbf->n[line]]=*ce;
  csbf->n[line]++;

#ifdef DBUG2
  dprintf(0,"addinter index=%d old=%d",ce->s.index,ce->s.oldindex);
#endif

 }

 return(err);
}


static void newelement(csbuffelement * ce,curvestr * c)
{
 memset(ce,0,sizeof(csbuffelement));
 ce->c=c;
}


os_error * simplifycsbuffer(csbuffstr * csbf)
{
 int        r;
 int        w;
 int        l;

 for(l=CSTOP;l<=CSBOT;l++)
 {
  w=0;
  for(r=0;r<csbf->n[l];r++)
  {
   if(w>0)
   {
    if(cscmp((csbuffelement*)&csbf->c[l][r],(csbuffelement*)&csbf->c[l][w-1]))
    {
     if(w!=r) csbf->c[l][w]=csbf->c[l][r];
     w++;
    }
   }
   else
   {
    w++;
   }
  }

  csbf->n[l]=w;
 }
 return(NULL);
}



/*****************************************************************************/


static os_error * csaddpoint(csbuffstr * csbf,csbuffelement * ce,int line,
                                     curvestr * tcurve,int nc,filestr * file)
{
 os_error      * err;
 int             i;

 err=NULL;
 i=nc;      /* compilers benefit */

 if(tcurve)
 {
  for(i=0;i<nc;i++)
  {
   if(insidecurve(ce->cx.x,ce->cx.y,&tcurve[i],NULL,file)) break;
  }
 }

 if(!tcurve || i<nc)   // if no tcurve or if point is inside tcurve
 {
  addintersection(csbf,ce,line);
 }

 return(err);
}






/* add all the max/minima of ccurve that lie in tcurve */

static os_error * csaddbox(csbuffstr * csbf,curvestr * ccurve,
                           curvestr * tcurve,int nc,filestr * file)
{
 os_error      * err;
 csbuffelement   ce;

 newelement(&ce,ccurve);

 ce.dx=0;
 ce.dy=1;

 ce.cx.x=ccurve->box.x;
 ce.cx.y=ccurve->box.y;

          err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);

 ce.cx.x+=ccurve->box.w;

 if(!err) err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);

 ce.cx.y+=ccurve->box.h;

 if(!err) err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);

 ce.cx.x-=ccurve->box.w;

 if(!err) err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);

 return(err);
}



static os_error * csaddboxr(csbuffstr * csbf,curvestr * ccurve,
                            curvestr * tcurve,int nc,filestr * file)
{
 os_error      * err;
 csbuffelement   ce;
 int             xradius;
 int             yradius;


 xradius=ccurve->data.rbox.xradius;
 if(xradius>ccurve->box.w/2) xradius=ccurve->box.w/2;

 yradius=ccurve->data.rbox.yradius;
 if(yradius>ccurve->box.h/2) yradius=ccurve->box.h/2;


 newelement(&ce,ccurve);

 ce.dx=0;
 ce.dy=1;

 ce.cx.x=ccurve->box.x+xradius;
 ce.cx.y=ccurve->box.y;

          err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);

 ce.cx.x+=ccurve->box.w-2*xradius;

 if(!err) err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);

 ce.cx.y+=ccurve->box.h;

 if(!err) err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);

 ce.cx.x-=ccurve->box.w-2*xradius;

 if(!err) err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);

 return(err);
}


static os_error * csaddellipse(csbuffstr * csbf,curvestr * ccurve,
                               curvestr * tcurve,int nc,filestr * file)
{
 os_error     * err;
 int            cx;
 int            cy;
 int            ht;
 int            wl;
 int            hb;
 int            wr;
 int            oldn;
 csbuffelement  ce;


 ht=ccurve->box.h/2;
 hb=ccurve->box.h-ht;
 wl=ccurve->box.w/2;
 wr=ccurve->box.w-wl;

 cx=ccurve->box.x+wl;
 cy=ccurve->box.y+ht;


 newelement(&ce,ccurve);

 ce.cx.x=cx;
 ce.cx.y=cy-ht;

 /* add max */

 ce.umin=0;
 ce.u=ce.umax=0x8000;
 ce.dx=(cy-ce.cx.y);
 ce.dy=0;

 oldn=csbf->n[CSBOT];

          err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);

 if(oldn!=csbf->n[CSBOT])
 {
  ce.u=ce.umin=0x8000;
  ce.umax=0x10000;
  ce.dx=(cy-ce.cx.y);
  ce.dy=0;
  if(!err) err=csaddpoint(csbf,&ce,CSBOT,NULL,nc,file);
 }

 /* add min */

 ce.cx.y=cy+hb;

 ce.dx=(cy-ce.cx.y);
 ce.dy=0;
 ce.umin=0;
 ce.u=ce.umax=0x8000;

 oldn=csbf->n[CSTOP];

 if(!err) err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);

 if(oldn!=csbf->n[CSTOP])
 {
  ce.u=ce.umin=0x8000;
  ce.umax=0x10000;
  ce.dx=(cy-ce.cx.y);
  ce.dy=0;
  if(!err) err=csaddpoint(csbf,&ce,CSTOP,NULL,nc,file);
 }

 return(err);
}




/* compare dy/dx for segment, and next segment  */
/* compare last segment with first              */
/* special case of run of segments with dy/dx=0 */


static int gradiant(int dy,int dx)
{
 if(dy==0)  return(0);
 else
 if(dx==0)
 {
  if(dy>=0) return(1);
  else      return(-1);
 }
 else
 if((dy ^ dx)<0) return(-1);
 else            return(1);
}



typedef struct csrootstr
{
 int         n;
 bezier      b;
 bezsplitstr split[3];
} csrootstr;



static void seguvals(segstr * seg,int second,csrootstr * root,
                                             csbuffelement * ce)
{
 if(seg->type==2)
 {
  if(second)
  {
   ce->umax=root->split[root->n-1].umax;
   ce->umin=root->split[root->n-1].umin;
  }
  else
  {
   ce->umax=root->split[0].umax;
   ce->umin=root->split[0].umin;
  }
/* dprintf(0,"seguvals x=%d y=%d umin=%x umax=%x",ce->cx.x,ce->cx.y,
                       ce->umin,ce->umax); */
 }
}



static os_error * csaddpath(csbuffstr * csbf,curvestr * ccurve,
                            curvestr * tcurve,int nc,filestr * file)
{
 os_error      * err;
 int             type;
 int             state;
 int             d1;
 int             d2;
 int             lastd2;
 int             lastd1;
 int             df;
 int             path;
 segstr          seg;
 segstr          fseg;
 segstr          oldseg;
 segstr          nextseg;
 int             finalindex;
 int             oldn;
 csbuffelement   ce;
 int             i;
 int             wrapped;
 csrootstr       froot;
 csrootstr       oldroot;
 csrootstr       root;


 err=NULL;

 finalindex=0;  /* compiler */
 df=0;          /* compiler */
 lastd1=0;      /* compiler */
 lastd2=0;      /* compiler */
 d1=0;          /* compiler */
 d2=0;          /* compiler */


 newelement(&ce,ccurve);

 path=ccurve->data.path.path;
 newsegment(&nextseg);

 while(1)
 {
  seg=nextseg;

  wrapped=0;
  state=0;

  while(1)
  {
   nextsegmentpath(&seg,path,file);
   if(seg.type==0) break;

   if(seg.type==3)         /* close path */
   {
    nextseg=seg;
    seg.index=seg.bindex;

    if(state==0) break;   /* give up */

    nextsegmentpath(&seg,path,file);

    wrapped=1;
   }

   type=seg.type;

   if(type==2) /* path segment may have local max/min */
   {
    bezchunks((bezier *)seg.cx,root.split,&root.n,&root.b);

    if(!wrapped)
    {
     oldn=0; /* compiler */

     for(i=1;i<root.n;i++)
     {
      ce.cx=root.split[i].b.left;
      ce.umin=root.split[i-1].umin;
      ce.u=ce.umax=root.split[i-1].umax;
      ce.s=seg;

      ce.dy=0;
      bezdev(&root.b,ce.u,&ce.dx,NULL,&ce.d2x,&ce.d2y);

#ifdef DBUG
      dprintf(0,"local index=%d",seg.index);
    
      dprintf(0,"add1 u=%x umin=%x umax=%x d2y=%x x=%d y=%d",
                  ce.u,ce.umin,ce.umax,ce.d2y,ce.cx.x,ce.cx.y); 
#endif


      if(ce.d2y>0)
      {
       oldn=csbf->n[CSBOT];
       err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);
      }
      else
      if(ce.d2y<0)
      {
       oldn=csbf->n[CSTOP];
       err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);
      }

      ce.u=ce.umin=root.split[i].umin;
      ce.umax=root.split[i].umax;


#ifdef DBUG
      dprintf(0,"add2 u=%x umin=%x umax=%x d2y=%x x=%d y=%d",
                 ce.u,ce.umin,ce.umax,ce.d2y,ce.cx.x,ce.cx.y); 
#endif


      if(ce.d2y>0)
      {
       if(oldn!=csbf->n[CSBOT])
       {
        err=csaddpoint(csbf,&ce,CSBOT,NULL,nc,file);
       }
      }
      else
      if(ce.d2y<0)
      {
       if(oldn!=csbf->n[CSTOP])
       {
        err=csaddpoint(csbf,&ce,CSTOP,NULL,nc,file);
       }
      } 
     }
    }

    d1=gradiant(seg.cx[1].y-seg.cx[0].y,1);    /* gradiant along path */
    d2=gradiant(seg.cx[3].y-seg.cx[2].y,1);

    if(d1==0) 
    {
     d1=2*seg.cx[0].y-4*seg.cx[1].y+2*seg.cx[2].y;
     if(d1>0) d1=1;
     else
     if(d1<0) d1=-1;
     else
     {
      d1=gradiant(seg.cx[3].y-seg.cx[0].y,1);
     }
    }

    if(d2==0)
    {
     d2=2*seg.cx[1].y-4*seg.cx[2].y+2*seg.cx[3].y;
     if(d2>0) d2=1;
     else
     if(d2<0) d2=-1;
     else
     {
      d2=gradiant(seg.cx[3].y-seg.cx[0].y,1);
     }
    }

   } 
   else
   if(type==1)
   {
    d1=d2=gradiant(seg.cx[1].y-seg.cx[0].y,1); /* gradiant along path */
   }


   if(type==1 || type==2)
   {
    if(state==0)
    {
     if(d1!=d2 || d1) /* not starting in flat bit */
     {
      df=d1;
      lastd1=d1;
      lastd2=d2;
      state=1;
      finalindex=seg.index;
     }
    }
    else     /* state !0 */
    {
     if(state==1)
     {
      if((d1>0 && lastd2<0) || (d1==lastd2 && !d1 && d2>0 && lastd1<0))
                                                           /*found a max*/
      {
       oldn=csbf->n[CSBOT];

       ce.cx=seg.cx[0];
       ce.u=0;
       ce.s=seg;
       seggrad(&seg,0,&ce.dx,&ce.dy);
       seguvals(&seg,0,&root,&ce);

#ifdef DBUG
       dprintf(0,"add max x=%d y=%d",ce.cx.x,ce.cx.y);
#endif

       err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);
       if(oldn!=csbf->n[CSBOT])
       {
        ce.u=0x10000;
        ce.s=oldseg;
        seggrad(&oldseg,1,&ce.dx,&ce.dy);
        seguvals(&oldseg,1,&oldroot,&ce);

        /* NB NULL tcurve forces addition */

        err=csaddpoint(csbf,&ce,CSBOT,NULL,nc,file);
       }
      }
      else
      if((d1<0 && lastd2>0) || (d1==lastd2 && !d1 && d2<0 && lastd1>0)) 
                                                           /*found a min*/
      {
       oldn=csbf->n[CSTOP];

       ce.cx=seg.cx[0];
       ce.u=0;
       ce.s=seg;
       seggrad(&seg,0,&ce.dx,&ce.dy);
       seguvals(&seg,0,&root,&ce);

#ifdef DBUG
       dprintf(0,"add min x=%d y=%d",ce.cx.x,ce.cx.y);
#endif

       err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);
       if(oldn!=csbf->n[CSTOP])
       {
        ce.u=0x10000;
        ce.s=oldseg;
        seggrad(&oldseg,1,&ce.dx,&ce.dy);
        seguvals(&oldseg,1,&oldroot,&ce);

        err=csaddpoint(csbf,&ce,CSTOP,NULL,nc,file);
       }
      }
      else
      if(d1==d2 && !d1)    /* found a flat bit */
      {
       df=lastd2;
       state=2;
       fseg=oldseg;
       froot=oldroot;
      }
     }
     else
     if(state==2) /* following flat top */
     {
      if(d1>0 && df<0)    /* found a flat max */
      {
       if(fseg.type==1) ce.cx=fseg.cx[1];
       else             ce.cx=fseg.cx[3];

       ce.u=0x10000;
       ce.s=fseg;
       seggrad(&fseg,1,&ce.dx,&ce.dy);
       seguvals(&fseg,1,&froot,&ce);

#ifdef DBUG
       dprintf(0,"add flat max x=%d y=%d",ce.cx.x,ce.cx.y);
#endif

       err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);

       ce.cx=seg.cx[0];
       ce.u=0;
       ce.s=seg;
       seggrad(&seg,0,&ce.dx,&ce.dy);
       seguvals(&seg,0,&root,&ce);

       err=csaddpoint(csbf,&ce,CSBOT,tcurve,nc,file);

       state=1;
      }
      else
      if(d1<0 && df>0)    /* found a flat min */
      {
       if(fseg.type==1) ce.cx=fseg.cx[1];
       else             ce.cx=fseg.cx[3];

       ce.u=0x10000;
       ce.s=fseg;
       seggrad(&fseg,1,&ce.dx,&ce.dy);
       seguvals(&fseg,1,&froot,&ce);

#ifdef DBUG
       dprintf(0,"add flat min x=%d y=%d",ce.cx.x,ce.cx.y);
#endif

       err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);

       ce.cx=seg.cx[0];
       ce.u=0;
       ce.s=seg;
       seggrad(&seg,0,&ce.dx,&ce.dy);
       seguvals(&seg,0,&root,&ce);

       err=csaddpoint(csbf,&ce,CSTOP,tcurve,nc,file);

       state=1;
      }
      else
      if(d2!=0)
      {
                         /* assume there is a local max */
       state=1;
      }
     }

     lastd1=d1;
     lastd2=d2;
    
     if(seg.index==finalindex) break;
    }
   }

   oldseg=seg;
   if(type==2) oldroot=root;
  }

  if(seg.type==0) break;
 }
 return(err);
}


/*****************************************************************************/

/* add all intersections of two boxes */


/* we have removed all intersections which occur at extrema                 */
/* this is in the expectation that another phase of the algorithm will pick */
/* these points up                                                          */

/* example two boxes of the same height and vertical posn                   */
/* which overlap horizontally                                               */

/* this has to be consistent with the same boxes made out of lines          */



static os_error * cscutboxbox(csbuffstr * csbf,curvestr * ccurve1,
                                               curvestr * ccurve2,
                             curvestr * tcurve,filestr * file)
{
 os_error    * err;
 int           x0;
 int           y0;
 int           x1;
 int           y1;
 csbuffelement cc;
 csbuffelement ct;

 err=NULL;

 newelement(&cc,ccurve1);
 newelement(&ct,ccurve2);

 x0=ccurve1->box.x;
 y1=ccurve1->box.y;
 x1=x0+ccurve1->box.w;
 y0=y1+ccurve1->box.h;


 if(y0>ccurve2->box.y && y0<(ccurve2->box.y+ccurve2->box.h))
 {
  if(x0<ccurve2->box.x && x1>ccurve2->box.x)
  {
   ct.cx.y=y0;
   ct.cx.x=ccurve2->box.x;
   ct.dy=1;
   ct.dx=0;
   csaddpoint(csbf,&ct,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&ct,CSBOT,tcurve,1,file);
  }

  if(x1>(ccurve2->box.x+ccurve2->box.w) && x0<(ccurve2->box.x+ccurve2->box.w))
  {
   ct.cx.y=y0;
   ct.cx.x=ccurve2->box.x+ccurve2->box.w;
   ct.dy=1;
   ct.dx=0;
   csaddpoint(csbf,&ct,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&ct,CSBOT,tcurve,1,file);
  }
 }



 if(y1>ccurve2->box.y && y1<(ccurve2->box.y+ccurve2->box.h))
 {
  if(x0<ccurve2->box.x && x1>ccurve2->box.x)
  {
   ct.cx.y=y1;
   ct.cx.x=ccurve2->box.x;
   ct.dy=1;
   ct.dx=0;
   csaddpoint(csbf,&ct,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&ct,CSBOT,tcurve,1,file);
  }

  if(x1>(ccurve2->box.x+ccurve2->box.w) && x0<(ccurve2->box.x+ccurve2->box.w))
  {
   ct.cx.y=y1;
   ct.cx.x=ccurve2->box.x+ccurve2->box.w;
   ct.dy=1;
   ct.dx=0;
   csaddpoint(csbf,&ct,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&ct,CSBOT,tcurve,1,file);
  }
 }



 if(x0>ccurve2->box.x && x0<(ccurve2->box.x+ccurve2->box.w))
 {
  if(y0>ccurve2->box.y && y1<ccurve2->box.y)
  {
   cc.cx.x=x0;
   cc.cx.y=ccurve2->box.y;
   cc.dy=1;
   cc.dx=0;
   csaddpoint(csbf,&cc,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&cc,CSBOT,tcurve,1,file);
  }

  if(y0>(ccurve2->box.y+ccurve2->box.h) && y1<(ccurve2->box.y+ccurve2->box.h))
  {
   cc.cx.x=x0;
   cc.cx.y=ccurve2->box.y+ccurve2->box.h;
   cc.dy=1;
   cc.dx=0;
   csaddpoint(csbf,&cc,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&cc,CSBOT,tcurve,1,file);
  }
 }

 if(x1>ccurve2->box.x && x1<(ccurve2->box.x+ccurve2->box.w))
 {
  if(y0>ccurve2->box.y && y1<ccurve2->box.y)
  {
   cc.cx.x=x1;
   cc.cx.y=ccurve2->box.y;
   cc.dy=1;
   cc.dx=0;
   csaddpoint(csbf,&cc,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&cc,CSBOT,tcurve,1,file);
  }

  if(y0>(ccurve2->box.y+ccurve2->box.h) && y1<(ccurve2->box.y+ccurve2->box.h))
  {
   cc.cx.x=x1;
   cc.cx.y=ccurve2->box.y+ccurve2->box.h;
   cc.dy=1;
   cc.dx=0;
   csaddpoint(csbf,&cc,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&cc,CSBOT,tcurve,1,file);
  }
 }

 return(err);
}













static os_error * cscutpathbox(csbuffstr * csbf,
                               curvestr * ccurve1,curvestr * refccurve1,
                               curvestr * ccurve2,curvestr * refccurve2,
                               curvestr * tcurve,filestr * file)
{
 os_error      * err;
 csbuffelement   cpath;
 csbuffelement   cbox;
 coordstr        ray[2];
 cutstr          result[MAXBCUT];
 curvestr      * box;
 curvestr      * path;
 int             cuts;
 int             type;
 int             n;
 int             xlo;
 int             xhi;
 int             ylo;
 int             yhi;
 int             pathk;
 segstr          seg;
 int             code;


 err=NULL;


 if(ccurve1->type==CBOX) 
 {
  box=ccurve1;
  path=ccurve2;
  newelement(&cpath,refccurve2);
  newelement(&cbox,refccurve1);
 }
 else
 {
  box=ccurve2;
  path=ccurve1;
  newelement(&cbox,refccurve2);
  newelement(&cpath,refccurve1);
 }




 xlo=box->box.x;
 xhi=xlo+box->box.w;
 ylo=box->box.y;
 yhi=ylo+box->box.h;



 ray[0].x=xlo;     /* a horizontal line */
 ray[0].y=ylo;
 ray[1].x=xhi;
 ray[1].y=ylo;


 pathk=path->data.path.path;
 newsegment(&seg);
 while(1)
 {
  nextsegmentpath(&seg,pathk,file);
  if((type=seg.type)==0) break;

  cuts=0;

  if(type==1) 
  {
   code=hintersectionc(ray,seg.cx,result);
   if(code==INTER) cuts++;


  }
  else
  if(type==2)
  {
   cuts=hbintersectionc(ray,seg.cx,result);


  }

  for(n=0;n<cuts;n++)
  {
   cpath.cx.x=result[n].c.x;
   cpath.cx.y=ylo;

   cpath.u=result[n].u2;
   cpath.umin=result[n].umin2;
   cpath.umax=result[n].umax2;

   cpath.dx=result[n].dx2;
   cpath.d2x=result[n].d2x2;
   cpath.dy=result[n].dy2;
   cpath.d2y=result[n].d2y2;

   cpath.s=seg;

#ifdef DBUG
   dprintf(0,"cutpath 1 hinter=%d x=%d y=%d index=%d",cpath.u,cpath.cx.x,cpath.cx.y,seg.index);
#endif

   csaddpoint(csbf,&cpath,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&cpath,CSBOT,tcurve,1,file);
  }
 }


 ray[0].x=xlo;    /* a horizontal line */
 ray[0].y=yhi;
 ray[1].x=xhi;
 ray[1].y=yhi;


 newsegment(&seg);
 while(1)
 {
  nextsegmentpath(&seg,pathk,file);
  if((type=seg.type)==0) break;

  cuts=0;

  if(type==1)
  {
   code=hintersectionc(ray,seg.cx,result);
   if(code==INTER) cuts++;

  }
  else
  if(type==2)
  {
   cuts=hbintersectionc(ray,seg.cx,result);


  }

  for(n=0;n<cuts;n++)
  {
   cpath.cx.x=result[n].c.x;
   cpath.cx.y=yhi;

   cpath.u=result[n].u2;
   cpath.umin=result[n].umin2;
   cpath.umax=result[n].umax2;

   cpath.dx=result[n].dx2;
   cpath.d2x=result[n].d2x2;
   cpath.dy=result[n].dy2;
   cpath.d2y=result[n].d2y2;

   cpath.s=seg;

#ifdef DBUG
   dprintf(0,"cutpath 2 hinter=%d",cpath.u);
#endif

   csaddpoint(csbf,&cpath,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&cpath,CSBOT,tcurve,1,file);
  }
 }


 ray[0].x=xlo;       /* a vertical line */
 ray[0].y=ylo;
 ray[1].x=xlo;
 ray[1].y=yhi;


 newsegment(&seg);
 while(1)
 {
  nextsegmentpath(&seg,pathk,file);
  if((type=seg.type)==0) break;

  cuts=0;

  if(type==1) 
  {
   code=vintersectionc(ray,seg.cx,result);
   if(code==INTER) cuts=1;


  }
  else
  if(type==2)
  {
   cuts=vbintersectionc(ray,seg.cx,result);


  }


  for(n=0;n<cuts;n++)
  {
   cpath.cx.x=xlo;
   cpath.cx.y=result[n].c.y;

   cpath.u=result[n].u2;
   cpath.umin=result[n].umin2;
   cpath.umax=result[n].umax2;

   cpath.dx=result[n].dx2;
   cpath.d2x=result[n].d2x2;
   cpath.dy=result[n].dy2;
   cpath.d2y=result[n].d2y2;

   cpath.s=seg;

#ifdef DBUG
   dprintf(0,"cutpath 1 vinter u=%x umin=%x umax=%x dy=%d x=%d y=%d",
           cpath.u,cpath.umin,cpath.umax,cpath.dy,cpath.cx.x,cpath.cx.y);
#endif

   if(cpath.dy)
   {
    csaddpoint(csbf,&cpath,CSTOP,tcurve,1,file);
    csaddpoint(csbf,&cpath,CSBOT,tcurve,1,file);
   }

   cbox.cx=cpath.cx;
   cbox.dx=result[n].dx1;
   cbox.dy=result[n].dy1;

   csaddpoint(csbf,&cbox,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&cbox,CSBOT,tcurve,1,file);
  }
 }


 ray[0].x=xhi;    /* a vertical line */
 ray[0].y=ylo;
 ray[1].x=xhi;
 ray[1].y=yhi;

 newsegment(&seg);
 while(1)
 {
  nextsegmentpath(&seg,pathk,file);
  if((type=seg.type)==0) break;

  cuts=0;

  if(type==1) 
  {
   code=vintersectionc(ray,seg.cx,result);
   if(code==INTER) cuts++;

  }
  else
  if(type==2)
  {
   cuts=vbintersectionc(ray,seg.cx,result);


  }

  for(n=0;n<cuts;n++)
  {
   cpath.cx.x=xhi;
   cpath.cx.y=result[n].c.y;

   cpath.u=result[n].u2;
   cpath.umin=result[n].umin2;
   cpath.umax=result[n].umax2;

   cpath.dx=result[n].dx2;
   cpath.d2x=result[n].d2x2;
   cpath.dy=result[n].dy2;
   cpath.d2y=result[n].d2y2;

   cpath.s=seg;

#ifdef DBUG
   dprintf(0,"cutpath 2 inter=%d",cpath.u);
#endif

   if(cpath.dy)
   {
    csaddpoint(csbf,&cpath,CSTOP,tcurve,1,file);
    csaddpoint(csbf,&cpath,CSBOT,tcurve,1,file);
   }

   cbox.cx=cpath.cx;
   cbox.dx=result[n].dx1;
   cbox.dy=result[n].dy1;

   csaddpoint(csbf,&cbox,CSTOP,tcurve,1,file);
   csaddpoint(csbf,&cbox,CSBOT,tcurve,1,file);
  }
 }
 return(err);
}





/* seems to miss case like horizontal line which goes through the joint */
/* of two vertical segments                                             */
/* skipping horizontal lines which go through extrema is OK though      */



static os_error * cscutpathpath(csbuffstr * csbf,
                                curvestr * ccurve1,curvestr * refccurve1,
                                curvestr * ccurve2,curvestr * refccurve2,
                                curvestr * tcurve,filestr * file)
{
 os_error      * err;
 csbuffelement   c1;
 csbuffelement   c2;
 cutstr          result[MAXBCUT];
 int             cuts;
 int             type1;
 int             type2;
 int             n;
 segstr          seg1;
 segstr          seg2;
 int             path1;
 int             path2;


 err=NULL;

 newelement(&c1,refccurve1);
 newelement(&c2,refccurve2);

 path1=ccurve1->data.path.path;
 path2=ccurve2->data.path.path;

 newsegment(&seg1);
 while(1)
 {
  nextsegmentpath(&seg1,path1,file);
  if((type1=seg1.type)==0) break;

  newsegment(&seg2);
  while(1)
  {
   nextsegmentpath(&seg2,path2,file);
   if((type2=seg2.type)==0) break;

   cuts=0;

   if(type1==1 && type2==1) 
   {
    cuts=intersection(seg1.cx,seg2.cx,result);
   }
   else
   if(type1==2 && type2==2)
   {
    cuts=bbintersection(seg1.cx,seg2.cx,result);
   }
   else                                  /* line first */
   if(type1==2 && type2==1)
   {
    cuts=bintersection(seg2.cx,seg1.cx,result);
   }
   else                                             
   if(type1==1 && type2==2)
   {
    cuts=bintersection(seg1.cx,seg2.cx,result);
   }

   for(n=0;n<cuts;n++)
   {
    c1.cx=result[n].c;
    c2.cx=c1.cx;

    if(result[n].type==INTER)
    {
     if(type1==2 && type2==1)
     {
      c1.u=result[n].u2;
      c1.umin=result[n].umin2;
      c1.umax=result[n].umax2;

      c2.u=result[n].u1;
      c2.umin=result[n].umin1;
      c2.umax=result[n].umax1;

      c1.dx=result[n].dx2;
      c2.dx=result[n].dx1;

      c1.dy=result[n].dy2;
      c2.dy=result[n].dy1;
     }
     else
     {
      c1.u=result[n].u1;
      c1.umin=result[n].umin1;
      c1.umax=result[n].umax1;

      c2.u=result[n].u2;
      c2.umin=result[n].umin2;
      c2.umax=result[n].umax2;

      c1.dx=result[n].dx1;
      c2.dx=result[n].dx2;

      c1.dy=result[n].dy1;
      c2.dy=result[n].dy2;
     }

     c1.s=seg1;
     c2.s=seg2;


#ifdef DBUG
     dprintf(0,"cutpathpath x=%d y=%d dy1=%d dy2=%d",
                                    c1.cx.x,c1.cx.y,c1.dy,c2.dy);
#endif

     if(c1.dy)
     {
      csaddpoint(csbf,&c1,CSTOP,tcurve,1,file);
      csaddpoint(csbf,&c1,CSBOT,tcurve,1,file);
     }

     if(c2.dy)
     {
      csaddpoint(csbf,&c2,CSTOP,tcurve,1,file);
      csaddpoint(csbf,&c2,CSBOT,tcurve,1,file);
     }
    }
   }
  }
 }
 return(err);
}


/*****************************************************************************/

/* add the extrema for ccurve */
/* only if inside tcurve      */


os_error * cextrema(csbuffstr * csbf,curvestr * ccurve,
                                     curvestr * tcurve,int nc,filestr * file)
{
 os_error * err;

 err=NULL;

 switch(ccurve->type)
 {
   case     CBOX:
                 csaddbox(csbf,ccurve,tcurve,nc,file);
                 break;

   case CELLIPSE:
                 csaddellipse(csbf,ccurve,tcurve,nc,file);
                 break;

   case    CBOXR:
                 csaddboxr(csbf,ccurve,tcurve,nc,file);
                 break;

   case    CPATH:
                 csaddpath(csbf,ccurve,tcurve,nc,file);
                 break;
 }

 return(err);
}



/* ccurve is the opaque clipping curve, anything lying inside it is obscured */
/* tcurve is the curve that we want clipped                                  */


// interesections of ccurve1 and ccurve2 that lie inside tcurve


os_error * cintersections(csbuffstr * csbf,curvestr * ccurve1,
                          curvestr * ccurve2,curvestr * tcurve,filestr * file)
{
 os_error * err;
 curvestr   xccurve1;
 curvestr   xccurve2;
 int        tempc1;
 int        tempc2;
         

 err=NULL;


 /* assert that if c1 and c2 don't overlap t then neither does intersection */

 if(tcurve && (!boxoverlap(&tcurve->box,&ccurve1->box) ||
               !boxoverlap(&tcurve->box,&ccurve2->box))) return(NULL);

 if(boxoverlap(&ccurve1->box,&ccurve2->box))
 {
  if(ccurve1->type==CBOX)
  {
   switch(ccurve2->type)
   {
    case     CBOX:
                  cscutboxbox(csbf,ccurve1,ccurve2,tcurve,file);
                  break;

    case CELLIPSE:
    case    CBOXR:
                  xccurve2=*ccurve2;
                  curvetopath(&xccurve2,file);
                  cscutpathbox(csbf,&xccurve2,ccurve2,ccurve1,ccurve1,
                                                             tcurve,file);
                  deletecurve(&xccurve2,file);
                  break;

    case    CPATH:
                  cscutpathbox(csbf,ccurve2,ccurve2,ccurve1,ccurve1,
                                                           tcurve,file);
                  break;
   }
  }
  else
  if(ccurve2->type==CBOX)
  {
   switch(ccurve1->type)
   {
    case     CBOX:
                  cscutboxbox(csbf,ccurve1,ccurve2,tcurve,file);
                  break;

    case CELLIPSE:
    case    CBOXR:
                  xccurve1=*ccurve1;
                  curvetopath(&xccurve1,file);
                  cscutpathbox(csbf,&xccurve1,ccurve1,ccurve2,ccurve2,
                                                             tcurve,file);
                  deletecurve(&xccurve1,file);
                  break;

    case    CPATH:
                  cscutpathbox(csbf,ccurve1,ccurve1,ccurve2,ccurve2,
                                                            tcurve,file);
                  break;
   }
  }
  else
  {
   xccurve1=*ccurve1;
   xccurve2=*ccurve2;
   tempc1=tempc2=0;

   if(ccurve1->type==CBOXR || ccurve1->type==CELLIPSE)
   {
    curvetopath(&xccurve1,file);
    tempc1=1;
   }

   if(ccurve2->type==CBOXR || ccurve2->type==CELLIPSE)
   {
    curvetopath(&xccurve2,file);
    tempc2=1;
   }

   /* ultimately path - path intersections */
  
   cscutpathpath(csbf,&xccurve1,ccurve1,&xccurve2,ccurve2,tcurve,file);

   if(tempc1) deletecurve(&xccurve1,file);
   if(tempc2) deletecurve(&xccurve2,file);
  }
 }

 return(err);
}

/*****************************************************************************/


/* ellipse */


static int cshellipsecheck(int i,int line,csbuffstr * csbfi,curvestr * c,
                                                            int cx,int x,int y)
{
 while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==y)
 {
  if(csbfi->c[line][i].c==c)
  {
   if(csbfi->c[line][i].cx.x>cx && x>cx) return(1);
   else
   if(csbfi->c[line][i].cx.x<cx && x<cx) return(1);
   else
   if(csbfi->c[line][i].cx.x==cx && x==cx) return(1);
  }
  i++;
 }

#ifdef DBUG
 dprintf(0,"cshe OK cx=%d x=%d y=%d",cx,x,y);
#endif

 return(0);
}




static os_error * cshellipse(curvestr * c,int y,filestr * file,
                 csbuffstr * csbfo,csbuffstr * csbfi,int indext,int indexb)
{
 os_error    * err;
 int           xlo;
 int           xhi;
 csbuffelement ce;
 int           miny;
 int           maxy;
 int           i;
 int           cx;
 int           cy;
 int           line;


 err=NULL;


 for(line=CSTOP;line<=CSBOT;line++)
 {
  if(line==CSTOP) i=indext;
  else            i=indexb;

  while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==y)
  {
   if(csbfi->c[line][i].c==c)
   {
    ce=csbfi->c[line][i];
    addintersection(csbfo,&ce,line);
   }
   i++;
  }
 }


 newelement(&ce,c);
 miny=c->box.y;
 maxy=c->box.y+c->box.h;
 cx=c->box.x+c->box.w/2;
 cy=c->box.y+c->box.h/2;
 ce.cx.y=y;

 ce.u=0;
 ce.umin=0;
 ce.umax=0x10000;


 if(y>miny && y<maxy)
 {
  err=ellipsewidth(c,y,&xlo,&xhi);

  if(xlo<xhi)
  {
   ce.cx.x=xlo;
   ce.dx=(cy-y);
   ce.dy=(xlo-cx);

   if(!cshellipsecheck(indext,CSTOP,csbfi,c,cx,xlo,y)) 
                                    addintersection(csbfo,&ce,CSTOP);
   if(!cshellipsecheck(indexb,CSBOT,csbfi,c,cx,xlo,y))
                                    addintersection(csbfo,&ce,CSBOT);

   ce.cx.x=xhi;
   ce.dy=(xhi-cx);

   if(!cshellipsecheck(indext,CSTOP,csbfi,c,cx,xhi,y))
                                    addintersection(csbfo,&ce,CSTOP);
   if(!cshellipsecheck(indexb,CSBOT,csbfi,c,cx,xhi,y))
                                    addintersection(csbfo,&ce,CSBOT);
  }
 }
 else
 if(y==miny)
 {                     /* adding a MAX */
  ce.cx.x=cx;
  ce.dy=0;
  ce.dx=(cy-y);

  if(!cshellipsecheck(indexb,CSBOT,csbfi,c,cx,cx,y))
                                   addintersection(csbfo,&ce,CSBOT);
 }
 else
 if(y==maxy)
 {                     /* adding a MIN */
  ce.cx.x=cx;
  ce.dy=0;
  ce.dx=(cy-y);

  if(!cshellipsecheck(indext,CSTOP,csbfi,c,cx,cx,y))
                                   addintersection(csbfo,&ce,CSTOP);
 }

 return(err);
 USE(file);
}






/* round box */

static int cshrboxcheck(int i,int line,csbuffstr * csbfi,curvestr * c,
                                                         int cx,int x,int y)
{
 while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==y)
 {
  if(csbfi->c[line][i].c==c)
  {
   if(csbfi->c[line][i].cx.x>cx && x>cx) return(1);
   else
   if(csbfi->c[line][i].cx.x<cx && x<cx) return(1);
   else
   if(csbfi->c[line][i].cx.x==cx && x==cx) return(1);
  }
  i++;
 }

 return(0);
}




static os_error * cshrbox(curvestr * c,int y,filestr * file,csbuffstr * csbfo,
                          csbuffstr * csbfi,int indext,int indexb)
{
 os_error    * err;
 int           xlo;
 int           xhi;
 int           ylo;
 int           yhi;
 csbuffelement ce;
 int           i;
 int           line;
 int           xradius;
 int           yradius;
 int           cx;


 err=NULL;


 for(line=CSTOP;line<=CSBOT;line++)
 {
  if(line==CSTOP) i=indext;
  else            i=indexb;

  while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==y)
  {
   if(csbfi->c[line][i].c==c)
   {
    ce=csbfi->c[line][i];
    addintersection(csbfo,&ce,line);
   }
   i++;
  }
 }


 newelement(&ce,c);

 ylo=c->box.y;
 yhi=ylo+c->box.h;

 cx=c->box.x+c->box.w/2;

 xradius=c->data.rbox.xradius;
 if(xradius>c->box.w/2) xradius=c->box.w/2;

 yradius=c->data.rbox.yradius;
 if(yradius>c->box.h/2) yradius=c->box.h/2;

 xlo=c->box.x+xradius;
 xhi=xlo+c->box.w-2*xradius;


 ce.cx.y=y;
 ce.dx=0;
 ce.dy=1;


 if(y>ylo && y<yhi)
 {
  err=rboxwidth(c,y,&xlo,&xhi);

  cprintf("rbox width %d %d",xlo,xhi);

  ce.cx.x=xlo;

  if(!cshrboxcheck(indext,CSTOP,csbfi,c,cx,xlo,y))
                                 addintersection(csbfo,&ce,CSTOP);

  if(!cshrboxcheck(indexb,CSBOT,csbfi,c,cx,xlo,y))
                                 addintersection(csbfo,&ce,CSBOT);

  ce.cx.x=xhi;

  if(!cshrboxcheck(indext,CSTOP,csbfi,c,cx,xhi,y))
                                 addintersection(csbfo,&ce,CSTOP);

  if(!cshrboxcheck(indexb,CSBOT,csbfi,c,cx,xhi,y))
                                 addintersection(csbfo,&ce,CSBOT);
 }
 else
 if(y==ylo)
 {
  ce.cx.x=xlo;

  if(!cshrboxcheck(indexb,CSBOT,csbfi,c,cx,xlo,y))
                                 addintersection(csbfo,&ce,CSBOT);

  ce.cx.x=xhi;

  if(!cshrboxcheck(indexb,CSBOT,csbfi,c,cx,xhi,y))
                                 addintersection(csbfo,&ce,CSBOT);
 }
 else
 if(y==yhi)
 {
  ce.cx.x=xlo;

  if(!cshrboxcheck(indext,CSTOP,csbfi,c,cx,xlo,y))
                                 addintersection(csbfo,&ce,CSTOP);

  ce.cx.x=xhi;

  if(!cshrboxcheck(indext,CSTOP,csbfi,c,cx,xhi,y))
                                 addintersection(csbfo,&ce,CSTOP);
 }

 return(err);
 USE(file);
}






static int cshpathcheck(int i,int line,csbuffstr * csbfi,curvestr * c,
                              coordstr * hresult,segstr * seg)

{
 while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==hresult->y)
 {
  if(csbfi->c[line][i].c==c)
  {
   if(seg->index==csbfi->c[line][i].s.index) 
   {
    return(1);
   }
  }
  i++;
 }
 return(0);
}



static int cshbcheck(int i,int line,csbuffstr * csbfi,csbuffelement * ce)
{
 while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==ce->cx.y)
 {
  if(csbfi->c[line][i].c==ce->c)
  {
   if(ce->s.index==csbfi->c[line][i].s.index &&
      ce->umin==csbfi->c[line][i].umin &&
      ce->umax==csbfi->c[line][i].umax) 
   {
#ifdef DBUG
    dprintf(0,"path check faulted line=%d",line);
#endif

    return(1);
   }
  }
  i++;
 }

#ifdef DBUG
 dprintf(0,"path check OK line=%d index=%d umin=%x umax=%x",
                         line,ce->s.index,ce->umin,ce->umax);
#endif

 return(0);
}





static os_error * cshpath(curvestr * c,int y,filestr * file,csbuffstr * csbfo,
                                     csbuffstr * csbfi,int indext,int indexb)
{
 os_error    * err;
 coordstr      ray[2];
 cutstr        result[MAXBCUT];
 coordstr      hresult;
 int           cuts;
 int           n;
 int           type;
 csbuffelement ce;
 segstr        seg;
 int           path;
 int           code;
 int           i;
 int           line;

 err=NULL;


 for(line=CSTOP;line<=CSBOT;line++)
 {
  if(line==CSTOP) i=indext;
  else            i=indexb;

  while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==y)
  {
   if(csbfi->c[line][i].c==c)
   {
    ce=csbfi->c[line][i];
    addintersection(csbfo,&ce,line);
   }
   i++;
  }
 }


 path=c->data.path.path;
 newsegment(&seg);

 ray[0].x=c->box.x-c->box.w;
 ray[0].y=y;
 ray[1].x=c->box.x+2*c->box.w;
 ray[1].y=y;

 newelement(&ce,c);

 while(1)
 {
  nextsegmentpath(&seg,path,file);

#ifdef DBUG2
  dprintf(0,"type=%d",seg.type);
#endif

  if((type=seg.type)==0) break;

  if(type==1) 
  {
   code=hintersectionr(ray,seg.cx,&hresult);

   ce.c=c;
   ce.cx.x=hresult.x;
   ce.cx.y=y;
   ce.s=seg;
   ce.u=0;
   ce.dy=seg.cx[1].y-seg.cx[0].y;
   ce.dx=seg.cx[1].x-seg.cx[0].x;

   if(code==INTER)
   {
    if(!cshpathcheck(indext,CSTOP,csbfi,c,&hresult,&seg)) 
                                         addintersection(csbfo,&ce,CSTOP);
    if(!cshpathcheck(indexb,CSBOT,csbfi,c,&hresult,&seg))
                                         addintersection(csbfo,&ce,CSBOT);
   }
   else
   if(code & INTERCO) /* ray must cut twice at ends of segment */
   {
    /* in flat land */



   }
   else
   if(code & INTER00) /* ray is going through first vertex */
   {
    /* if previous was not flat can work out cut or maxmin   */
    /* if previous was flat, then this marks end of flat bit */

    if(seg.cx[1].y<y)
    {
     if(!cshpathcheck(indext,CSTOP,csbfi,c,&hresult,&seg)) 
                                         addintersection(csbfo,&ce,CSTOP);
    }
    else
    {
     if(!cshpathcheck(indexb,CSBOT,csbfi,c,&hresult,&seg))
                                         addintersection(csbfo,&ce,CSBOT);
    }
   }
   else
   if(code & INTER01) /* ray is going through second vertex */
   {
    /* if next is not flat can work out cut or maxmin */
    /* if next is flat, then this marks start of flat bit */

    if(seg.cx[0].y<y)
    {
     if(!cshpathcheck(indext,CSTOP,csbfi,c,&hresult,&seg)) 
                                         addintersection(csbfo,&ce,CSTOP);
    }
    else
    {
     if(!cshpathcheck(indexb,CSBOT,csbfi,c,&hresult,&seg))
                                         addintersection(csbfo,&ce,CSBOT);
    }
   }
  }
  else
  if(type==2)
  {
   cuts=hbintersectionc(ray,seg.cx,result);

#ifdef DBUG
   dprintf(0,"cuts=%d index=%d y0=%d y3=%d",cuts,seg.index,
              seg.cx[0].y,seg.cx[3].y);
#endif

   if(cuts)
   {
    for(n=0;n<cuts;n++)
    {
     ce.cx.x=result[n].c.x;
     ce.cx.y=y;

     ce.u=result[n].u2;
     ce.umin=result[n].umin2;
     ce.umax=result[n].umax2;

     ce.dx=result[n].dx2;
     ce.d2x=result[n].d2x2;
     ce.dy=result[n].dy2;
     ce.d2y=result[n].d2y2;

     ce.s=seg;

     code=result[n].type;

#ifdef DBUG
     dprintf(0,"cshpath hinter code=%x u=%d x=%d y=%d",code,ce.u,
                                                        ce.cx.x,ce.cx.y);

     if(code)
     {
      dprintf(0,"interX %d index=%d old=%d",code,ce.s.index,ce.s.oldindex);
     }
#endif



     if(code==INTER)
     {
      if(!cshbcheck(indext,CSTOP,csbfi,&ce)) addintersection(csbfo,&ce,CSTOP);

      if(!cshbcheck(indexb,CSBOT,csbfi,&ce)) addintersection(csbfo,&ce,CSBOT);
     }
     else
     if(code & INTERMX2)
     {
      if(ce.d2y>0 && ce.umin<ce.umax)
      {
       if(!cshbcheck(indexb,CSBOT,csbfi,&ce)) addintersection(csbfo,&ce,CSBOT);
      }
      else
      if(ce.d2y<0 && ce.umin<ce.umax)
      {
       if(!cshbcheck(indext,CSTOP,csbfi,&ce)) addintersection(csbfo,&ce,CSTOP);
      }
     }
     else
     if(code & INTER00)
     {
      if(ce.dy<0)
      {
       if(!cshbcheck(indext,CSTOP,csbfi,&ce)) addintersection(csbfo,&ce,CSTOP);
      }
      else
      if(ce.dy>0)
      {
       if(!cshbcheck(indexb,CSBOT,csbfi,&ce)) addintersection(csbfo,&ce,CSBOT);
      }
     }
     else
     if(code & INTER01)
     {
      if(ce.dy>0)
      {
       if(!cshbcheck(indext,CSTOP,csbfi,&ce)) addintersection(csbfo,&ce,CSTOP);
      }
      else
      if(ce.dy<0)
      {
       if(!cshbcheck(indexb,CSBOT,csbfi,&ce)) addintersection(csbfo,&ce,CSBOT);
      }
     }
    }
   }
  }
 }

 return(err);
}





static int cshboxcheck(int i,int line,csbuffstr * csbfi,curvestr * c,
                                                            int x,int y)
{

 while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==y)
 {
  if(csbfi->c[line][i].c==c)
  {
   if(csbfi->c[line][i].cx.x==x) 
   {
  /*  dprintf(0,"cshbox faulted c=%x x=%d y=%d line=%d",c,x,y,line); */

    return(1);
   }
  }
  i++;
 }

/* dprintf(0,"cshbox allowed c=%x x=%d y=%d line=%d",c,x,y,line); */

 return(0);
}




static os_error * cshbox(curvestr * c,int y,filestr * file,csbuffstr * csbfo,
                             csbuffstr * csbfi,int indext,int indexb)
{
 os_error    * err;
 int           xlo;
 int           xhi;
 int           ylo;
 int           yhi;
 csbuffelement ce;
 int           i;
 int           line;

 err=NULL;


 for(line=CSTOP;line<=CSBOT;line++)
 {
  if(line==CSTOP) i=indext;
  else            i=indexb;

  while(i<csbfi->n[line] && csbfi->c[line][i].cx.y==y)
  {
   if(csbfi->c[line][i].c==c)
   {
    ce=csbfi->c[line][i];
    addintersection(csbfo,&ce,line);
   }
   i++;
  }
 }


 newelement(&ce,c);


 ylo=c->box.y;
 yhi=ylo+c->box.h;
 xlo=c->box.x;
 xhi=xlo+c->box.w;


 ce.cx.x=xlo;
 ce.cx.y=y;
 ce.dx=0;
 ce.dy=1;


 if(y>ylo && y<yhi)
 {
  if(!cshboxcheck(indext,CSTOP,csbfi,c,xlo,y))addintersection(csbfo,&ce,CSTOP);
  if(!cshboxcheck(indexb,CSBOT,csbfi,c,xlo,y))addintersection(csbfo,&ce,CSBOT);

  ce.cx.x=xhi;

  if(!cshboxcheck(indext,CSTOP,csbfi,c,xhi,y))addintersection(csbfo,&ce,CSTOP);
  if(!cshboxcheck(indexb,CSBOT,csbfi,c,xhi,y))addintersection(csbfo,&ce,CSBOT);
 }
 else
 if(y==ylo)
 {
  if(!cshboxcheck(indexb,CSBOT,csbfi,c,xlo,y))addintersection(csbfo,&ce,CSBOT);

  ce.cx.x=xhi;

  if(!cshboxcheck(indexb,CSBOT,csbfi,c,xhi,y))addintersection(csbfo,&ce,CSBOT);
 }
 else
 if(y==yhi)
 {
  if(!cshboxcheck(indext,CSTOP,csbfi,c,xlo,y))addintersection(csbfo,&ce,CSTOP);

  ce.cx.x=xhi;

  if(!cshboxcheck(indext,CSTOP,csbfi,c,xhi,y))addintersection(csbfo,&ce,CSTOP);
 }

 return(err);
 USE(file);
}






/* i is the index in csbf of what is causing this line to be calculated */

os_error * cshline(curvestr * c,int y,filestr * file,csbuffstr * csbfo,
                                csbuffstr * csbfi,int indext,int indexb)
{
 os_error * err;

 err=NULL;

 switch(c->type)
 {
      case CBOX:
                cshbox(c,y,file,csbfo,csbfi,indext,indexb);
                break;

     case CPATH:
                cshpath(c,y,file,csbfo,csbfi,indext,indexb);
                break;

     case CBOXR:
                cshrbox(c,y,file,csbfo,csbfi,indext,indexb);
                break;

  case CELLIPSE:
                cshellipse(c,y,file,csbfo,csbfi,indext,indexb);
                break;

 }

 return(err);
}








/*****************************************************************************/

#define SEGCHUNK (16*sizeof(segstr))


os_error * clearsegments(segstr ** segp)
{
 return(flex_free((flex_ptr)segp));
}


os_error * outputsegments(int path,int u1,segstr * seg1,int x1,int y1,
                                   int u2,segstr * seg2,int x2,int y2,
                          segstr ** segp,int * n,int * ssizep,filestr * file)
{
 os_error * err;
 int        ns;
 int        wrapped;
 segstr     seg;
 bezier     a;
 bezier     b;
 bezier     c;
 int        ssize;


#ifdef DBUG2
 dprintf(0,"segments path=%d",path);
 dprintf(0,"x1=%d y1=%d u1=%d",x1,y1,u1);
 dprintf(0,"x2=%d y2=%d u2=%d",x2,y2,u2);
 dprintf(0,"index1=%d index2=%d",seg1->oldindex,seg2->oldindex);
#endif


 err=flex_alloc((flex_ptr)segp,SEGCHUNK);
 if(!err)
 {
  if(seg1->oldindex==seg2->oldindex) /* special case */
  {
   if(seg1->type==1)
   {
    seg.cx[0].x=x1;
    seg.cx[0].y=y1;
    seg.cx[1].x=x2;
    seg.cx[1].y=y2;
    seg.type=1;
    (*segp)[0]=seg;
    *n=1;
    *ssizep=3;
   }
   else
   {
    bezsplit3((bezier *)seg1->cx,&a,&b,&c,u1,u2);

    seg.cx[0].x=x1;
    seg.cx[0].y=y1;

    seg.cx[1]=b.leftc;
    seg.cx[2]=b.rightc;

    seg.cx[3].x=x2;
    seg.cx[3].y=y2;

    if(y1<y2)
    {
     if(seg.cx[1].y<y1) seg.cx[1].y=y1;
     if(seg.cx[2].y>y2) seg.cx[2].y=y2;
    }
    else
    {
     if(seg.cx[1].y>y1) seg.cx[1].y=y1;
     if(seg.cx[2].y<y2) seg.cx[2].y=y2;
    }

    seg.type=2;
    (*segp)[0]=seg;
    *n=1;
    *ssizep=7;
   }
  }
  else
  {
   ns=0;
   ssize=0;

   /* split first segment */

   if(seg1->type==1)
   {
    seg.cx[0].x=x1;
    seg.cx[0].y=y1;
    seg.cx[1]=seg1->cx[1];
    seg.type=1;
    (*segp)[ns++]=seg;
    ssize+=3;
   }
   else
   if(seg1->type==2)
   {
    seg.cx[0].x=x1;
    seg.cx[0].y=y1;


    bezsplit2((bezier *)seg1->cx,&a,&b,u1);
    seg.cx[1]=b.leftc;
    seg.cx[2]=b.rightc;
    seg.cx[3]=b.right;

    if(y1<y2)
    {
     if(seg.cx[1].y<y1) seg.cx[1].y=y1;
    }
    else
    {
     if(seg.cx[1].y>y1) seg.cx[1].y=y1;
    }

    seg.type=2;
    (*segp)[ns++]=seg;
    ssize+=7;
   }

   seg=*seg1;
   wrapped=0;



   while(1)
   {
    nextsegmentpath(&seg,path,file);

    if(seg.type==3)  /* was 0 */
    {
     if(wrapped) break; /* can wrap-but twice would be wrong */
     seg.index=seg.bindex;
     wrapped=1;
    }
    else
    {
     if(seg.oldindex==seg2->oldindex) break;

     err=flex_chunk((flex_ptr)segp,(ns+2)*sizeof(segstr),SEGCHUNK);
     (*segp)[ns++]=seg;

     if(seg.type==1) ssize+=3;
     else            ssize+=7;
    }
   }


   /* split last segment */

   if(seg2->type==1)
   {
    seg.cx[0]=seg2->cx[0];
    seg.cx[1].x=x2;
    seg.cx[1].y=y2;
    seg.type=1;
    (*segp)[ns++]=seg;
    ssize+=3;
   }
   else
   if(seg2->type==2)
   {
    bezsplit2((bezier *)seg2->cx,&a,&b,u2);
    seg.cx[0]=a.left;
    seg.cx[1]=a.leftc;
    seg.cx[2]=a.rightc;

    seg.cx[3].x=x2;
    seg.cx[3].y=y2;

    if(y1<y2)
    {
     if(seg.cx[2].y>y2) seg.cx[2].y=y2;
    }
    else
    {
     if(seg.cx[2].y<y2) seg.cx[2].y=y2;
    }

    seg.type=2;
    (*segp)[ns++]=seg;
    ssize+=7;
   }
   *n=ns;
   *ssizep=ssize;
  }
 }

 return(err);
}



int outputsegmentpath(int *  q,segstr * seg,int n,int backwards)
{
 int   i;
 int * p;

 p=q;

 if(backwards)
 {
  for(i=n-1;i>=0;i--)
  {
   if(seg[i].type==2)
   {
    *q++=DRAWCURVE;
    *q++=seg[i].cx[2].x;
    *q++=seg[i].cx[2].y;
    *q++=seg[i].cx[1].x;
    *q++=seg[i].cx[1].y;
    *q++=seg[i].cx[0].x;
    *q++=seg[i].cx[0].y;
   }
   else
   if(seg[i].type==1)
   {
    *q++=DRAWDRAW;
    *q++=seg[i].cx[0].x;
    *q++=seg[i].cx[0].y;
   }
  }
 }
 else
 {
  for(i=0;i<n;i++)
  {
   if(seg[i].type==2)
   {
    *q++=DRAWCURVE;
    *q++=seg[i].cx[1].x;
    *q++=seg[i].cx[1].y;
    *q++=seg[i].cx[2].x;
    *q++=seg[i].cx[2].y;
    *q++=seg[i].cx[3].x;
    *q++=seg[i].cx[3].y;
   }
   else
   if(seg[i].type==1)
   {
    *q++=DRAWDRAW;
    *q++=seg[i].cx[1].x;
    *q++=seg[i].cx[1].y;
   }
  }
 }


 return(q-p);
}





























#ifdef NEVER





static os_error * pathwidth(curvestr * c,int y,filestr * file,
                                                   int * xlo,int * xhi)
{
 pathstr * p;
 int     * q;
 coordstr  ray[2];
 cutstr    result[MAXBCUT];
 coordstr  output[2];
 int       cuts;
 int       type;
 int       n;
 int       i;
 int       total;
 segstr    seg;

 p=fob_find(c->data.path.path,file);

 q=p->data;
 newsegment(&seg);
 cuts=0;
 total=0;

 ray[0].x=c->box.x-1;
 ray[0].y=y;
 ray[1].x=c->box.x+c->box.w+1;
 ray[1].y=y;

 while(1)
 {
  nextsegment(&seg,q);
  if((type=seg.type)==0) break;

  if(type==1) n=intersection(ray,seg.cx,result);
  else
  if(type==2) n=bintersection(ray,seg.cx,result);
  else        n=0;

  if(n)
  {
   cuts+=n;
   i=0;
   while(total<2 && i<n)
   {
    output[total++]=result[i++].c;
   }
  }
 }

 if(cuts==2)
 {
  *xlo=MIN(output[0].x,output[1].x);
  *xhi=MAX(output[0].x,output[1].x);
 }
 else
 {
  *xlo=*xhi+1;
 }


/* dprintf(1,"cuts=%d xlo=%d xhi=%d",cuts,*xlo,*xhi); */


 return(NULL); 
}



/* input a simply connected curve and a y coordinate */
/* output the curve width at this point              */
/* xlo>xhi pointoutside curve                        */

os_error * curvewidth(filestr * file,curvestr * c,int y,int * xlo,int * xhi)
{
 os_error * err;

 err=NULL;

 if(y>=c->box.y && y<=(c->box.y+c->box.h))
 {
  switch(c->type)
  {
      case CBOX:
                *xlo=c->box.x;
                *xhi=c->box.x+c->box.w;
                break;

     case CBOXR:
                return(rboxwidth(c,y,xlo,xhi));
                break;

  case CELLIPSE:
                return(ellipsewidth(c,y,xlo,xhi));
                break;

     case CPATH:
                return(pathwidth(c,y,file,xlo,xhi));
                break;
  }
 }
 else
 {
  *xlo=1;
  *xhi=0;
 }

 return(err);
}















/* input a simply connected curve and a y coordinate */
/* input the height of curve that we want            */
/* input the span at height y                        */

/* output max height possible                        */
/* output max span   possible                        */


static os_error * boxheight(curvestr * c,boxstr * box)
{
 boxstr minbox;

 getminbox2(&minbox,&c->box,box);
 *box=minbox;
 return(NULL);
}


static os_error * rboxheight(curvestr * c,boxstr * box)
{
 boxstr minbox;
 int    xlo;
 int    xhi;

 getminbox2(&minbox,&c->box,box);

 rboxwidth(c,minbox.y+minbox.h,&xlo,&xhi);

 xlo=MAX(minbox.x,xlo);
 xhi=MIN(minbox.x+minbox.w,xhi);

 minbox.x=xlo;
 minbox.w=xhi-xlo;

 *box=minbox;
 return(NULL);
}


static os_error * ellipseheight(curvestr * c,boxstr * box)
{
 boxstr minbox;
 int    xlo;
 int    xhi;

 getminbox2(&minbox,&c->box,box);

 ellipsewidth(c,minbox.y+minbox.h,&xlo,&xhi);

 xlo=MAX(minbox.x,xlo);
 xhi=MIN(minbox.x+minbox.w,xhi);

 minbox.x=xlo;
 minbox.w=xhi-xlo;

 *box=minbox;
 return(NULL);
}


static os_error * pathheight(filestr * file,curvestr * c,boxstr * box)
{

 USE(file);
 USE(c);
 USE(box);



#ifdef NEVER

 os_error * err;
 pathstr  * p;
 int      * q;
 coordstr   chunk[4];
 coordstr   ray1[2];
 coordstr   ray2[2];
 cutstr     result[MAXBCUT];
 coordstr   output[2];
 int        cuts;
 int        type;
 int        n;
 int        i;
 int        total;
 segstr     seg;

 p=fob_find(c->path,file);

 q=p->data;
 cuts=0;
 total=0;

 ray1[0].x=c->box.x-1;
 ray1[0].y=box->y;
 ray1[1].x=c->box.x+c->box.w+1;
 ray1[1].y=box->y;

 ray2[0].x=c->box.x-1;
 ray2[0].y=box->y+box->h;
 ray2[1].x=c->box.x+c->box.w+1;
 ray2[1].y=box->y+box->h;


 while((type=nextsegment(chunk,&q))!=0)
 {
  if((type=seg.type)==0) break;

  inside=(chunk[0].y>=box->y)&&(chunk[0].y<(box->y+box->h));

  if(type==1) n=intersection(ray,chunk,result);
  else
  if(type==2) n=bintersection(ray,chunk,result);
  else        n=0;

  if(inside && !n)
  {  /* see what impact curve extent has */



  }

 }

 return(err);

#endif


 return(NULL);

}


os_error * curveheight(filestr * file,curvestr * c,boxstr * box)
{
 os_error * err;
 int        y;

 err=NULL;

 y=box->y;
 if(y>=c->box.y && y<=(c->box.y+c->box.h))
 {
  switch(c->type)
  {
      case CBOX:
                err=boxheight(c,box);
                break;

     case CBOXR:
                err=rboxheight(c,box);
                break;

  case CELLIPSE:
                err=ellipseheight(c,box);
                break;

     case CPATH:
                err=pathheight(file,c,box);
                break;
  }
 }
 else
 {
  box->w=box->h=-1;
 }

 return(err);
}



#endif











os_error * curveinit(void)
{

// return(bezinit());

 return(NULL);
}









