/*->h.curve */


typedef struct clipbuff
{
 int * path;
 int   n;
} clipbuff;



typedef struct segstr
{
 coordstr cx[4];
 int      oldindex;
 int      index;
 int      type;
 coordstr bx;
 int      bindex;
} segstr;



typedef struct csbuffelement
{
 coordstr   cx;
 curvestr * c;

 int        dx;   /* 1==increasing u, increases x */
 int        dy;   /* 1==increasing u, increases y */
 int        d2x;
 int        d2y;

 int        u;
 int        umin;
 int        umax;

 segstr     s;

} csbuffelement;


#define CSTOP 0
#define CSBOT 1


typedef struct csbuffstr
{
 int             n[2];
 csbuffelement * c[2];

} csbuffstr;





typedef struct curvestylestr
{
 colourvalue    linecolour;      /* outline colour  */
 colourvalue    fillcolour;      /* fill colour     */
 pathstyle      pstyle;
 fillstr        fill;
 int            spare;

} curvestylestr;





extern       void newsegment(segstr * seg);
extern os_error * nextsegment(segstr * seg,int * q);
extern os_error * nextsegmentpath(segstr * seg,int path,filestr * file);


extern int pathisbox(int * p,boxstr * box);
extern int pathisline(int * p,boxstr * box);

extern int        curveisline(curvestr * c,filestr * file,boxstr * box);
extern os_error * curvetoline(curvestr * c,filestr * file);


extern int ellipsearc(int * p,curvestr * c,int sx,int sy,int ex,int ey,
                                                         int clockwise);

extern os_error * copycurve(curvestr * dest,filestr * dfile,
                                             curvestr * src,filestr * sfile);
extern os_error * deletecurve(curvestr * c,filestr * file);
extern os_error * newcurve(curvestr * c,int type,int ** xpath,int size,
                                                              filestr * file);

extern os_error * makecurve(curvestr * c,int shape,int data,filestr * file);

extern os_error * nullcurve(curvestr * c);

extern os_error * transformcurve(transformstr * tr,curvestr * c,
                                           filestr * file);


extern void correctcurve(transformstr * xnew,transformstr * old, 
                                                     curvestr * curve);

extern os_error * rendercurve(viewredrawstr * v,curvestr * c,
                                    curvestylestr * cstyle,transformstr * t);

extern os_error * truecurvebox(curvestr * c,pathstyle * cstyle,
                               filestr * file,transformstr * tr,boxstr * box);



#define OUTLINEEOR      (1<<16)
#define OUTLINECVFILL   (1<<17)
#define OUTLINECVLINE   (1<<18)
#define OUTLINEWIMPFILL (1<<19)
#define OUTLINEWIMPLINE (1<<20)
#define OUTLINECVBACK   (1<<21)
#define OUTLINEUSEPATH  (1<<22)


extern os_error * outlinecurve(viewredrawstr * v,curvestr * c,int wimpcolour,
                                         colourvalue * cv,transformstr * t);

extern os_error * outlinebox(viewredrawstr * v,curvestr * c,int wimpcolour,
                                         colourvalue * cv,transformstr * t);

extern os_error * outlinesections(viewredrawstr * v,int * p,int wimpcolour,
                                  colourvalue * c,int width,transformstr * t);

extern os_error * outlineline(viewredrawstr * v,curvestr * c,int wimpcolour,
                              colourvalue * cv,transformstr * t,int swop);


extern os_error * outlinecross(viewredrawstr * v,boxstr * box,int wimpcolour,
                                                         colourvalue * cv);



extern int insidecurve(int x,int y,curvestr * c,
                                   transformstr * tr,filestr * file);

extern int nearcurve(int x,int y,int dx,int dy,curvestr * c,
                                        transformstr * tr,filestr * file);

extern void getpathbox(boxstr * box,int * p);
extern os_error *parallelcurve(filestr * file,curvestr * f,curvestr * c,int d);
extern os_error *parallelcurve2(filestr * file,curvestr * f,curvestr * c,
                                                           int d,int * border);

extern void setcurveradius(curvestr * f,int radius,transformstr * tr);

extern os_error * clipvline(curvestr * f,boxstr * box,
                                              filestr * file,clipbuff * buff);

extern os_error * cliphline(curvestr * f,boxstr * box,
                                              filestr * file,clipbuff * buff);

extern os_error * clipopen(clipbuff * buffer);
extern os_error * clipclose(clipbuff * buffer);
extern os_error * clipreset(clipbuff * buffer);




extern os_error * rboxwidth(curvestr * c,int y,int * xlo,int * xhi);
extern int rboxarc(int * p,curvestr * c,int sx,int sy,int ex,int ey,
                                                      int clockwise);


extern void ellipsepath(int * p,curvestr * c,transformstr * t);
extern void linepath(int * p,curvestr * c,int swop);

extern os_error * ellipsewidth(curvestr * c,int y,int * xlo,int * xhi);
extern os_error * curvewidth(filestr * file,curvestr * c,int y,
                                                 int * xlo,int * xhi);

extern os_error * curveheight(filestr * file,curvestr * c,boxstr * box);

extern os_error * curvebox(filestr * file,curvestr * c,int ylo,int yhi);


extern os_error * cextrema(csbuffstr * csbf,curvestr * ccurve,
                                 curvestr * tcurve,int nc,filestr * file);



extern os_error * cintersections(csbuffstr * csbf,curvestr * ccurve1,
                         curvestr * ccurve2,curvestr * tcurve,filestr * file);


extern os_error * createcsbuffer(csbuffstr * csbf);
extern os_error * deletecsbuffer(csbuffstr * csbf);
extern os_error * sortcsbuffer(csbuffstr * csbf);
extern void       clearcsbuffer(csbuffstr * csbf);

extern os_error * cshline(curvestr * c,int y,filestr * file,csbuffstr * csbfo,
                                   csbuffstr * csbfi,int indext,int indexb);


extern os_error * cmpcurves(curvestr * c1,curvestr * c2,
                               int * match,filestr * file);

extern os_error * simplifycsbuffer(csbuffstr * csbf);

extern os_error * curvetopath(curvestr * c,filestr * file);
extern void       curvepath(curvestr * c,int * path);



extern os_error * clearsegments(segstr ** segp);

extern os_error * outputsegments(int path,int u1,segstr * seg1,int x1,int y1,
                                   int u2,segstr * seg2,int x2,int y,
                          segstr ** segp,int * n,int * ssizep,filestr * file);

extern int outputsegmentpath(int *  q,segstr * seg,int n,int backwards);

extern os_error * parallelpath2(filestr * file,int path,int ** spathp,
                                               int * pathk,int ** pathp,int d);

extern int        pathsize(int * p);
extern os_error * pathflex(int ** pathpp,int path,filestr * file);
extern void       pathtransform(int * q,transformstr * tr);

extern int      * clippath;
extern int        getclipcurve(int fp);
extern os_error * setupclipcurve(curvestr * c,transformstr * tr,filestr *file);
extern os_error * setclipcurve(curvestr * c,transformstr * tr,viewredrawstr * v);
extern os_error * clearclipcurve(void);


extern os_error * curveinit(void);


