/*->c.geometry */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "transform.h"


#include "err.h"
#include "xext.h"
#include "poll.h"
#include "temp.h"
#include "alloc.h"
#include "fsx.h"
#include "trans.h"
#include "xmath.h"
#include "colour.h"
#include "dbhi.h"

#include "etc.h"



#include "constants.h"
#include "str.h"
#include "mhi.h"

#include "geometry.h"



/* point (x,y) inside box */

int insidebox(int x,int y,boxstr * box)
{
 return(
        x>=box->x &&
        x<=(box->x+box->w) &&
        y>=box->y &&
        y<=(box->y+box->h)
       );
}


int nearlyinsidebox(int x,int y,int dx,int dy,boxstr * box)
{
 boxstr xbox;

 xbox=*box;

 xbox.x-=dx;
 xbox.w+=2*dx;
 xbox.y-=dy;
 xbox.h+=2*dy;

 return(insidebox(x,y,&xbox));
}


void box2coord(coordstr * coord,boxstr * box)
{
 coord[0].x=box->x;
 coord[0].y=box->y;

 coord[1].x=box->x+box->w;
 coord[1].y=box->y;

 coord[2].x=box->x+box->w;
 coord[2].y=box->y+box->h;

 coord[3].x=box->x;
 coord[3].y=box->y+box->h;
}



void abox2coord(coordstr * coord,aboxstr * box)
{
 coord[0].x=box->x0;
 coord[0].y=box->y0;

 coord[1].x=box->x1;
 coord[1].y=box->y0;

 coord[2].x=box->x1;
 coord[2].y=box->y1;

 coord[3].x=box->x0;
 coord[3].y=box->y1;
}


void box2abox(boxstr * box,aboxstr * abox)
{
 abox->x0=box->x;
 abox->x1=box->x+box->w;
 abox->y0=box->y;
 abox->y1=box->y+box->h;
}


void abox2box(aboxstr * abox,boxstr * box)
{
 box->x=abox->x0;
 box->w=abox->x1-abox->x0;
 box->y=abox->y0;
 box->h=abox->y1-abox->y0;
}


void getmaxbox(boxstr * maxbox,boxstr * box,transformstr * tr)
{
 coordstr coord[4];
 int minx;
 int miny;
 int maxx;
 int maxy;
 int i;

 box2coord(coord,box);
 tr_transformset(coord,4,tr);

 minx=maxx=coord[0].x;
 miny=maxy=coord[0].y;

 for(i=1;i<4;i++)
 {
  if(minx>coord[i].x) minx=coord[i].x;
  else
  if(maxx<coord[i].x) maxx=coord[i].x;

  if(miny>coord[i].y) miny=coord[i].y;
  else
  if(maxy<coord[i].y) maxy=coord[i].y;
 }

 maxbox->x=minx;
 maxbox->y=miny;
 maxbox->w=maxx-minx;
 maxbox->h=maxy-miny;
}



void getmaxabox(aboxstr * maxbox,aboxstr * box,transformstr * tr)
{
 coordstr coord[4];
 int minx;
 int miny;
 int maxx;
 int maxy;
 int i;

 abox2coord(coord,box);
 tr_transformset(coord,4,tr);

 minx=maxx=coord[0].x;
 miny=maxy=coord[0].y;

 for(i=1;i<4;i++)
 {
  if(minx>coord[i].x) minx=coord[i].x;
  else
  if(maxx<coord[i].x) maxx=coord[i].x;

  if(miny>coord[i].y) miny=coord[i].y;
  else
  if(maxy<coord[i].y) maxy=coord[i].y;
 }

 maxbox->x0=minx;
 maxbox->y0=miny;
 maxbox->x1=maxx;
 maxbox->y1=maxy;
}


void getsumabox(aboxstr * maxbox,aboxstr * box)
{
 maxbox->x0=MIN(maxbox->x0,box->x0);
 maxbox->x1=MAX(maxbox->x1,box->x1);
 maxbox->y0=MIN(maxbox->y0,box->y0);
 maxbox->y1=MAX(maxbox->y1,box->y1);
}


void getsumbox(boxstr * maxbox,boxstr * box)
{
 aboxstr maxabox;
 aboxstr abox;

 box2abox(maxbox,&maxabox);
 box2abox(box,&abox);
 getsumabox(&maxabox,&abox);
 abox2box(&maxabox,maxbox);
}



void getmaxinvabox(aboxstr * maxbox,aboxstr * box,transformstr * tr)
{
 coordstr coord[4];
 int minx;
 int miny;
 int maxx;
 int maxy;
 int i;

 abox2coord(coord,box);
 tr_inverseset(coord,4,tr);

 minx=maxx=coord[0].x;
 miny=maxy=coord[0].y;

 for(i=1;i<4;i++)
 {
  if(minx>coord[i].x) minx=coord[i].x;
  else
  if(maxx<coord[i].x) maxx=coord[i].x;

  if(miny>coord[i].y) miny=coord[i].y;
  else
  if(maxy<coord[i].y) maxy=coord[i].y;
 }

 maxbox->x0=minx;
 maxbox->y0=miny;
 maxbox->x1=maxx;
 maxbox->y1=maxy;
}




int getminabox(aboxstr * minbox,aboxstr * box1,aboxstr * box2)
{
 minbox->x0=MAX(box1->x0,box2->x0);
 minbox->x1=MIN(box1->x1,box2->x1);
 minbox->y0=MAX(box1->y0,box2->y0);
 minbox->y1=MIN(box1->y1,box2->y1);

 if(minbox->x1<minbox->x0 || minbox->y0>minbox->y1) return(0);
 else                                               return(1);
}



/* find min area of two  boxes */
/* return 1, overlap 0 don't   */


int getminbox2(boxstr * minbox,boxstr * box1,boxstr * box2)
{
 int x10;
 int x11;
 int x20;
 int x21;
 int y10;
 int y11;
 int y20;
 int y21;
 int x0;
 int x1;
 int y0;
 int y1;

 x10=box1->x;
 x11=box1->x+box1->w;
 x20=box2->x;
 x21=box2->x+box2->w;
 y10=box1->y;
 y11=box1->y+box1->h;
 y20=box2->y;
 y21=box2->y+box2->h;

 x0=MAX(x10,x20);
 x1=MIN(x11,x21);
 y0=MAX(y10,y20);
 y1=MIN(y11,y21);

 minbox->x=x0;
 minbox->y=y0;
 minbox->w=x1-x0;
 minbox->h=y1-y0;

 if(minbox->w<0 || minbox->h<0) return(0);
 else                           return(1);
}





int boxoverlap(boxstr * box1,boxstr * box2)
{
 int x10;
 int x11;
 int x20;
 int x21;
 int y10;
 int y11;
 int y20;
 int y21;

 x10=box1->x;
 x11=box1->x+box1->w;
 x20=box2->x;
 x21=box2->x+box2->w;

 if(x10>x21 || x11<x20) return(0);

 y10=box1->y;
 y11=box1->y+box1->h;
 y20=box2->y;
 y21=box2->y+box2->h;

 if(y10>y21 || y11<y20) return(0);

 return(1);
}



void aboxshift(aboxstr * abox,int xshift,int yshift)
{
 abox->x0+=xshift;
 abox->x1+=xshift;
 abox->y0+=yshift;
 abox->y1+=yshift;
}



/* is box 1 inside box 2 ? */

int boxinbox(boxstr * box1,boxstr * box2)
{
 int x10;
 int x11;
 int x20;
 int x21;
 int y10;
 int y11;
 int y20;
 int y21;

 x10=box1->x;
 x11=box1->x+box1->w;
 x20=box2->x;
 x21=box2->x+box2->w;
 y10=box1->y;
 y11=box1->y+box1->h;
 y20=box2->y;
 y21=box2->y+box2->h;

 return(x10>=x20 && x11<=x21 && y10>=y20 && y11<=y21);
}


int cmpbox(boxstr * b1,boxstr * b2)
{
 return(b1->x==b2->x && b1->y==b2->y &&
        b1->w==b2->w && b1->h==b2->h);
}



/*****************************************************************************/


void getboxbits(boxstr * box,transformstr * tr,boxbitstr * boxbits)
{
 coordstr          coords[4];
 int               temp;

 box2coord(coords,box);
 tr_transformset(coords,4,tr);

 boxbits->orgx=coords[0].x;
 boxbits->orgy=coords[0].y;
 boxbits->org2x=coords[3].x;
 boxbits->org2y=coords[3].y;
 boxbits->width =pythag(coords[1].x-coords[0].x,coords[1].y-coords[0].y);
 boxbits->height=pythag(coords[3].x-coords[0].x,coords[3].y-coords[0].y);

 if(coords[1].y==coords[0].y && coords[1].x==coords[0].x)
 {
  boxbits->angle=scaleatan2(coords[0].x-coords[3].x,coords[3].y-coords[0].y);
  boxbits->skew=0;
 }
 else
 {
  boxbits->angle=scaleatan2(coords[1].y-coords[0].y,coords[1].x-coords[0].x);
  if(coords[3].x==coords[0].x && coords[3].y==coords[0].y)
  {
   boxbits->skew=0;
  }
  else
  {
   temp=scaleatan2(coords[3].x-coords[0].x,coords[3].y-coords[0].y);
   boxbits->skew=boxbits->angle+temp;
  }
 }

 while(boxbits->skew>=(180*0x10000))  boxbits->skew-=360*0x10000;
 while(boxbits->skew<=(-180*0x10000)) boxbits->skew+=360*0x10000;

 /* angle of 180 can give skew of 540 (180+360) */

 if(boxbits->skew>90*0x10000 || boxbits->skew<-90*0x10000)
 {
  boxbits->height=-boxbits->height;
 /* boxbits->angle+=180*0x10000; */
  boxbits->skew+=180*0x10000;
 }

/*
 if(((coords[1].x-coords[0].x)^(coords[3].y-coords[1].y))<0)
 {
  boxbits->width=-boxbits->width;
  boxbits->angle+=180*0x10000;
  boxbits->skew+=180*0x10000;
 }
*/

 boxbits->cx=(coords[0].x+coords[2].x)/2;
 boxbits->cy=(coords[0].y+coords[2].y)/2;
}



static boxbitstr oldboxbits;
static boxbitstr newboxbits;


void setboxtrans(boxbitstr * old,boxbitstr * xnew)
{
 oldboxbits=*old;
 newboxbits=*xnew;
}


/* not happy with this, maybe should set absolute values for transform */
/* there is an assumption in what follows that c is what the t         */
/* transform applies to                                                */


int boxtrans(transformstr * r,transformstr * t,
                                    curvestr * c,transformstr * ct)
{
 coordstr     centre;
 transformstr temp1;
 transformstr temp2;
 transformstr temp3;
 transformstr rtransform;
 int          type;

 type=0;

 *r=*t;

 if(oldboxbits.orgx!=newboxbits.orgx)
 {
  r->e=t->e+(newboxbits.orgx-oldboxbits.orgx);
  type|=TR_XSHIFT;
 }

 if(oldboxbits.orgy!=newboxbits.orgy)
 {
  r->f=t->f+(newboxbits.orgy-oldboxbits.orgy);
  type|=TR_YSHIFT;
 }

 if(oldboxbits.angle!=newboxbits.angle)
 {
  /* find centre */

  centre.x=c->box.x+c->box.w/2;
  centre.y=c->box.y+c->box.h/2;

  tr_setidentity(&rtransform);
  tr_transformset(&centre,1,ct);     /* now got centre in real coords */

  tr_setrotationa(&rtransform,newboxbits.angle-oldboxbits.angle);

  temp1=*r;
  temp1.e-=centre.x;
  temp1.f-=centre.y;

  tr_multiply(r,&temp1,&rtransform);

  r->e+=centre.x;
  r->f+=centre.y;
  type|=TR_ROTATE;
 }

 if(oldboxbits.skew!=newboxbits.skew)
 {
  transformbitstr b;

  tr_setidentity(&rtransform);
  rtransform.c=tr_tan(newboxbits.skew,0x10000)-tr_tan(oldboxbits.skew,0x10000);

  centre.x=c->box.x;
  centre.y=c->box.y;
  tr_transformset(&centre,1,ct);

  tr_getbits(r,&b);
  b.angle=-b.angle;
  tr_setbits(&temp3,&b,TR_ROTATE);

  temp1=*r;
  temp1.e-=centre.x;
  temp1.f-=centre.y;

  tr_multiply(&temp2,&temp1,&temp3);

  tr_multiply(&temp1,&temp2,&rtransform);

  b.angle=-b.angle;
  tr_setbits(&temp3,&b,TR_ROTATE);

  tr_multiply(r,&temp1,&temp3);

  r->e+=centre.x;
  r->f+=centre.y;
  type|=TR_SKEWXY;
 }

 if(oldboxbits.width!=newboxbits.width)   type|=TR_XSCALE;
 if(oldboxbits.height!=newboxbits.height) type|=TR_YSCALE;

 if(type & TR_SCALE)
 {
  tr_setscale(&rtransform,oldboxbits.width,newboxbits.width,
                          oldboxbits.height,newboxbits.height);

  tr_setshift(&temp1,-c->box.x,-c->box.y);
  tr_multiply(&temp2,&temp1,&rtransform);
  tr_setshift(&temp1,c->box.x,c->box.y);
  tr_multiply(&temp3,&temp2,&temp1);
  temp1=*r;
  tr_multiply(r,&temp3,&temp1);
 }

 return(type);
}





void getboxtransform(transformstr * tr,curvestr * c,int skew,int angle)
{
 transformstr tr1;
 transformstr tr2;
 transformstr tr3;
 transformstr tr4;
 transformstr tr5;
 transformstr trr1;
 transformstr trr2;
 transformstr trr3;
 transformstr trr4;


 if(skew || angle)
 {
  tr_setshift(&tr1,-c->box.x,-c->box.y);

  tr_setscale(&tr2,0x10000,0x10000,0x10000,tr_cos(skew,0x10000));

  tr_setidentity(&tr3);
  tr_setskew(&tr3,skew);

  tr_setidentity(&tr4);
  tr_setrotationa(&tr4,angle);

  tr_setshift(&tr5,c->box.x,c->box.y);

  tr_multiply(&trr1,tr,&tr1);
  tr_multiply(&trr2,&trr1,&tr2);
  tr_multiply(&trr3,&trr2,&tr3);
  tr_multiply(&trr4,&trr3,&tr4);
  tr_multiply(tr,&trr4,&tr5);
 }
}



int cmpboxbits(boxbitstr * b1,boxbitstr * b2)
{
 return(memcmp(b1,b2,sizeof(boxbitstr)));
}


/***************************************************************************/


int splitboxes(boxstr * back,boxstr * front,boxstr * boxes)
{
 int      bxlo;
 int      bxhi;
 int      bylo;
 int      byhi;
 int      fxlo;
 int      fxhi;
 int      fylo;
 int      fyhi;
 int      loy;
 int      hiy;
 boxstr * oboxes;

 oboxes=boxes;

 bxlo=back->x;
 bxhi=bxlo+back->w;
 bylo=back->y;
 byhi=bylo+back->h;

 fxlo=front->x;
 fxhi=fxlo+front->w;
 fylo=front->y;
 fyhi=fylo+front->h;


 if(bxhi<fxlo || bxlo>fxhi || bylo>fyhi || byhi<fylo)
 {
  /* don't overlap */
  *boxes=*back;
  return(1);
 }
 else
 {
  if(bylo<fylo) /* top strip */
  {
   *boxes=*back;
   boxes->h=fylo-bylo;
   boxes++;
   loy=fylo;
  }
  else
  {
   loy=bylo;
  }

  if(byhi>fyhi) /* bot strip */
  {
   *boxes=*back;
   boxes->h=byhi-fyhi;
   boxes->y=hiy=fyhi;
   boxes++;
  }
  else
  {
   hiy=byhi;
  }

  if(bxlo<fxlo) /* left strip */
  {
   *boxes=*back;
   boxes->y=loy;
   boxes->w=fxlo-bxlo;
   boxes->h=hiy-loy;
   boxes++;
  }

  if(bxhi>fxhi) /* right strip */
  {
   *boxes=*back;
   boxes->y=loy;
   boxes->w=bxhi-fxhi;
   boxes->x=fxhi;
   boxes->h=hiy-loy;
   boxes++;
  }
 }

 return(boxes-oboxes);
}



