/*->c.xcolour */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <ctype.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "transform.h"
#include "deb.h"

#include "err.h"
#include "etc.h"
#include "bits.h"
#include "xmath.h"
#include "bf.h"
#include "poll.h"
#include "xprocess.h"
#include "dbhi.h"





#include "str.h"
#include "constants.h"

#include "object.h"
#include "file.h"
#include "view.h"
#include "undo.h"

#include "dscript.h"


#include "xredraw.h"

#include "xxcolour.h"
#include "xcolour.h"
#include "chart.h"
#include "cconv.h"
#include "colsep.h"



/* NB MIN SEQ is greater than MAXDEF */

#define MAXDEFCOLOURS 10
#define MINCOLOURSEQ  0x20


/*               3  +    5      +    5     +    5     */

#define MODEL  0,0,0, 0,0,0,0,0, 0,0,0,0,0, 0,0,0,0,0
#define MODELP              0,0, 0,0,0,0,0, 0,0,0,0,0
#define MODEL4P               0, 0,0,0,0,0, 0,0,0,0,0

#define MP DEFOVERPRINT



static colourstr defcolours[MAXDEFCOLOURS]=
{               /*  n  flags          pal */

 "Black",           1, COLOURLOCKED|COLOURPROCESS,  
                                      0,          MOD3LEN,CMYKMODEL,
                                      0,0,0,MAXCV,  MODEL4P,MP,0,

 "White",           2, COLOURLOCKED,  0xFFFFFF00, MOD3LEN,RGBMODEL,
                                      MAXCV,MAXCV,MAXCV,   MODELP,MP,0,

 "Transparent",     3, COLOURLOCKED|COLOURTRANS,
                                      0xFFFFFF00, MOD3LEN,RGBMODEL,
                                      MAXCV,MAXCV,MAXCV,   MODELP,MP,0,

 "Red",             4, 0,             0x0000FF00, MOD3LEN,RGBMODEL,
                                      MAXCV,0x0000,0x0000, MODELP,MP,0,

 "Green",           5, 0,             0xFF0000,   MOD3LEN,RGBMODEL,
                                      0x0000,MAXCV ,0x0000,MODELP,MP,0,

 "Blue",            6, 0,             0xFF000000, MOD3LEN,RGBMODEL,
                                      0x0000,0x0000,MAXCV, MODELP,MP,0,

 "Cyan",            7, COLOURLOCKED|COLOURPROCESS,
                                      0xFFFF0000, MOD3LEN,CMYKMODEL,
                                      MAXCV,0,0,0,  MODEL4P,MP,0,

 "Magenta",         8, COLOURLOCKED|COLOURPROCESS,
                                      0xFF00FF00, MOD3LEN,CMYKMODEL,
                                      0,MAXCV,0,0,  MODEL4P,MP,0,

 "Yellow",          9, COLOURLOCKED|COLOURPROCESS,
                                      0x00FFFF00, MOD3LEN,CMYKMODEL,
                                      0,0,MAXCV,0,  MODEL4P,MP,0,

 "Registration",   10, COLOURLOCKED|COLOURALLPLATES,
                                      0,          MOD4LEN,CMYKMODEL,
                                      MAXCV,MAXCV,MAXCV,MAXCV,MODEL4P,MP,0,
};



#define UCSAVE    0x7000
#define UCDELETE  0x7001
#define UCUPDATES 0x7002
#define UCUPDATEF 0x7003



int renamecolours;







static int cmpcolour(const void * colour1,const void * colour2)
{
 return(((colourstr*)colour1)->n-((colourstr*)colour2)->n);
}


static int cmpcolourdata(const void * c1,const void * c2)
{
 colourstr * colour1=(colourstr*)c1;
 colourstr * colour2=(colourstr*)c2;
 int         code;

           code=(colour1->flags & (~COLOURLOCKED))-
                (colour2->flags & (~COLOURLOCKED));

/* dprintf(0,"cc1 %d",code);  */

 if(!code) code=(colour1->overprintlimit-colour2->overprintlimit);
/* dprintf(0,"cc2 %d",code);  */

 if(!code) code=memcmp(&colour1->model,&colour2->model,sizeof(modelstr));
/* dprintf(0,"cc3 %d",code);  */

 return(code);
}



int cmpcolourdataf(colourvalue * c1,filestr * file1,
                                 colourvalue * c2,filestr * file2)
{
 colourstr colour1;
 colourstr colour2;

 findcolours(c1->data.tint.key,&colour1,file1);
 findcolours(c2->data.tint.key,&colour2,file2);

 return(cmpcolourdata(&colour1,&colour2));
}



os_error * copycolourmap(filestr * file,colourbasestr ** newcolourbase)
{
 os_error      * err;
 colourbasestr * colourbase;
 int             colours;
 int             size;
 int             i;

 if(file)
 {
  colours=file->hdr.colours;

  colourbase=(colourbasestr*)fob_find(colours,file);
  size=sizeof(colourbasestr)+(colourbase->n-1)*sizeof(colourstr);

  err=flex_alloc((flex_ptr)newcolourbase,size);
  if(!err)
  {
   colourbase=(colourbasestr*)fob_find(colours,file);
   XMEMCPYX(*newcolourbase,colourbase,size);
  }
 }
 else
 {
  size=sizeof(colourbasestr)+(MAXDEFCOLOURS-1)*sizeof(colourstr);
  err=flex_alloc((flex_ptr)newcolourbase,size);
  if(!err)
  {
   colourbase=*newcolourbase;
   colourbase->n=MAXDEFCOLOURS;
   colourbase->sequence=MINCOLOURSEQ;
   for(i=0;i<MAXDEFCOLOURS;i++) colourbase->colour[i]=defcolours[i];
  }
 }

 return(err);
}




static os_error * copyfilecolourmap(filestr * sfile,filestr * dfile)
{
 os_error      * err;
 colourbasestr * colourbase;
 int             colours;

 colours=sfile->hdr.colours;

 err=fob_copy((fobjptr)&colourbase,dfile,colours,sfile);
 colours=colourbase->key;

 dfile->hdr.colours=colours;

 return(err);
}




static os_error * savecolour(colourstr * colour,filestr * file)
{
 colourstr temp;
 temp=*colour;
 return(undowritefn(file,UCSAVE,&temp,sizeof(colourstr)));
}


/* tagged to UCDELETE */

static os_error * undeletecolour(filestr * file)
{
 int             key;
 int             colours;
 colourstr     * colour;
 colourstr       temp;
 colourbasestr * colourbase;
 int             n;

 undoread(file,&key,sizeof(int));

 colours=file->hdr.colours;

 colourbase=(colourbasestr*)fob_find(colours,file);

 temp.n=key;
 colour=(colourstr*)bsearch(&temp,&colourbase->colour[0],colourbase->n,
                                             sizeof(colourstr),cmpcolour);

 n=colour-colourbase->colour;
 if(colour)
 {
  temp=*colour;

  memmove(colour,colour+1,(colourbase->n-n-1)*sizeof(colourstr));
  colourbase->n--;
  fob_change(colours,file,sizeof(colourbasestr)+
                 (colourbase->n-1)*sizeof(colourstr),(fobjptr)&colourbase);
  undowritefn(file,UCSAVE,&temp,sizeof(colourstr));
 }

 return(NULL);
}


/* tagged to UCSAVE */

static os_error * unsavecolour(filestr * file)
{
 colourstr       temp;
 colourstr       temp2;
 colourbasestr * colourbase;
 int             colours;
 colourstr     * colour;
 int             n;

 undoread(file,&temp,sizeof(colourstr));

 colours=file->hdr.colours;

 colourbase=(colourbasestr*)fob_find(colours,file);

 colour=(colourstr*)bnearest(&temp,colourbase->colour,colourbase->n,
                                              sizeof(colourstr),cmpcolour);
 n=colour-colourbase->colour;
 if(n<colourbase->n && colour->n==temp.n)
 {
  temp2=*colour;
  *colour=temp;
  undowritefn(file,UCSAVE,&temp2,sizeof(colourstr));
 }
 else
 {
  fob_change(colours,file,sizeof(colourbasestr)+
                       colourbase->n*sizeof(colourstr),(fobjptr)&colourbase);

  memmove(colourbase->colour+n+1,colourbase->colour+n,
                              sizeof(colourstr)*(colourbase->n-n));

  colourbase->colour[n]=temp;
  colourbase->n++;

  undowritefn(file,UCDELETE,&temp.n,sizeof(int));
 }

 return(NULL);
}


/* tagged to UCUPDATES */
/* this function is always called last in undo/redo */

static os_error * unupdatecolours(filestr * file) 
{
 undowritefn(file,UCUPDATEF,NULL,0);
 event(EVENT_COLOURCHANGE,file,NULL);
 refreshfile(file);
 return(NULL);
}


/* tagged to UCUPDATEF */

static os_error * unupdatecolourf(filestr * file)
{
 undowritefn(file,UCUPDATES,NULL,0);
 return(NULL);
}



os_error * updatecolourmap(filestr * file,colourbasestr ** newcolourbase)
{
 os_error      * err;
 colourbasestr * colourbase;
 int             colours;
 int             newsize;
 int             oldsize;
 int             n;
 colourstr     * rs;
 colourstr     * ws;
 colourstr     * ns;
 int             ri;
 int             wi;
 int             ni;
 int             changes;
 int             key;


 err=modfile(file,0);
 if(!err)
 {
  undostart(file);

  undowritefn(file,UCUPDATES,NULL,0);

  colours=file->hdr.colours;

  fob_lock((fobjptr)&colourbase,colours,file);

  newsize=(*newcolourbase)->n;
  oldsize=colourbase->n;
  ri=wi=ni=0;
  changes=0;

  while(ri<oldsize && wi<oldsize && ni<newsize)
  {
   rs=&colourbase->colour[ri];
   ns=&(*newcolourbase)->colour[ni];

   if(ns->n>rs->n)
   {
    /* scrapping a colour */
    savecolour(rs,file);
    ri++;
   }
   else
   if(ns->n==rs->n)
   {
    if(cmpcolourdata(rs,ns) || cstrcmp(rs->name,ns->name))
    {
     /* scrap old colour */
     savecolour(rs,file);
     ws=&colourbase->colour[wi];
     *ws=(*newcolourbase)->colour[ni];
     changes++;
    }
    else
    {
     if(ri!=wi) colourbase->colour[wi]=*rs;
    }

    wi++;
    ri++;
    ni++;
   }
  }  


  /* scrap from r to end of buffer */

  while(ri<oldsize)
  {
   rs=&colourbase->colour[ri];
   savecolour(rs,file);
   ri++;
   changes++;
  }

  n=newsize-ni;
  oldsize=sizeof(colourbasestr)+(wi+n-1)*sizeof(colourstr);
  err=fob_change(colours,file,oldsize,(fobjptr)&colourbase);

  while(n--)
  {
   /* add these colours */
   key=(*newcolourbase)->colour[ni].n;
   undowritefn(file,UCDELETE,&key,sizeof(int));
   colourbase->colour[wi]=(*newcolourbase)->colour[ni];
   wi++;
   ni++;
  }

  colourbase->n=wi;
 
  colourbase->sequence=(*newcolourbase)->sequence;

  colourbase->charthdr=(*newcolourbase)->charthdr;

  event(EVENT_COLOURCHANGE,file,NULL);

  if(changes) err=refreshfile(file);

  fob_unlock((fobjptr)&colourbase,file);

  undowritefn(file,UCUPDATEF,NULL,0);
  undoend(file);
 }

 return(err);
}


os_error * losecolourmap(filestr * file,colourbasestr ** newcolourbase)
{
 flex_free((flex_ptr)newcolourbase);
 return(NULL);
 USE(file);
}


os_error * newcolourmap(filestr * file,int defaults)
{
 os_error      * err;
 int             i;
 int             handle;
 colourbasestr * colourbase;
 int             ncolours;

 if(defaults) ncolours=MAXDEFCOLOURS;
 else         ncolours=0;


 err=fob_create(sizeof(colourbasestr)+(ncolours-1)*sizeof(colourstr),
                                        file,OCOLOUR,(fobjptr)&colourbase);
 if(!err)
 {
  handle=colourbase->key;
  {
   file->hdr.colours=handle;
   err=fob_mod(handle,file,(fobjptr)&colourbase);
   if(!err)
   {
    colourbase->charthdr=defltcharthdr;
    colourbase->n=ncolours;
    colourbase->sequence=MINCOLOURSEQ;
    for(i=0;i<ncolours;i++) colourbase->colour[i]=defcolours[i];
   }
  }
 }
 return(err);
}



os_error * deletecolour(colourbasestr ** colourbase,int n)
{
 os_error * err;
 err=flex_midextend((flex_ptr)colourbase,
          sizeof(colourbasestr)+(n*sizeof(colourstr)),-(int)sizeof(colourstr));
 if(!err) (*colourbase)->n--;

 return(err);
}


os_error * createcolour(colourbasestr ** colourbase)
{
 os_error  * err;
 int         n;
 colourstr * colour;

 n=(*colourbase)->n;

 err=flex_extend((flex_ptr)colourbase,sizeof(colourbasestr)+(n*sizeof(colourstr)));
 if(!err)
 {
  (*colourbase)->n++;
  colour=&(*colourbase)->colour[n];
  memset(colour,0,sizeof(colourstr));
  colour->n=(*colourbase)->sequence++;
 }
 return(err);
}





int palfromcv(colourvalue * cv)
{
 if(cv->data.tintdata.tag==TINTTRANS) return(-1);
 else
 if(cv->data.tintdata.tag==TINTPAL)
 {
  return((cv->data.tintdata.pal1<<17)|(cv->data.tintdata.pal0<<8));
 }
 return(0);
}

void cvfrompal(colourvalue * cv,int pal)
{
 if(pal==-1)
 {
  cv->data.tintdata.tag=TINTTRANS;
  cv->data.tintdata.pal0=0;
  cv->data.tintdata.pal1=0;
 }
 else
 {
  cv->data.tintdata.tag=TINTPAL;
  cv->data.tintdata.pal0=(pal>>8)&0x1FF;
  cv->data.tintdata.pal1=(pal>>17)&0x7FFF;
 }
}



/* pal=(1-tint)*WHITE+tint*pal */

int tintpal(int tint,int pal)
{
 int r;
 int g;
 int b;



 if(pal!=-1)
 {
  r=(pal>>8) & 0xFF;
  g=(pal>>16)& 0xFF;
  b=(pal>>24)& 0xFF;

  r=((MAXTINT-tint)*255+tint*r)/MAXTINT;
  g=((MAXTINT-tint)*255+tint*g)/MAXTINT;
  b=((MAXTINT-tint)*255+tint*b)/MAXTINT;

  pal=(r<<8)|(g<<16)|(b<<24);
 }

//	cprintf("tintpal %x tint %x",pal,tint);

 return(pal);
}




os_error * findcolourkey(int key,colourbasestr * colourbase,colourstr ** csp)
{
 colourstr       temp;
 temp.n=key;
 *csp=(colourstr*)bsearch(&temp,&colourbase->colour[0],colourbase->n,
                                          sizeof(colourstr),cmpcolour);
 return(NULL);
}



os_error * findcoloursp(int key,colourstr ** csp,filestr * file)
{
 colourbasestr * colourbase;
 int             colours;
 colourstr     * colour;
 colourstr       temp;


 temp.n=key;

 if(file)
 {
  colours=file->hdr.colours;
  colourbase=(colourbasestr*)fob_find(colours,file);


  colour=(colourstr*)bsearch(&temp,&colourbase->colour[0],colourbase->n,
                                          sizeof(colourstr),cmpcolour);

  if(colour) *csp=colour;
  else       *csp=&colourbase->colour[0];
 }
 else
 {
  colour=(colourstr*)bsearch(&temp,&defcolours,MAXDEFCOLOURS,
                                          sizeof(colourstr),cmpcolour);
  if(colour) *csp=colour;
  else       *csp=defcolours;
 }

 return(NULL);
}


os_error * findcolours(int key,colourstr * cs,filestr * file)
{
 os_error  * err;
 colourstr * csp;

 err=findcoloursp(key,&csp,file);

 *cs=*csp;

 return(err);
}



int vfindcolour2(viewstr * view,filestr * file,colourvalue * cv)
{
 colourstr * cs;
 int         pal;
 int         transparent;


//	cprintf("vfind %x",*((int*)cv));
//	cprintf("pal0 %x tag %x pal1 %x",cv->data.tintdata.pal0,cv->data.tintdata.tag,cv->data.tintdata.pal1);


 if(cv->data.tintdata.tag==TINTTRANS || cv->data.tintdata.tag==TINTPAL)
 {
//		cprintf("use data");

  pal=palfromcv(cv);
  if(pal==-1) return(-1);
  if(view && view->process)
  {
   ((processfn3)view->process)((int)view,&pal,&pal,1,PROCESST_PAL,
                                                     PROCESST_PAL,NULL,NULL);
  }
  return(pal);
 }
 else
 {
//		cprintf("look up");

  findcoloursp(cv->data.tint.key,&cs,file);
  transparent=(cs->flags & COLOURTRANS) ||
               ((cs->flags & COLOURNOPRINT) && view && view->printview);

  if(transparent) return(-1);

  if(view && view->process)
  {
   ((processfn3)view->process)((int)view,&pal,cv,1,PROCESST_PAL,
                                                  PROCESST_COLVAL,NULL,NULL);
   return(pal);
  }
  else
  {
   return(tintpal(cv->data.tint.tint,colourgetpal(cs,view,file)));
  }
 }
}




int vcolourtrans(viewstr * view,filestr * file,colourvalue * cv)
{
 colourstr * cs;
 int         transparent;
 int         pal;

 if(cv->data.tintdata.tag==TINTTRANS || cv->data.tintdata.tag==TINTPAL)
 {
  pal=palfromcv(cv);
  if(pal!=-1)
  {
   if(view && view->process)
   {
    ((processfn3)view->process)((int)view,&pal,&pal,1,PROCESST_PAL,
                                                      PROCESST_PAL,NULL,NULL);
   }
  }
  transparent=(pal==-1);
 }
 else
 {
  findcoloursp(cv->data.tint.key,&cs,file);
  transparent=((cs->flags & COLOURTRANS) || ((cs->flags & COLOURNOPRINT) && view && view->printview));
  if(!transparent)
  {
   if(view && view->process)
   {
    ((processfn3)view->process)((int)view,&pal,cv,1,PROCESST_PAL,PROCESST_COLVAL,NULL,NULL);
    transparent=(pal==-1);
   }
  }
 }
 return(transparent);
}


int transparentcolour(colourvalue * cv,filestr * file)
{
 colourstr * cs;

 if(cv->data.tintdata.tag==TINTTRANS || cv->data.tintdata.tag==TINTPAL)
 {
  return(cv->data.tintdata.tag==TINTTRANS);
 }
 else
 {
  findcoloursp(cv->data.tint.key,&cs,file);
  return(cs->flags & COLOURTRANS);
 }
}



#ifdef NEVER

int vfindcolour2t(viewstr * view,filestr * file,colourvalue * cv,
                                                       int * transparent)
{
 colourstr * cs;
 int         pal;

 
 if(cv->data.tintdata.tag==TINTTRANS || cv->data.tintdata.tag==TINTPAL)
 {
  pal=palfromcv(cv);
  *transparent=(pal==-1);
  if(pal==-1) return(-1);
  if(view && view->process)
  {
   ((processfn2)view->process)((int)view,&pal,&pal,1,PROCESST_PAL,
                                                     PROCESST_PAL);
  }
  return(pal);
 }
 else
 {
  findcoloursp(cv->data.tint.key,&cs,file);
  *transparent=(cs->flags & COLOURTRANS) ||
               ((cs->flags & COLOURNOPRINT) && view && view->printview);

  if(*transparent) return(-1);

  if(view && view->process)
  {
   ((processfn2)view->process)((int)view,&pal,cv,1,PROCESST_PAL,
                                                  PROCESST_COLVAL);
   return(pal);
  }
  else
  {
   return(tintpal(cv->data.tint.tint,colourgetpal(cs,view,file)));
  }
 }
}

#endif


int vfindcolour(viewstr * view,colourvalue * cv)
{
 return(vfindcolour2(view,view?view->file:NULL,cv));
}



/* nb, cv may be a pointer to something that can move */

os_error * vsetcolourfile(colourvalue * cv,int * transparent,filestr * file)
{
 int pal;

 pal=vfindcolour2(NULL,file,cv);
 if(pal==-1)
 {
  *transparent=1;
  return(NULL);
 }
 else
 {
  *transparent=0;
//  cprintf("vset colour %x",pal);
  return(rsetgcol(pal,0));
 }
}



/* nb, cv may be a pointer to something that can move */

os_error * vsetcolour(colourvalue * cv,int * transparent,viewstr * view)
{
 int pal;

 pal=vfindcolour(view,cv);
 if(pal==-1)
 {
  *transparent=1;
  return(NULL);
 }
 else
 {
  *transparent=0;
//  cprintf("vset colour %x",pal);
  return(rsetgcol(pal,0));
 }
}




int findcolourindex(colourbasestr * colourbase,int colourn)
{
 colourstr   target;
 colourstr * colour;

 target.n=colourn;

 colour=(colourstr*)bsearch(&target,&colourbase->colour[0],colourbase->n,sizeof(colourstr),
                                                                    cmpcolour);
 if(colour) return(colour-&colourbase->colour[0]);
 else       return(0);
}



os_error * findcolourname(filestr * file,colourvalue * c,char * p)
{
 colourbasestr * colourbase;
 int             colours;
 colourstr       target;
 colourstr     * colour;

 colours=file->hdr.colours;

 target.n=c->data.tint.key;

 colourbase=(colourbasestr*)fob_find(colours,file);

 colour=(colourstr*)bsearch(&target,&colourbase->colour[0],colourbase->n,
                                                sizeof(colourstr),cmpcolour);

 if(colour) strcpy(p,colour->name);
 else       *p=0;

 return(NULL);
}



int findcolourbynameb(char * name,colourbasestr * colourbase,colourstr * cs)
{
 int i;

 for(i=0;i<colourbase->n;i++)
 {
  if(!cstrcmp(colourbase->colour[i].name,name))
  {
   if(cs) *cs=colourbase->colour[i];
   return(i);
  }
 }

 return(-1);
}


int findcolourbyname(char * name,filestr * file,colourstr * cs)
{
 colourbasestr * colourbase;
 int             colours;

 colours=file->hdr.colours;
 colourbase=(colourbasestr*)fob_find(colours,file);

 return(findcolourbynameb(name,colourbase,cs));
}





static void nextcolourname(char * name,int * n)
{
 int    len;
 char * p;

 len=strlen(name);

 if(isdigit(name[len-1]) && isdigit(name[len-2]))
 {
  if(*n==0) *n=(name[len-2]-'0')*10+(name[len-1]-'0');
  p=name+(len-2);
 }
 else
 {
  if(len>(COLNAMELEN-3)) len=(COLNAMELEN-3);
  p=name+len;
 }

 *n=(*n)+1;

 sprintf(p,"%02d",*n);
}



/* turn name into a unique colour name */

void newcolournameb(char * name,colourbasestr * colourbase)
{
 int n;

 n=0;

 while(1)
 {
  nextcolourname(name,&n);
  if(findcolourbynameb(name,colourbase,NULL)<0) break;
 }
}


/*
void newcolourname(char * name,filestr * file)
{
 colourbasestr * colourbase;
 int             colours;


 colours=file->hdr.colours;

 colourbase=fob_find(colours,file);

 newcolournameb(name,colourbase);
}
*/


os_error * addcolour(colourstr * cs,filestr * file,int * key,int overwrite)
{
 os_error      * err;
 colourbasestr * colourbase;
 int             colours;
 colourstr       oldcolour;
 colourstr     * csp;

 colours=file->hdr.colours;

 if(overwrite && findcolourbyname(cs->name,file,&oldcolour)>=0)
 {
  err=findcoloursp(oldcolour.n,&csp,file);
  cs->n=oldcolour.n;
  *csp=*cs;
  csp->flags|=(oldcolour.flags & (COLOURLOCKED|COLOURALLPLATES)); // all plates added
  if(key) *key=oldcolour.n;
 }
 else
 {
  colourbase=(colourbasestr*)fob_find(colours,file);

  err=fob_change(colours,file,sizeof(colourbasestr)+
                       colourbase->n*sizeof(colourstr),(fobjptr)&colourbase);
  if(!err)
  {
   cs->n=colourbase->sequence++;
   if(key) *key=cs->n;
   colourbase->colour[colourbase->n++]=*cs;
  }
 }

 return(err);
}



os_error * mapcolour(filestr * sfile,filestr * dfile,colourvalue * cv,int mode)
{
 os_error  * err;
 colourstr   cs;
 colourstr   cd;
 int         cnew;
 int         n;

 err=NULL;


 if(sfile!=dfile)
 {
  err=findcolours(cv->data.tint.key,&cs,sfile);

  if(mode & LOCOVERWRITE)
  {
   addcolour(&cs,dfile,&cnew,1);
  }
  else
  {
   if(findcolourbyname(cs.name,dfile,&cd)>=0) cnew=cd.n;
   else                                       cnew=0;

   if(cnew)
   {
    if(cmpcolourdata(&cs,&cd))
    {
     cs.flags&=~(COLOURLOCKED|COLOURALLPLATES); /* prob of copying black etc. all plates added */

     if(renamecolours)
     {
      n=0;
      while(1)
      {
       nextcolourname(cs.name,&n);
       if(findcolourbyname(cs.name,dfile,&cd)>=0) 
       {
        if(!cmpcolourdata(&cs,&cd))
        {
         cnew=cd.n;
         break;
        }
       }
       else
       {
        err=addcolour(&cs,dfile,&cnew,0);
        break;
       }
      }
     }
    }
   }
   else
   {
    addcolour(&cs,dfile,&cnew,0);
   }
  }
  cv->data.tint.key=cnew;
 }

 return(err);
}




int cmpcolourval(colourvalue * c1,colourvalue * c2)
{
 int temp;

           temp=(c1->data.tint.key-c2->data.tint.key);
 if(!temp) temp=(c1->data.tint.tint-c2->data.tint.tint);
 if(!temp) temp=(c1->data.tint.overprint-c2->data.tint.overprint);

 return(temp);
}



os_error * getcolour(int pal,int trans,colourvalue * cv,filestr * file)
{
 os_error      * err;
 colourbasestr * colourbase;
 int             colours;
 int             i;
 int             n;
 int             r;
 int             g;
 int             b;
 int             R;
 int             G;
 int             B;
 int             rr;
 int             gr;
 int             br;
 int             rp;
 int             gp;
 int             bp;
 int             rx;
 int             gx;
 int             bx;
 colourstr     * cs;
 colourstr       cx;
 int             key;


 err=NULL;

 if(trans)
 {
  cv->data.tint.key=KEYTRANS;
  cv->data.tint.tint=MAXTINT;
 }
 else
 {
  colours=file->hdr.colours;
  colourbase=(colourbasestr*)fob_find(colours,file);
  n=colourbase->n;

  r=(pal>>8)&0xFF;                                                                
  g=(pal>>16)&0xFF;
  b=(pal>>24)&0xFF;

  rp=r=(r*100)/255;
  gp=g=(g*100)/255;
  bp=b=(b*100)/255;

  rx=(r*MAXCV)/100;
  gx=(g*MAXCV)/100;
  bx=(b*MAXCV)/100;

  r=MAXCV-rx;
  g=MAXCV-gx;
  b=MAXCV-bx;


  for(i=0;i<n;i++)           /* skip reg black   */
  {
   cs=&colourbase->colour[i];

   if(cs->n==KEYTRANS || cs->n==KEYREG) continue;   /* skip transparent */


   if(cs->model.no==RGBMODEL)
   {
    R=cs->model.data.quad.type.rgb.r;
    G=cs->model.data.quad.type.rgb.g;
    B=cs->model.data.quad.type.rgb.b;

    if(R==rx && G==gx && B==bx)
    {
     cv->data.tint.key=cs->n;
     cv->data.tint.tint=MAXTINT;
     break;
    }
    else
    if(rx>=R && gx>=G && bx>=B)
    {
     R=MAXCV-R;
     G=MAXCV-G;
     B=MAXCV-B;

     if(!R && !r) rr=-2;
     else
     if(!R) rr=-1;
     else
     if(!r) rr=0;
     else   rr=scale(MAXTINT,r,R);

     if(!G && !g) gr=-2;
     else
     if(!G) gr=-1;
     else
     if(!g) gr=0;
     else   gr=scale(MAXTINT,g,G);

     if(!B && !b) br=-2;
     else
     if(!B) br=-1;
     else
     if(!b) br=0;
     else   br=scale(MAXTINT,b,B);


/*     dprintf(0,"n=%d rr=%x gr=%x br=%x",cs->n,rr,gr,br);
     dprintf(1,"r=%x g=%x b=%x R=%x G=%x B=%x",r,g,b,R,G,B);  */


     if(rr>0)
     {
      if(br==-2) br=rr;
      if(gr==-2) gr=rr;
     }
     else
     if(br>0)
     {
      if(rr==-2) rr=br;
      if(gr==-2) gr=br;
     }
     else
     if(gr>0)
     {
      if(br==-2) br=gr;
      if(rr==-2) rr=gr;
     }

     if(rr==gr && rr==br && rr>0)
     {
      cv->data.tint.key=cs->n;
      cv->data.tint.tint=rr;
      break;
     }
    }
   }
  }

  if(i>=n)    /* add a new colour */
  {
   memset(&cx,0,sizeof(colourstr));
   sprintf(cx.name,"RGB%02d%02d%02d",rp,gp,bp);
   cx.model.no=RGBMODEL;
   cx.model.data.quad.type.rgb.r=rx;
   cx.model.data.quad.type.rgb.g=gx;
   cx.model.data.quad.type.rgb.b=bx;
   cx.model.len=12;

   bx=(bx*255)/MAXCV;
   gx=(gx*255)/MAXCV;
   rx=(rx*255)/MAXCV;

   cx.model.pal=(bx<<24)|(gx<<16)|(rx<<8);

   err=addcolour(&cx,file,&key,0);

   cv->data.tint.key=key;
   cv->data.tint.tint=MAXTINT;
  }
 }

 return(err);
}


static os_error * dsputcoloursub(colourstr * c,buffer * bf)
{
 os_error * err;
 char       temp[256];

 err=bf_printf(bf,"COL_%02x={colour \"%s\" ",c->n,unquotec(c->name,temp));

 if(c->model.no==RGBMODEL)
 {
  bf_printf(bf,"{rgb 0x%x 0x%x 0x%x}",c->model.data.quad.type.rgb.r,
                                      c->model.data.quad.type.rgb.g,
                                      c->model.data.quad.type.rgb.b);
 }
 else
 if(c->model.no==HSVMODEL)
 {
  bf_printf(bf,"{hsv 0x%x 0x%x 0x%x}",c->model.data.quad.type.hsv.h,
                                      c->model.data.quad.type.hsv.s,
                                      c->model.data.quad.type.hsv.v);
 }
 else
 if(c->model.no==CMYKMODEL)
 {
  bf_printf(bf,"{cmyk 0x%x 0x%x 0x%x 0x%x}",c->model.data.quad.type.cmyk.c,
                                            c->model.data.quad.type.cmyk.m,
                                            c->model.data.quad.type.cmyk.y,
                                            c->model.data.quad.type.cmyk.k);
 }

 if(c->flags & COLOURCHART)
 {
  bf_printf(bf,"{chart \"%s\" \"%s\"}",
                         ((chartmodelstr*)&c->model)->charthdr.name,
                         ((chartmodelstr*)&c->model)->chart.name);
 }

 if(c->flags & COLOURTRANS)  bf_printf(bf,"{transparent 1}");
 if(c->flags & COLOURNOPRINT)  bf_printf(bf,"{noprint  1}");
 if(c->flags & COLOURLOCKED) bf_printf(bf,"{lock 1}");
 if(c->flags & COLOURPLATE)  bf_printf(bf,"{spot 1}");
 if(c->flags & COLOURALLPLATES) bf_printf(bf,"{allplates 1}");
 if(c->flags & COLOUROVERPRINT) bf_printf(bf,"{overprint 1}");
 bf_printf(bf,"{overprintlimit 0x%x}",c->overprintlimit);

 bf_printfeol(bf,"}");

 return(err);
}




static os_error * colourdsevent(int eventn,int user,filestr * file,buffer * bf)
{
 os_error      * err;
 colourbasestr * colourbase;
 int             i;
 int             colours;

 colours=file->hdr.colours;

 colourbase=(colourbasestr*)fob_find(colours,file);

 err=bf_printfeoln(bf,2,"// Colours");

 for(i=0;i<colourbase->n;i++)
 {
  dsputcoloursub(colourbase->colour+i,bf);
 }

 bf_printfeol(bf,"");

 return(err);

 USE(user);
 USE(eventn);
}



os_error * dsputcolour(int key,filestr * file,buffer * bf)
{
 os_error  * err;
 colourstr * cs;

 err=findcoloursp(key,&cs,file);
 if(!err) err=dsputcoloursub(cs,bf);

 return(err);
}



/* do all this to cope with colours which have been deleted */

static char cvbuff[48];

char * dscvstring(colourvalue * cv,filestr * file)
{
 char      * p;
 colourstr * cs;

 p=cvbuff;
 findcoloursp(cv->data.tint.key,&cs,file);
 sprintf(p,"{colourvalue COL_%02x 0x%x %d}",cs->n,
                                            cv->data.tint.tint,
                                            cv->data.tint.overprint);

 return(p);
}


char * dscvstring2(colourvalue * cv,filestr * file)
{
 char      * p;
 colourstr * cs;

 p=cvbuff;
 findcoloursp(cv->data.tint.key,&cs,file);
 sprintf(p,"COL_%02x 0x%x %d",cs->n,
                              cv->data.tint.tint,
                              cv->data.tint.overprint);

 return(p);
}





#define UNDOS 4

static undofnstr undos[UNDOS]=
{
 unsavecolour,    NULL, UCSAVE,       NULL,
 undeletecolour,  NULL, UCDELETE,     NULL,
 unupdatecolours, NULL, UCUPDATES,    "{UCOLOUR}",
 unupdatecolourf, NULL, UCUPDATEF,    "{UCOLOUR}",
};





#define DSTRGB         0x201
#define DSTHSV         0x202
#define DSTCMYK        0x203
#define DSTCHART       0x204
#define DSTTRANSPARENT 0x205
#define DSTSPOT        0x206
#define DSTOVERPRINT   0x207
#define DSTOVERPRINTLT 0x208
#define DSTNOPRINT     0x209
#define DSTALLPLATES   0x20A





static os_error * dscolour(int fp,int argc,objectloc * loc)
{
 os_error *    err;
 colourstr     colour;
 int           sp;
 int           nsp;
 dsobjectstr * ds;
 dsobjectstr * ds1;
 int           key;




 err=dsneeddoc(loc);
	if(!err)
	{
  sp=fp+2; /* skip leading type/len */

  memset(&colour,0,sizeof(colourstr));

  while(argc--)
  {
   ds=(dsobjectstr*)(dsstack+sp);
   nsp=sp+ds->len/sizeof(int);
   ds1=(dsobjectstr*)(dsstack+sp+2);

   switch(ds->type)
   {
    case     DSTNOPRINT:
                        colour.flags|=ds1->data.integer?COLOURNOPRINT:0;
                        break;

    case DSTTRANSPARENT:
                        colour.flags|=ds1->data.integer?COLOURTRANS:0;
                        break;

    case        DSTLOCK:
                        colour.flags|=ds1->data.integer?COLOURLOCKED:0;
                        break;

    case   DSTALLPLATES:
                        colour.flags|=ds1->data.integer?COLOURALLPLATES:0;
                        break;


    case        DSTSPOT:
                        colour.flags|=ds1->data.integer?COLOURPLATE:0;
                        break;

    case   DSTOVERPRINT:
                        colour.flags|=ds1->data.integer?COLOUROVERPRINT:0;
                        break;

    case DSTOVERPRINTLT:
                        colour.overprintlimit=ds1->data.integer;
                        break;


    case         DSTSTR:
                        err=xzstrncpye(colour.name,ds->data.string,sizeof(colour.name));
                        break;

    case         DSTRGB:
                        colour.model.no=RGBMODEL;
                        colour.model.len=MOD3LEN;
                        colour.model.data.quad.type.rgb.r=ds1->data.integer;
                        ds1=(dsobjectstr*)(dsstack+sp+2+3);
                        colour.model.data.quad.type.rgb.g=ds1->data.integer;
                        ds1=(dsobjectstr*)(dsstack+sp+2+6);
                        colour.model.data.quad.type.rgb.b=ds1->data.integer;
                        break;

    case         DSTHSV:
                        colour.model.no=HSVMODEL;
                        colour.model.len=MOD3LEN;
                        colour.model.data.quad.type.hsv.h=ds1->data.integer;
                        ds1=(dsobjectstr*)(dsstack+sp+2+3);
                        colour.model.data.quad.type.hsv.s=ds1->data.integer;
                        ds1=(dsobjectstr*)(dsstack+sp+2+6);
                        colour.model.data.quad.type.hsv.v=ds1->data.integer;
                        break;

    case        DSTCMYK:
                        colour.model.no=CMYKMODEL;
                        colour.model.len=MOD4LEN;
                        colour.model.data.quad.type.cmyk.c=ds1->data.integer;
                        ds1=(dsobjectstr*)(dsstack+sp+2+3);
                        colour.model.data.quad.type.cmyk.m=ds1->data.integer;
                        ds1=(dsobjectstr*)(dsstack+sp+2+6);
                        colour.model.data.quad.type.cmyk.y=ds1->data.integer;
                        ds1=(dsobjectstr*)(dsstack+sp+2+9);
                        colour.model.data.quad.type.cmyk.k=ds1->data.integer;
                        break;

   case        DSTCHART:
                        colour.flags|=COLOURCHART;
                        err=xzstrncpye(((chartmodelstr*)&colour.model)->charthdr.name,ds1->data.string,sizeof(((chartmodelstr*)&colour.model)->charthdr.name));
                        ds1=(dsobjectstr*)(dsstack+sp+2+(ds1->len/sizeof(int)));
																								if(!err) err=xzstrncpye(((chartmodelstr*)&colour.model)->chart.name,ds1->data.string,sizeof(((chartmodelstr*)&colour.model)->chart.name));
                        break;

   }

   sp=nsp;
			if(err) break;
  }

		if(!err)
		{
   colour.model.pal=modelgetpalnaive(NULL,loc->file,colour.model.no,
                                               &colour.model.data.quad);

   err=addcolour(&colour,loc->file,&key,1);
 
   ds=(dsobjectstr*)(dsstack+fp);
   ds->data.integer=key;
   ds->len=3*sizeof(int);
		}
	}
 return(err);
}



static os_error * dscname(int fp,int argc,objectloc * loc)
{
 dsobjectstr * ds;
 colourstr     cs;

 ds=(dsobjectstr*)(dsstack+fp+2);

 findcolourbyname(ds->data.string,loc->file,&cs);

 ds=(dsobjectstr*)(dsstack+fp);
 ds->data.integer=cs.n;
 ds->len=3*sizeof(int);

 return(NULL);

 USE(argc);
}





static int dsthreeintlist[]={
                           DSTINT|DSTCOMPULSORY,
                           DSTINT|DSTCOMPULSORY,
                           DSTINT|DSTCOMPULSORY,
                           DSTENDLIST
                          };


static int dsfourintlist[]={
                           DSTINT|DSTCOMPULSORY,
                           DSTINT|DSTCOMPULSORY,
                           DSTINT|DSTCOMPULSORY,
                           DSTINT|DSTCOMPULSORY,
                           DSTENDLIST
                          };


static int dscolourlist[]={
                           DSTENDLIST
                          };



static dscreator dsdefs[]=
{
 0,0,0,0, "colour",  dscolour, 0, DSTCOLOUR,     dscolourlist,
 0,0,0,0, "colourname",dscname,0, DSTCOLOUR,     dsfourintlist,
 0,0,0,0, "rgb",     NULL,     0, DSTRGB,        dsthreeintlist,
 0,0,0,0, "hsv",     NULL,     0, DSTHSV,        dsthreeintlist,
 0,0,0,0, "cmyk",    NULL,     0, DSTCMYK,       dsfourintlist,
 0,0,0,0, "chart",   NULL,     0, DSTCHART,      dsthreeintlist,
 0,0,0,0, "transparent",NULL,  0, DSTTRANSPARENT,dscolourlist,
 0,0,0,0, "spot",    NULL,     0, DSTSPOT,       dscolourlist,
 0,0,0,0, "overprint",NULL,    0, DSTOVERPRINT,  dscolourlist,
 0,0,0,0, "overprintlimit",NULL,0,DSTOVERPRINTLT,dscolourlist,
 0,0,0,0, "noprint",  NULL,     0, DSTNOPRINT,    dscolourlist,
 0,0,0,0, "allplates",NULL,     0, DSTALLPLATES,  dscolourlist,
};


/*****************************************************************************/


static os_error * colourmkstyle(int eventn,int user,filestr * file,
                                                    filestr * ofile)
{
 return(copyfilecolourmap(ofile,file));
 USE(eventn);
 USE(user);
}



static os_error * colournewevent(int eventn,int user,filestr * file,
                                                     filestr * deffile)
{
 if(deffile) return(copyfilecolourmap(deffile,file));
 else        return(newcolourmap(file,1));

 USE(eventn);
 USE(user);
}



static os_error * colourbareevent(int eventn,int user,filestr * file,
                                                      filestr * deffile)
{
 return(newcolourmap(file,1));

 USE(eventn);
 USE(deffile);
 USE(user);
}




os_error * colourinit(void)
{
 os_error * err;

 err=chartinit();
 if(!err) err=undoaddblock(undos,UNDOS);

 if(!err) err=addevent(EVENT_DSSAVE1,(eventfn)colourdsevent,0);
 if(!err) err=addevent(EVENT_MKSTYLE1,(eventfn)colourmkstyle,0);
 if(!err) err=addevent(EVENT_NEWDOC1,(eventfn)colournewevent,0);
 if(!err) err=addevent(EVENT_BAREDOC1,(eventfn)colourbareevent,0);

 dsaddblock(dsdefs,sizeof(dsdefs)/sizeof(dscreator));

 return(err);
}


