#include "stdafx.h"



#include "os.h"
#include "sprite.h"
#include "wimp.h"
#include "bbc.h"
#include "flex.h"

#include "etc.h"
#include "deb.h"
#include "transform.h"

#include "colour.h"
#include "sprx.h"
#include "xmath.h"
#include "wos.h"


#include "oserr.h"
#include "osglue.h"
#include "osgraphics.h"

#include "xprocess.h"

#include "imlib.h"

#include "imrender.h"



typedef struct bmphdrstr
{
 BITMAPINFOHEADER bmi;
 RGBQUAD          pal[256];


} bmphdrstr;




typedef struct spanstr
{
 int dx0;
 int dx1;
 int sx0;
 int sx1;
 int sy0;
 int sy1;
 int n;

} spanstr;





static void osgetmpal(int bpp,RGBQUAD * quad)
{
 int * paltab;

 paltab=(int*)quad;

 paltab[0]=QUADFROMPAL(0xFFFFFF00);
 paltab[(1<<bpp)-1]=QUADFROMPAL(0);
}



static int bpp1leftmask[32]=
{
 0x00000080,
 0x000000C0,
 0x000000E0,
 0x000000F0,
 0x000000F8,
 0x000000FC,
 0x000000FE,
 0x000000FF,
 0x000080FF,
 0x0000C0FF,
 0x0000E0FF,
 0x0000F0FF,
 0x0000F8FF,
 0x0000FCFF,
 0x0000FEFF,
 0x0000FFFF,
 0x0080FFFF,
 0x00C0FFFF,
 0x00E0FFFF,
 0x00F0FFFF,
 0x00F8FFFF,
 0x00FCFFFF,
 0x00FEFFFF,
 0x00FFFFFF,
 0x80FFFFFF,
 0xC0FFFFFF,
 0xE0FFFFFF,
 0xF0FFFFFF,
 0xF8FFFFFF,
 0xFCFFFFFF,
 0xFEFFFFFF,
 0xFFFFFFFF
};



static int bpp1rightmask[33]=
{
 0x01000000,         // right most bit
 0x03000000,
 0x07000000,
 0x0F000000,
 0x1F000000,
 0x3F000000,
 0x7F000000,
 0xFF000000,
 0xFF010000,
 0xFF030000,
 0xFF070000,
 0xFF0F0000,
 0xFF1F0000,
 0xFF3F0000,
 0xFF7F0000,
 0xFFFF0000,
 0xFFFF0100,
 0xFFFF0300,
 0xFFFF0700,
 0xFFFF0F00,
 0xFFFF1F00,
 0xFFFF3F00,
 0xFFFF7F00,
 0xFFFFFF00,
 0xFFFFFF01,
 0xFFFFFF03,
 0xFFFFFF07,
 0xFFFFFF0F,
 0xFFFFFF1F,
 0xFFFFFF3F,
 0xFFFFFF7F,
 0xFFFFFFFF
};






static void setmaskbitsscaledspan(int dx,int dy,int dspanx,int dbpp,int * dbits,int sx,int ex)
{
 int        * pwrowltw;
 int          shiftw;
 int          shiftltw;
 int          dbytewidth;
 int        * pww;
 int          dxlt;
 int          dspanxp;


 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;

 dxlt=dx+dspanx;
 if(dx<sx)
 {
  dx=sx;
 }

 if(dxlt>(ex-1)) dxlt=ex-1;
 dspanxp=dxlt-dx;

 pww=((int*)(((char*)dbits)+dbytewidth*dy))+(dx-sx)/32;
 pwrowltw=pww+(((dx-sx)&0x1F)+(dxlt-dx))/32;
 shiftw=(dx-sx)&0x1F;
 shiftltw=(((dx-sx)&0x1F)+(dxlt-dx))&0x1F;

 if(pww!=pwrowltw)
 {
  *pww=(*pww)|bpp1rightmask[31-shiftw];
  pww++;

  *pwrowltw=(*pwrowltw)|bpp1leftmask[shiftltw];
  pwrowltw--;

  if(pwrowltw>=pww) memset(pww,0xFF,(pwrowltw-pww+1)*sizeof(int));
 }
 else
 {
//  cprintf("left %x right %x %x",bpp1rightmask[31-shiftw],bpp1leftmask[shiftltw],(bpp1rightmask[31-shiftw] & bpp1leftmask[shiftltw]));
//  cprintf("shiftw %d shiftlw %d",shiftw,shiftltw);
  
  *pww=(*pww)|(bpp1rightmask[31-shiftw] & bpp1leftmask[shiftltw]);
 }

}








static void setbitsscaledspan(POINT * pts,int * sbits,int sbpp,int sbytewidth,int dx,int dy,int dspanx,int dbpp,int * dbits,processstr * prc,imagestr * im,int sx,int ex)
{
 char       * pr;
 char       * pw;
 char       * pwrowlt;

 int        * pwrowltw;

 int          shiftr;
 int          shiftw;
 int          shiftltw;
 int          cr;
 int          cw;
 int          crf;

 int          xdisc;
 int          ydisc;
 int          dbytewidth;
 int        * prw;
 int        * pww;

 int          incx;
 int          incy;
 int          sspanx;
 int          sspany;

 int          inc1x;
 int          isum;

 int          dxlt;
 int          dspanxp;
 int          nx;
 int          ny;


 dbytewidth=(((ex-sx)*dbpp+31)&~31)/8;

 sspanx=pts[1].x-pts[0].x;
 if(sspanx>=0)
 {
  inc1x=1;
  incx=sbpp;
 }
 else
 {
  inc1x=-1;
  incx=-sbpp;
  sspanx=-sspanx;
 }

 sspany=pts[1].y-pts[0].y;
 if(sspany>=0) incy=sbytewidth;
 else
 {
  incy=-sbytewidth;
  sspany=-sspany;
 }


 dxlt=dx+dspanx;
 if(dx<sx)
 {
  xdisc=(sx-dx)*sspanx;
  nx=(xdisc/dspanx);
  xdisc-=nx*dspanx;

  ydisc=(sx-dx)*sspany;
  ny=(ydisc/dspanx);
  ydisc-=ny*dspanx;

  dx=sx;
 }
 else
 {
  nx=ny=xdisc=ydisc=0;
 }

 if(dxlt>(ex-1)) dxlt=ex-1;
 dspanxp=dxlt-dx;


 if(sbpp==1)
 {
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+incy*ny))+(pts[0].x+incx*nx)/32;
  shiftr=((pts[0].x+incx*nx) & 0x1F);
  shiftr=(7-(shiftr & 7))+(shiftr & 24);
  incx=inc1x*15;

  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/32;
  shiftw=((dx-sx) & 0x1F);
  shiftw=(7-(shiftw & 7))+(shiftw & 24);
  pwrowltw=pww+(((dx-sx) & 0x1F)+(dxlt-dx+1))/32;
  shiftltw=(((dx-sx) & 0x1F)+(dxlt-dx+1)) & 0x1F;
  shiftltw=(7-(shiftltw & 7))+(shiftltw & 24);

  cr=crf=0;
  cw=0;

  if(incx>0)
  {
   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }
    
    cw|=((cr>>shiftr)&0x1)<<shiftw;

    if(shiftw & 7) shiftw--;
    else           shiftw+=15;

    if(shiftw>=32)
    {
     shiftw=7;
     *pww++=cw;
     cw=0;
    }
   
    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;

     if(shiftr & 7) shiftr--;
     else           shiftr+=15;

     if(shiftr>=32)
     {
      shiftr=7;
      prw++;
      crf=0;
     }
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)((char*)prw+incy);
     crf=0;
    } 
   }
   if(shiftw!=7) *pww=cw;
  }
  else
  {
   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }

    cw|=((cr>>shiftr)&0x1)<<shiftw;

    if(shiftw & 7) shiftw--;
    else           shiftw+=15;

    if(shiftw>=32)
    {
     shiftw=7;
     *pww++=cw;
     cw=0;
    }
   
    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;

     if((shiftr & 7)!=7) shiftr++;
     else                shiftr-=15;

     if(shiftr<0)
     {
      shiftr=24;
      prw--;
      crf=0;
     }
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)((char*)prw+incy);
     crf=0;
    } 
   }

   if(shiftw!=7) *pww=cw;
  }
 }
 else
 if(sbpp==4)
 {
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y+incy*ny))+(pts[0].x*4+incx*nx)/32;
  shiftr=((pts[0].x*4+incx*nx) & 0x1F);
  if(shiftr & 4) shiftr-=4;
  else           shiftr+=4;

  if(inc1x<0)
  {
   if(shiftr & 4) incx=-12;
   else           incx=4;
   isum=-8;
  }
  else
  {
   if(shiftr & 4) incx=-4;
   else           incx=12;
   isum=8;
  }

  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/8;
  shiftw=((dx-sx) & 0x7)*4;
  if(shiftw & 4) shiftw-=4;
  else           shiftw+=4;

  pwrowltw=pww+(((dx-sx) & 0x7)+(dxlt-dx+1))/8;
  shiftltw=((((dx-sx) & 0x7)+(dxlt-dx+1)) & 0x7)*4;
  if(shiftltw & 4) shiftltw-=4;
  else             shiftltw+=4;

  cr=crf=0;
  cw=0;


//  cprintf("shiftr %d incx %d isum %d inc1x %d",shiftr,incx,isum,inc1x);

  while(pww<pwrowltw || shiftw!=shiftltw)
  {
   if(crf==0)
   {
    cr=*prw;
    crf=1;
   }
   
   cw|=((cr>>shiftr)&0xF)<<shiftw;

   if(shiftw & 4) shiftw-=4;
   else           shiftw+=12;

   if(shiftw>=32)
   {
    shiftw=4;
    *pww++=cw;
    cw=0;
   }
   
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    shiftr+=incx;
    incx=isum-incx;
    if(shiftr>=32)
    {
     shiftr=4;
     prw++;
     crf=0;
    }
    else
    if(shiftr<0)
    {
     shiftr=24;
     prw--;
     crf=0;
    }
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    prw=(int*)((char*)prw+incy);
    crf=0;
   } 
  }

  if(shiftw!=4) *pww=cw;
 }
 else
 if(sbpp==8)
 {
//  cprintf("imrender 8 spanx %d",dspanx);

  pr=(char*)sbits+sbytewidth*pts[0].y+pts[0].x;
  pr+=incy*ny+inc1x*nx;
  pw=(char*)dbits+dbytewidth*dy+(dx-sx);
  pwrowlt=pw+(dxlt-dx);

  while(pw<=pwrowlt)
  {
   *pw++=*pr;
    
   xdisc+=sspanx;
   while(xdisc>=dspanx)
   {
    xdisc-=dspanx;
    pr+=inc1x;
   }

   ydisc+=sspany;
   while(ydisc>=dspanx)
   {
    ydisc-=dspanx;
    pr+=incy;
   } 
  }
 }
 else
 if(sbpp==16 || sbpp==15)
 {
//  cprintf("scaledspan 16");
  prw=((int*)((char*)sbits+sbytewidth*pts[0].y))+(pts[0].x*16+incx*nx)/32;
  prw=(int*)((char*)prw+incy*ny);
  shiftr=((pts[0].x*16+incx*nx) & 0x1F);


  pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/2;
  shiftw=((dx-sx) & 0x1)*16;
  pwrowltw=pww+(((dx-sx) & 0x1)+(dxlt-dx+1))/2;
  shiftltw=((((dx-sx) & 0x1)+(dxlt-dx+1)) & 0x1)*16;
		pw=((char*)pww)+shiftw/8;

  crf=cr=0;
  cw=0;
  
  if(prc && prc->mode==PROCESSBLACK) memset(pw,0,(dspanxp+1)*2);
  else
  if(prc && prc->mode==PROCESSWHITE) memset(pw,0xFF,(dspanxp+1)*2);
  else
  {
   while(pww<pwrowltw || shiftw!=shiftltw)
   {
    if(crf==0)
    {
     cr=*prw;
     crf=1;
    }
    
    if(shiftw==0)
    {
     if(shiftr==0)  cw=(cr & 0xFFFF);
     else           cw=((cr>>16) & 0xFFFF);
    }
    else
    {
     if(shiftr==0) cw|=((cr & 0xFFFF)<<16);
     else          cw|=(cr & 0xFFFF0000);
    }

    shiftw+=16;

    if(shiftw>=32)
    {
     shiftw=0;
     *pww++=cw;
    }
   
    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;
     shiftr+=incx;
     if(shiftr>=32)
     {
      shiftr=0;
      prw++;
      crf=0;
     }
     else
     if(shiftr<0)
     {
      shiftr=16;
      prw--;
      crf=0;
     }
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)((char*)prw+incy);
     crf=0;
    } 
   }

   if(shiftw) *pww=cw;


   if(im->usergbmask)
   {
    int          r;
    int          g;
    int          b;
    int          rshift=im->rshift;
    int          gshift=im->gshift;
    int          bshift=im->bshift;
    int          rmask=im->rmask;
    int          gmask=im->gmask;
    int          bmask=im->bmask;
    unsigned int c1;
    int          c2;

//    cprintf("16 bit mask sort rm %x gm %x bm %x rs %x gs %x bs %x",
//            rmask,gmask,bmask,rshift,gshift,bshift);

    pww=((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/2;
    if(shiftltw) pwrowltw++;
    while(pww<pwrowltw)
    {
     c1=(unsigned int)*pww;
     r=(c1 & rmask)>>rshift;
     g=(c1 & gmask)>>gshift;
     b=(c1 & bmask)>>bshift;
     c2=b|(g<<5)|(r<<10);

     c1=c1>>16;
     r=(c1 & rmask)>>rshift;
     g=(c1 & gmask)>>gshift;
     b=(c1 & bmask)>>bshift;
     c2|=(b<<16)|(g<<(5+16))|(r<<(10+16));

     *pww++=c2;
    }
   }

   if(prc && prc->mode==PROCESSFN)
   {
    prc->process(prc->handle,((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/2,
                             ((int*)((char*)dbits+dbytewidth*dy))+(dx-sx)/2,
                             dspanxp+1,
                             PROCESST_BGR16,PROCESST_BGR16,NULL,NULL);
                    
   }
  }
 }
 else
 if(sbpp==24)
 {

// if(dy==2) cprintf("sspanx %d sspany %d dspanx %d",sspanx,sspany,dspanx);

  if(prc && prc->mode==PROCESSBLACK) memset(((char*)dbits)+dbytewidth*dy+(dx-sx)*3,0,(dspanxp+1)*3);
  else
  if(prc && prc->mode==PROCESSWHITE) memset(((char*)dbits)+dbytewidth*dy+(dx-sx)*3,0xFF,(dspanxp+1)*3);
  else
  {
   pr=((char*)sbits)+sbytewidth*pts[0].y+pts[0].x*3;
   
   incx=inc1x*3;
   pr+=incx*nx;
   pr+=incy*ny;

   pw=((char*)dbits)+dbytewidth*dy+(dx-sx)*3;
   pwrowlt=pw+(dxlt-dx)*3;

   cr=-1;

   while(pw<=pwrowlt)
   {
    if(cr==-1)
    {
     cr =(pr[0])<<16;
     cr|=(pr[1])<<8;
     cr|=pr[2];
    }
    
    *pw++=(char)((cr & 0xFF0000)>>16);
    *pw++=(char)((cr & 0xFF00)>>8);
    *pw++=(char)((cr & 0xFF));

    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;
     pr+=incx;
     cr=-1;
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     pr+=incy;
     cr=-1;
    } 
   }


   if(prc && prc->mode==PROCESSFN)
   {
    prc->process(prc->handle,((char*)dbits)+dbytewidth*dy+(dx-sx)*3,
                             ((char*)dbits)+dbytewidth*dy+(dx-sx)*3,
                             dspanxp+1,
                             PROCESST_BGR24,PROCESST_BGR24,NULL,NULL);
                
   }
  }
 }
 else
 if(sbpp==32)
 {

//  cprintf("spbpp 32 class %d",im->ipal.palclass);

//  cprintf("spanx %d spany %d dspanx %d",sspanx,sspany,dspanx);
  if(prc && prc->mode==PROCESSBLACK) memset(((char*)dbits)+dbytewidth*dy+(dx-sx)*4,0,(dspanxp+1)*4);
  else
  if(prc && prc->mode==PROCESSWHITE) memset(((char*)dbits)+dbytewidth*dy+(dx-sx)*4,0xFF,(dspanxp+1)*4);
  else
  {
   prw=(int*)(((char*)sbits)+sbytewidth*pts[0].y+pts[0].x*sizeof(int));
   prw+=inc1x*nx;
   prw=(int*)((char*)prw+incy*ny);

   
   pww=(int*)(((char*)dbits)+dbytewidth*dy+(dx-sx)*sizeof(int));
   pwrowltw=pww+(dxlt-dx);

   cw=cr=0;


   while(pww<=pwrowltw)
   {
    if(cw==0)
    {
     cr=*prw;
     cw=1;
    }

    *pww++=cr;

    xdisc+=sspanx;
    while(xdisc>=dspanx)
    {
     xdisc-=dspanx;
     prw+=inc1x;
     cw=0;
    }

    ydisc+=sspany;
    while(ydisc>=dspanx)
    {
     ydisc-=dspanx;
     prw=(int*)(((char*)prw)+incy);
     cw=0;
    }
   } 

   if(im->usergbmask)
   {
    int          r;
    int          g;
    int          b;
    int          rshift=im->rshift;
    int          gshift=im->gshift;
    int          bshift=im->bshift;
    int          rmask=im->rmask;
    int          gmask=im->gmask;
    int          bmask=im->bmask;
    unsigned int c1;

    cprintf("32 bit mask sort");

    pww=(int*)(((char*)dbits)+dbytewidth*dy+(dx-sx)*sizeof(int));
    while(pww<=pwrowltw)
    {
     c1=(unsigned int)*pww;
     r=(c1 & rmask)>>rshift;
     g=(c1 & gmask)>>gshift;
     b=(c1 & bmask)>>bshift;
  
     *pww++=b|(g<<8)|(r<<16);
    }
   }

   if(prc && prc->mode==PROCESSFN)
   {
    prc->process(prc->handle,(int*)(((char*)dbits)+dbytewidth*dy+(dx-sx)*sizeof(int)),
                             (int*)(((char*)dbits)+dbytewidth*dy+(dx-sx)*sizeof(int)),
                             dspanxp+1,
                             PROCESST_BGR32,im->ipal.palclass==CL_CMYK?PROCESST_CMYK:PROCESST_BGR32,NULL,NULL);
            
   }
   else
   if(prc && prc->cmyk2bgr && im->ipal.palclass==CL_CMYK)
   {
    prc->cmyk2bgr(prc->handle2,(int*)(((char*)dbits)+dbytewidth*dy+(dx-sx)*sizeof(int)),
                               (int*)(((char*)dbits)+dbytewidth*dy+(dx-sx)*sizeof(int)),
                               dspanxp+1,NULL,NULL);
   }
  }
 }
}


/*
            1
           / \
          /   \
         /     \
       0 \      \ 2
          \     /
           \   /
            \ /
             3
*/

static int getspan(POINT * dpts,POINT * spts,int xpix,int ypix,int y,POINT * pts)
{
 int n;

 n=0;

 if(y==pts[0].y && pts[0].y==pts[1].y)
 {
//  cprintf("clause1");
  
  dpts[0].x=pts[0].x;
  dpts[1].x=pts[1].x;
  spts[0].x=0;
  spts[0].y=0;
  spts[1].x=xpix-1;
  spts[1].y=0;
  n=2;
 }
 else
 if(y==pts[1].y && pts[1].y==pts[2].y)
 {
//  cprintf("clause 2");

  dpts[0].x=pts[1].x;
  dpts[1].x=pts[2].x;
  spts[0].x=xpix-1;
  spts[0].y=0;
  spts[1].x=xpix-1;
  spts[1].y=ypix-1;
  n=2;
 
 }
 else
 if(y==pts[2].y && pts[2].y==pts[3].y)
 {
//  cprintf("clause3");

  dpts[0].x=pts[2].x;
  dpts[1].x=pts[3].x;
  spts[0].x=xpix-1;
  spts[0].y=ypix-1;
  spts[1].x=0;
  spts[1].y=ypix-1;
  n=2;
 }
 else
 if(y==pts[3].y && pts[3].y==pts[0].y)
 {
//  cprintf("clause4");

  dpts[0].x=pts[3].x;
  dpts[1].x=pts[0].x;
  spts[0].x=0;
  spts[0].y=ypix-1;
  spts[1].x=0;
  spts[1].y=0;
  n=2;
 }
 else
 {
  if((y>=pts[0].y && y<pts[1].y) || (y<=pts[0].y && y>pts[1].y))
  {
//   cprintf("clause1a");
   
   dpts[n].x=pts[0].x+scale(pts[1].x-pts[0].x,y-pts[0].y,pts[1].y-pts[0].y);
   spts[n].x=scale(xpix-1,y-pts[0].y,pts[1].y-pts[0].y);  // xpix-1
   spts[n].y=0;
   n++;
  }

  if((y>=pts[1].y && y<pts[2].y) || (y<=pts[1].y && y>pts[2].y))
  {
//   cprintf("clause2a");
   
   dpts[n].x=pts[1].x+scale(pts[2].x-pts[1].x,y-pts[1].y,pts[2].y-pts[1].y);
   spts[n].x=xpix-1;
   spts[n].y=scale(ypix-1,y-pts[1].y,pts[2].y-pts[1].y);   // ypix-1
   n++;
  }

  if((y>=pts[2].y && y<pts[3].y) || (y<=pts[2].y && y>pts[3].y))
  {
//   cprintf("clause3a");

   dpts[n].x=pts[3].x+scale(pts[2].x-pts[3].x,y-pts[3].y,pts[2].y-pts[3].y);
   spts[n].x=scale(xpix-1,y-pts[3].y,pts[2].y-pts[3].y);  // xpix-1
   spts[n].y=ypix-1;
   n++;
  }

  if((y>=pts[3].y && y<pts[0].y) || (y<=pts[3].y && y>pts[0].y))
  {
//   cprintf("clause4a");

   dpts[n].x=pts[0].x+scale(pts[3].x-pts[0].x,y-pts[0].y,pts[3].y-pts[0].y);
   spts[n].x=0;
   spts[n].y=scale(ypix-1,y-pts[0].y,pts[3].y-pts[0].y);  // ypix-1
   n++;
  }
 }


 if(n==1)
 {
  dpts[1]=dpts[0];
  spts[1]=spts[0];
 }

 if((n>1) && (dpts[0].x>dpts[1].x))
 {
//  cprintf("span swop");
  SWOP(dpts[0].x,dpts[1].x);
  SWOP(spts[0].x,spts[1].x);
  SWOP(spts[0].y,spts[1].y);
//  dpts[1].x-=1;
 }

 return(n);
}




static void setbitsscaled4(int * sbits,int sbpp,int sbytewidth,int dbpp,int sypix,int * dbits,processstr * pr,imagestr * im,int sx,int ex,int sy,int ey,spanstr * spans)
{
 int   dx;
 int   dspanx;
 int   dy;
// POINT dpts[2];
 POINT spts[2];
// int   n;

// cprintf("setspritesbitsscaled4 sxpix %d dxpix %d sypix %d dypix %d",sxpix,dxpix,sypix,dypix);

// cprintf("x0 %d x1 %d x2 %d x3 %d",pts[0].x,pts[1].x,pts[2].x,pts[3].x);



 for(dy=sy;dy<ey;dy++)
 {
  if(spans[dy-sy].n==2)
  {
   dx=spans[dy-sy].dx0;
   spts[0].x=spans[dy-sy].sx0;
   spts[0].y=spans[dy-sy].sy0;
   spts[1].x=spans[dy-sy].sx1;
   spts[1].y=spans[dy-sy].sy1;

   if(im->topdown)
   {
    spts[0].y=(sypix-1)-spts[0].y;
    spts[1].y=(sypix-1)-spts[1].y;
   }

   dspanx=spans[dy-sy].dx1-dx;
 
   setbitsscaledspan(spts,sbits,sbpp,sbytewidth,dx,(ey-sy-1)-(dy-sy),dspanx,dbpp,dbits,pr,im,sx,ex);
   
  }
 }
}



static void setmaskbitsscaled4(int dbpp,int * dbits,int sx,int ex,int sy,int ey,spanstr * spans)
{
 int   dx;
 int   dspanx;
 int   dy;


 for(dy=sy;dy<ey;dy++)
 {
  if(spans[dy-sy].n==2)
  {
   dx=spans[dy-sy].dx0;

   dspanx=spans[dy-sy].dx1-dx;
 
   setmaskbitsscaledspan(dx,(ey-sy-1)-(dy-sy),dspanx,dbpp,dbits,sx,ex);
  }
 }
}








static void setspans(POINT * pts,int sxpix,int sypix,int sx,int ex,int sy,int ey,spanstr * spans)
{
 int   dy;
 POINT dpts[2];
 POINT spts[2];
 int   n;

 for(dy=sy;dy<ey;dy++)
 {
  n=getspan(dpts,spts,sxpix,sypix,dy,pts);

  spans[dy-sy].n=n;
  if(n==2)
  {
   if(dpts[0].x>=ex || dpts[1].x<sx)
   {
    spans[dy-sy].n=0;
   }
   else
   {
    spans[dy-sy].dx0=dpts[0].x;
    spans[dy-sy].dx1=dpts[1].x;
    spans[dy-sy].sx0=spts[0].x;
    spans[dy-sy].sy0=spts[0].y;
    spans[dy-sy].sx1=spts[1].x;
    spans[dy-sy].sy1=spts[1].y;
   }
  }
 }
}



/*
            1
           / \
          /   \
         /     \
       0 \      \ 2
          \     /
           \   /
            \ /
             3
*/



// we may be here because regular or because processed

os_error * imrenderbox(framestr * frame,hdcstr * hdcptr,POINT * pts,processstr * pr)
{
 os_error      * err;
 int             i;
 int             xmin;
 int             xmax;
 int             ymin;
 int             ymax;
 int             xpix;
 int             ypix;
 int           * bits;
 bmphdrstr       bmi;
 int             sbytewidth;
 int             spbpp;
 int           * dbits;
 HBITMAP         hbm;
 HBITMAP         hbmask;
 HDC             hdcmem;
 HDC             hdcmemmask;
 int             dbpp;
 int             dxpix;
 int             dypix;
 int             x;
 int             y;
 imagestr      * im;
 int             nc;
 graphicssavestr grsaves;
// void          * handle;
 RECT            clipbox;
 int             code;
 spanstr       * spans;
 int             sx;
 int             ex;
 int             sy;
 int             ey;
 int             regular;
 int             w98code;




	w98code=(ISW9X) || (pr && (pr->isprinter==ISPRINTERNORM));	// because PlgBlt does not seem to work well on PCL6 driver
																																									               		// Vredenburch rotated JPEG's test

// int             time=GetTickCount();
 

 err=NULL;

	cprintf("im render box w98code %d",w98code);
	if(pr) cprintf("isprinter %d",pr->isprinter);


 code=GetClipBox(hdcptr->hdc,&clipbox);
 if(code!=NULLREGION)
 {
//  cprintf("botleft %d topright %d",PtVisible(hdcptr->hdc,clipbox.left,clipbox.bottom-1),
//                                   PtVisible(hdcptr->hdc,clipbox.right-1,clipbox.top));

// clipbox dimensions >=left <right >=top <bottom visible
// spts spans are inclusive they run from 0 to xpix-1,ypix-1

  regular=((pts[0].y==pts[3].y && pts[0].x==pts[1].x) ||
           (pts[0].x==pts[3].x && pts[0].y==pts[1].y));
   

//		cprintf("regular %d",regular);

  im=&frame->im[0];

  xmin=xmax=pts[0].x;
  ymin=ymax=pts[0].y;

  for(i=1;i<4;i++)
  {
   xmin=MIN(xmin,pts[i].x);
   xmax=MAX(xmax,pts[i].x);
   ymin=MIN(ymin,pts[i].y);
   ymax=MAX(ymax,pts[i].y);
  }


// cprintf("cb %d left %d right %d top %d bot %d",code,clipbox.left,clipbox.right,clipbox.top,clipbox.bottom);
// cprintf("xmin %d xmax %d ymin %d ymax %d",xmin,xmax,ymin,ymax);

  for(i=0;i<4;i++)
  {
   pts[i].x-=xmin;
   pts[i].y-=ymin;
  }

  x=xmin;
  y=ymin;



  bits=(int*)im[0].data;
  spbpp=im[0].bpp;
  sbytewidth=im[0].wwidth;


// cprintf("spbpp %d",spbpp);

  xpix=im[0].xpix;
  ypix=im[0].ypix;
  dxpix=xmax-xmin+1;    // this is so x,y can go from ==pts[] ==the other pts[]
  dypix=ymax-ymin+1;    // in other words we pts[] includes all the pixels
  
  dbpp=spbpp;

// cprintf("render box xpix %d ypix %d bpp %d bytewidth %d",xpix,ypix,spbpp,sbytewidth);
// cprintf("xmin %d xmax %d ymin %d ymax %d",xmin,xmax,ymin,ymax);


// 	bbc_saveclippath(&handle);
//  bbc_startpath();

  sy=clipbox.top-ymin;
  if(sy<0) sy=0;

  ey=clipbox.bottom-ymin;
  if(ey>dypix) ey=dypix;

  sx=clipbox.left-xmin;
  if(sx<0) sx=0;

  ex=clipbox.right-xmin;
  if(ex>dxpix) ex=dxpix;

// sx,sy are inclusive, i.e. fill pixels at sx,sy

  if(sx<ex && sy<ey)
  {
   graphicssave(hdcptr,&grsaves);
   graphicspixel(hdcptr);


   flex_alloc((flex_ptr)&spans,(ey-sy)*sizeof(spanstr));
  
   setspans(pts,xpix,ypix,sx,ex,sy,ey,spans);
 
// bbc_move(pts[0].x+xmin,pts[0].y+ymin);
// bbc_draw(pts[1].x+xmin,pts[1].y+ymin);
// bbc_draw(pts[2].x+xmin,pts[2].y+ymin);
// bbc_draw(pts[3].x+xmin,pts[3].y+ymin);
// bbc_draw(pts[0].x+xmin,pts[0].y+ymin);

//  bbc_endpath();

//  bbc_gclip(CLIP_AND);

   memset(&bmi,0,sizeof(bmi));

   osgetmpal(1,&bmi.pal[0]);


   bmi.bmi.biWidth=(ex-sx);
   bmi.bmi.biHeight=(ey-sy);
   bmi.bmi.biXPelsPerMeter=0;
   bmi.bmi.biYPelsPerMeter=0;
   bmi.bmi.biBitCount=(unsigned short)1;
   bmi.bmi.biPlanes=1;
   bmi.bmi.biCompression=BI_RGB;
   bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
   bmi.bmi.biSizeImage=0;
   bmi.bmi.biClrImportant=0;
   bmi.bmi.biClrUsed=0;


   if(!regular)
   {
    err=XCreateDIBSection(hdcptr->hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbmask);
    if(!err)
    {
     setmaskbitsscaled4(bmi.bmi.biBitCount,dbits,sx,ex,sy,ey,spans);


					if(w98code)
					{
      hdcmemmask=CreateCompatibleDC(hdcptr->hdc);
      SelectObject(hdcmemmask,hbmask);
      BitBlt(hdcptr->hdc,x+sx,y+sy,(ex-sx),(ey-sy),hdcmemmask,0,0,SRCAND);  // and mask
      DeleteDC(hdcmemmask);
      DeleteObject(hbmask);
					}
    }
   }


   memset(&bmi,0,sizeof(bmi));

 
// cprintf("renderbox bpp %d process %x",frame->im[0].bpp,pr);
// if(pr) cprintf("process mode %d",pr->mode);

   if(frame->im[0].bpp<=8)
   {
    nc=(1<<frame->im[0].bpp);

    if(pr)
    {
     switch(pr->mode)
     {
      case     PROCESSFN:
                         pr->process(pr->handle,&bmi.pal[0],&frame->im[0].ipal.word[0],nc,
                                      PROCESST_BGR32,PROCESST_BGR32,NULL,NULL);
                         break;

      case  PROCESSBLACK:
                         for(i=0;i<nc;i++) *((int*)(&bmi.pal[i]))=0x0;
                         break;
 
      case  PROCESSWHITE:
                         for(i=0;i<nc;i++) *((int*)(&bmi.pal[i]))=0xFFFFFF;
                         break;

                 default:
                         XMEMCPYX(&bmi.pal[0],&frame->im[0].ipal.word[0],sizeof(int)*(1<<frame->im[0].bpp));
                         break;
     }
    }
    else
    {
     XMEMCPYX(&bmi.pal[0],&frame->im[0].ipal.word[0],sizeof(int)*(1<<frame->im[0].bpp));
    }
   }



   bmi.bmi.biWidth=(ex-sx);
   bmi.bmi.biHeight=(ey-sy);
   bmi.bmi.biXPelsPerMeter=0;
   bmi.bmi.biYPelsPerMeter=0;
   bmi.bmi.biBitCount=(unsigned short)dbpp;
   bmi.bmi.biPlanes=1;
   bmi.bmi.biCompression=BI_RGB;
   bmi.bmi.biSize=sizeof(BITMAPINFOHEADER);
   bmi.bmi.biSizeImage=0;
   bmi.bmi.biClrImportant=0;
   bmi.bmi.biClrUsed=0;

   err=XCreateDIBSection(hdcptr->hdc,(BITMAPINFO*)&bmi,DIB_RGB_COLORS,(void**)&dbits,NULL,0,&hbm);
   if(!err)
   {
    setbitsscaled4(bits,spbpp,sbytewidth,dbpp,ypix,dbits,pr,im,sx,ex,sy,ey,spans);
    

    SetStretchBltMode(hdcptr->hdc,COLORONCOLOR);

//  GetObject(hbm,sizeof(BITMAP),&bitmap);

    hdcmem=CreateCompatibleDC(hdcptr->hdc);

    SelectObject(hdcmem,hbm);

				if(w98code || regular)
				{
     BitBlt(hdcptr->hdc,x+sx,y+sy,(ex-sx),(ey-sy),hdcmem,0,0,regular?SRCCOPY:SRCPAINT);
				}
				else
				{
					POINT dpts[3];
					int code;

					dpts[0].x=x+sx;
					dpts[0].y=y+sy;
					dpts[1].x=dpts[0].x+(ex-sx);
					dpts[1].y=dpts[0].y;
					dpts[2].x=dpts[0].x;
					dpts[2].y=dpts[0].y+(ey-sy);

																																				

     code=PlgBlt(hdcptr->hdc,          // handle to destination DC
            dpts, // destination vertices
            hdcmem,           // handle to source DC
            0,            // x-coord of source upper-left corner
            0,            // y-coord of source upper-left corner
            (ex-sx),           // width of source rectangle
            (ey-sy),          // height of source rectangle
            hbmask,      // handle to bitmask
            0,            // x-coord of bitmask upper-left corner
            0             // y-coord of bitmask upper-left corner
           );
//					cprintf("PlgBlt %d",code);

#ifdef NEVER

MaskBlt(
  hdcptr->hdc,     // handle to destination DC
  x+sx,      // x-coord of destination upper-left corner
  y+sy,      // y-coord of destination upper-left corner 
  (ex-sx),      // width of source and destination
  (ey-sy),     // height of source and destination
  hdcmem,      // handle to source DC
  0,       // x-coord of upper-left corner of source
  0,       // y-coord of upper-left corner of source
  hbmask, // handle to monochrome bit mask
  0,       // horizontal offset into mask bitmap
  0,       // vertical offset into mask bitmap
  MAKEROP4(SRCCOPY,SRCPAINT)      // raster operation code
);

#endif
				}

  
    DeleteDC(hdcmem);
    DeleteObject(hbm);

				if(!w98code)
				{
     DeleteObject(hbmask);
				}

   }

// 	bbc_loadclippath(handle);

   flex_free((flex_ptr)&spans);

  	graphicsload(hdcptr,&grsaves);
  }
//  cprintf("imbox ticks %d",GetTickCount()-time);
 }
 return(err);
}




